#Quick PDF Fuzzer (and memtracker) to be used with GPF
#Jared DeMott, VDA Labs, LLC

from pydbg import *
from pydbg.defines import *
import pgraph
import utils
import sys
import os
import time
import struct
import sys
from threading import Thread
import smtplib
import random

class MemTrack(Thread):
    def __init__(self,counter):

        self.counter = counter
        self.hook_safe = False
        self.memcpy_avg = 0
        self.hit_counter = 0
        self.malloc_total = 0
        
        Thread.__init__(self)
        
    ######################################################################    
    def run(self):
        global num_pdfs
                  
        self.dbg = pydbg()
        self.dbg.get_debug_privileges()
        
        # Let's set some breakpoint handlers
        self.dbg.set_callback(EXCEPTION_BREAKPOINT, self.handler_hide)
        self.dbg.set_callback(LOAD_DLL_DEBUG_EVENT, self.dll_load_handler)
        self.dbg.set_callback(EXCEPTION_ACCESS_VIOLATION,self.accessv_handler)
        self.dbg.set_callback(CREATE_PROCESS_DEBUG_EVENT, self.dbg_cleaner)
        
        # Fire up the process
        print "Loading %s" % self.counter
        self.dbg.load("C:\Program Files\Adobe\Reader 8.0\Reader\AcroRd32.exe", ".\_messedup.pdf\\"+str(self.counter)+".pdf")

        #Or attach
        #os.spawnl(os.P_NOWAIT, "C:\Program Files\Adobe\Reader 8.0\Reader\AcroRd32.exe", str(self.counter)+".pdf")
        #os.system("\"C:\Program Files\Adobe\Reader 8.0\Reader\AcroRd32.exe \" "+str(self.counter)+".pdf")
        #look through the process list, if not found, start it
        #look_for = "AcroRd32.exe"
        #attached = False
        #time.sleep(.05)
        #for j in range(10):
        #    for (pid, proc_name) in self.dbg.enumerate_processes():
        #        if proc_name.lower() == look_for.lower():
        #            try:            
        #                self.dbg.attach(pid)
        #                print "Attached to %d" % pid
        #                attached = True
        #            except:
        #                print "Couldn't attach!"
        #            break
        #    if attached:
        #        break
        #    else:
        #        time.sleep(.05)
        
        # Let's run the process, and fire our monitoring thread to terminate the loaded Acro reader in 5 seconds.
        try:
            print "Running attached process..."
            self.dbg.run()
        except:
            print "Failed to run attached process!"
        
        # Let's take an average of the MemCpy to see if we can affect it or not
        try:
            print "MemCpy Hits: %s" % self.hit_counter
            print "MemCpy Sum: %s " % self.memcpy_avg
            print "MemCpy Avg: %s " % (self.memcpy_avg / self.hit_counter) 
            print "\nMalloc Totals: %s " % self.malloc_total
        except:
            print "No MemCpy hits!"
            pass
        
        self.memcpy_avg = 0
        self.hit_counter = 0
        self.dbg = None

        self.counter = self.counter + 1
            
    #####################################################################
    ##    Hooking callbacks
    #####################################################################
    def memcpyEnter(self,dbg,args):
        self.hit_counter = self.hit_counter + 1
        #print "Hit MemCpy => Source: %x => Dest: %x => Size: %x" % (args[0],args[1],args[2])
        self.memcpy_avg = self.memcpy_avg + args[2]
        return DBG_CONTINUE
    
    ######################################################################
    def mallocExit(self,dbg,args,ret):
        #print "Returned this: %x" % ret
        return DBG_CONTINUE
    
    ######################################################################
    def mallocEnter(self,dbg,args):
        self.malloc_total = self.malloc_total + args[0]
        return DBG_CONTINUE
    
    ######################################################################
    # This is our kickoff for the monitoring thread to kill the process
    def dbg_cleaner(self,dbg):
        MonitorDbg(self.dbg).start()
        return DBG_CONTINUE   
     
    #####################################################################
    def handler_hide (self, dbg):

        if dbg.first_breakpoint:
            try:
                dbg.hide_debugger()
            except:
                print "Failed to hide debugger"
            print "Debugger hidden."
            return DBG_CONTINUE

    ######################################################################
    # We need to make sure the msvcrt.dll is loaded before we can install hooks
    def dll_load_handler (self,dbg):
    
        last_dll = dbg.get_system_dll(-1)
        
        if last_dll.name == "msvcrt.dll":
            self.safe_hooks(dbg)
        return DBG_CONTINUE
    
    ######################################################################
    # This is a straight up av_handler, will email me what happened, makes our life easier.
    def accessv_handler(self,dbg):
       crash_bin = utils.crash_binning.crash_binning()
       crash_bin.record_crash(dbg)
       output = "On the %d.pdf file we hit a problem:" % self.counter
       output += crash_bin.crash_synopsis()
       self.notify(output)
    
       logfilename = "log_%d.txt" % self.counter
       logfile = open(logfilename, "w")
       print "---------CRASH %d------------" % x
       logfile.write(output)
       logfile.close()
       print output
       print "---------DONE WITH CRASH %d------------" % x
       self.dbg.terminate_process()

       #set flag
       global found_crash
       found_crash = True
    
       return None

    ######################################################################
    # This is where we define our hooks and callbacks    
    def safe_hooks(self,dbg):    
        
        # Here let's set our hooks on the interesting functions
        self.ntmemcpy = self.dbg.func_resolve("ntdll","memcpy")
        self.msmemcpy = self.dbg.func_resolve("msvcrt","memcpy")
        self.msmalloc = self.dbg.func_resolve("msvcrt", "malloc")
        
        #print "Hooked MS Runtime memcpy at: %x" % self.msmemcpy
        #print "Hooked NTDLL memcpy at: %x" % self.ntmemcpy
        #print "Hooked MS Runtime malloc at %x" % self.msmalloc
            
        self.hooks = utils.hook_container()
        self.hooks.add(self.dbg,self.ntmemcpy,3,self.memcpyEnter,None)
        self.hooks.add(self.dbg,self.msmemcpy,3,self.memcpyEnter,None)
        self.hooks.add(self.dbg,self.msmalloc,1,self.mallocEnter,self.mallocExit)
        
        return DBG_CONTINUE

    ######################################################################
    # Send me an email when an access violation occurs.
    # Note: I have this commented out.
    def notify(self,output):
            
        smtpserver = 'mail.msu.edu'
        recipients = ['demottja@msu.edu',]
        sender = 'demottja@msu.edu'
        mssg = "From:demottja@msu.edu\r\n\r\nTo:\r\n\r\nFile Fuzzer Exploded!.exe"+output
        session = smtplib.SMTP(smtpserver)
        session.sendmail(sender, recipients, mssg)
        session.quit()

###########################################################################################################################        
class MonitorDbg(Thread):
    def __init__(self, dbg):
        self.dbg = dbg
        Thread.__init__(self)
    def run(self):
        time.sleep(5.5)
        global safe_run
        try:
            self.dbg.terminate_process()
            print "Terminated process"
        except:
            print "Failed to terminate process!"
            pass
        
        safe_run = True

safe_run = True    
i = 0
num_pdfs = 2000
#This parameter defines how far away from the baseline memcpy amounts you are willing to go.
#This is handy to reduce logging output. This is a percentage, so 10 = 10%
memcpy_threshold = 10
found_crash = False


while i < num_pdfs:
    MemTrack(i).start()
    time.sleep(7)
    if found_crash == False:
        print "Test complete. Deleting %d.pdf" % i
        try:
            os.remove(".\_messedup.pdf\%d.pdf" % i)
        except:
            print "Unable to find or remove file"
    elif found_crash == True:
        print "Crash detected! Not deleting file %d.pdf" % i
    print "=========================================================================="

    i = i + 1
    found_crash = False
    
    
    
