/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#define FUNCS 0
#define BASIC_BLOCKS 1

struct pools * create_pools(u_int, u_int, u_int, u_int, struct seed *, int, int, int);

char * recieve_type_information(char * host, char * port, int display_level, int print_choice);
struct pools * types_create_pools(u_int total_pools, u_int sessions_per_pool, u_int legs_per_session,
                        u_int toks_per_leg, struct seed * seeds, int sessions_fixed_or_max,
                        int legs_fixed_or_max, int toks_fixed_or_max, char * types);

struct pool * create_next_pool(struct pool * pool, int position);
struct session * create_next_session(struct session * session, int position);
struct leg * create_next_leg(struct leg * leg, int position, int PROXY);
struct seed * create_next_seed(struct seed * seed, int position);

struct seed * read_in_seeds(struct seed * seeds, char * filename);
int typestr_to_num(char * type);

int paimei_send_target_tag(int, int, int, int, int);
int is_paimei_ready(int, int);
int paimei_send_app_err(int, int);
int paimei_send_done(int, int, int);
int paimei_send_cc_option(int, int, int, int);
int paimei_read_cc_session(struct session * session, int paimei_fd, int display_level, int stalk_type);
int paimei_read_cc_pool(struct pool * pool, int paimei_fd, int display_level, int stalk_type);

int play_session(struct session *, int, char *, char *, char *, int, int, int, int, int, int, char *, struct seed *, char *, int);

int evolve_write(int, u_char *, u_int, int *, int *, void *, int, u_int);
int evolve_read(int, int *, int *, int *, int *, struct leg *, u_int, void *, u_int, int, int);

int combine_toks_to_leg(struct leg *, struct seed * seeds, char * types, int proxy);

int print_pools(struct pools *);
int print_pool(struct pool * pool);
int print_session(struct session * session);
int print_leg(struct leg * leg);
int print_evolution(struct pools * pools, int);

struct pools * process_generation(struct pools * pools, int generations, int stalk_type,
        int session_mutation_rate, int pool_crossover_rate, int pool_mutation_rate, int display_level,
	struct seed *, u_int max_sessions_per_pool, u_int max_legs_per_session, u_int max_toks_per_leg,
	int, int, int, char *, int);

struct fitness_list * create_fitness_list(struct session *, int);
int sort_fitness_list(struct fitness_list *, int);
struct fitness_list * new_head(struct fitness_list *, int);
int swap(struct fitness_list *, struct fitness_list *);
int print_fitness_list(struct fitness_list *);
int destroy_fitness_list(struct fitness_list * fitness_list);

int sessions_crossover(struct pool *, struct fitness_list *, int, u_int, u_int, int);
int dup_session(struct session * dup_session, struct session * orig_session, int session_position, int proxy);
struct session * pick_session(struct session *, struct fitness_list *, int);
struct session * breed(struct session *, struct session *, struct session *, int, int, u_int, u_int, int);
int dup_leg(struct leg *, struct leg *, int);
int dup_leg_to(struct leg *, struct leg *, int);
int dup_leg_from(struct leg *, int, struct leg *, int, int, u_int);
int limited_copy_tok(struct Tok *, struct Tok *);
int destroy_sessions(struct session *);
int tok_type_mutation(struct leg *);
int tok_data_mutation(struct leg *, struct seed *);

int fuzz_to_mutate(struct Tok * tok, struct seed *);

int save(MYSQL * mysql, u_int mysql_id, int generation, struct pools * pools, int, int);
struct pools * restore(MYSQL * mysql, u_int mysql_id, int generation, int PROXY);

struct pools * pool_crossover(struct pools *, struct fitness_list *, int display_level, u_int, int,u_int, int, u_int,
		int, struct seed *, char *, int);
int dup_pool(struct pool * dup_pool, struct pool * orig_pool, int pool_position, int proxy);
struct pool * pick_pool(struct pool * pool, struct fitness_list * fitness_list, int size);
struct pool * breed_pools(struct pool * a, struct pool * b, struct pool * new_pool, int position,
                        int size, u_int, int, u_int, int, u_int, int, struct seed *, char *, int);

int add_random_session(struct pool * pool, u_int max_session_per_pool, int, u_int max_legs_per_session, int, u_int max_toks_per_leg, int, struct seed * seeds);
struct session * insert_session(struct pool * pool, int insert_position);

int remove_random_session(struct pool * pool);
int delete_session(struct pool * pool, int delete_position);
int destroy_session(struct session * session);

int add_type_session(struct pool * pool, u_int max_session_per_pool, int sessions_fixed_or_max,
        char * types, struct seed * seeds);

int init_session(struct session * session, int legs_fixed_or_max, int max_legs_per_session, int toks_fixed_or_max,
        int max_toks_per_leg, struct seed * seeds);

int init_type_session(struct session * session, char * types, struct seed * seeds);

struct session * types_breed(struct session * a, struct session * b, struct session * kids, int position, int size, int proxy);

MYSQL * _mysql_query(char * query, int query_len, MYSQL * mysql, char * mysql_host, char *mysql_user,
                char * mysql_passwd, char * dbname);

int delete_all_hits_and_crashes(MYSQL * mysql, char * mysql_host, char * mysql_user, char * mysql_passwd, char * dbname,
        u_int gpf_id, int current_generation, struct pools * pools, int display_level, int stalk_type);

int delete_hits_and_crashes(MYSQL * mysql, char * mysql_host, char * mysql_user, char * mysql_passwd, char * dbname,
        u_int gpf_id, int current_generation, struct pools * pools, int display_level, int stalk_type);

int calculate_fitness(MYSQL * mysql, char * mysql_host, char * mysql_user, char * mysql_passwd, char * dbname,
        u_int gpf_id, int current_generation, struct pools * pools, int display_level, int stalk_type);
