/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*fixed or max?*/
#define FIXED 0
#define MAXI 1

/*pool/session crashed*/
#define NO 0
#define ACCESS_VIOLATION 1
#define OTHER 2

/*session play*/
#define APP_ERR -2
#define FATAL -3

/*session errors*/
#define NO 0
#define YES 1

/*leg errors*/
#define NONE 0
#define WRITE_ERROR 1
#define READ_ERROR_TIMEOUT 2
#define READ_ERROR_DEADSOCKET 3
#define COMBINE_ERROR 4
#define SAVE_ERROR 5

#define READ 0
#define WRITE 1

#define STRING 0                // an ascii string that ends in a 0x00. (also known as a properly terminated string).
#define MIXED 1                 // the data has ascii strings mixed with binary data
#define ASCII 2                 // general ascii text 0x09,0x0a,0x0d 0x20-0x7f
#define ASCII_CMD 3             // i.e. USER for ftp
#define ASCII_CMDVAR 4          // i.e. jared for ftp username
#define ASCII_SPACE 5           // the ' ' between cmd and cmdvar as in "USER jared"
#define ASCII_END 6             // the \r\n after the cmdvar as in "USER jared\r\n", "USER jared\n", or \r\n\r\n like http, etc.
#define SESSION_ID 7            // also called the tranaction ID in DNS. Set by either client or server and must be repected.
#define FLAGS 8                 // describes this particlar message or session
#define PROTO_ID 9              //
#define LEN 10                  // many protocols follow the len|data|len|data... format. tok->data=length of the "covered" toks.
#define HASH 11                 // the next "covered" items are hashed
#define BINARY 12               // general binary data 0x00-0xff
#define BINARY_END 13           // typically a 0x00 after a ascii string of data, but could be other things
#define WIDESTRING_2 14         // a wide (unicode or 2byte) ascii string
#define WIDESTRING_4 15         // a wide (unicode or 4byte) ascii string
#define UNDEFINED 16           

struct pools {
	int total_pools; 		// total number of pools
        struct pool * first_pool; 
};

struct pool {
        int position;  			// identifier in a group of pools
        int total_sessions;		// total number of sessions in this pool
	int hit_funcs;
	int total_funcs;
	double func_fitness;		// func_fitness = hits + diversity
	int hit_blocks;
	int total_blocks;
	double block_fitness;		// block_fitness = hits + diversity
	struct session * first_session;
        struct pool * prev; 
        struct pool * next;
};

/*
notice the session and pool structs are similar.  For some operations/funcs
it may be ok to reuse code.
*/
struct session {
        int position;  			// sequence, or position, in a list of sessions
        int total_legs;  		// current number of legs in this session (this will change over time)
	int hit_funcs;
	int total_funcs;
	double func_fitness;		// func_fitness = hits + diversity
	int hit_blocks;
	int total_blocks;
	double block_fitness;		// block_fitness = hits + diversity
	struct leg * first_leg;
        struct session * prev; 
        struct session * next;
	char error;			// indicates a play error.  Should be set to NO by default;
};

struct leg {
        int position;  			// sequence number, or position, in a chain of pkts
        int total_toks;  		// current number of toks in this leg (this will change over time)
        char type;       		// READ || WRITE
	int data_len;			// total size of this leg (will be zero if read, and have yet to do read)
	char * data;			// either previously read in data, or concatenated toks of this leg
	struct Tok * first_tok;
        struct leg * prev; 
        struct leg * next;
	char error;			// indicates a play error.  Should be set to NONE by default;
};

//need to add a subtype for hashes, or I guess I could reuse a different field
//when this is done, many other funcs will need to be updated
//:mysql:-> are the names for this feild in the mysqldb
struct Tok {
        int dataLen;		// -1 = not yet defined. :mysql:->data_len 
        char * data;		// NULL = not yet defined. 
        char type;     		// UNDEFINED = not yet defined, else see defines above.
        char numeric;  		// -1=not defined; 0=no;1=yes  0x30-0x39
        char bracket;  		// -1=not defined; 0=no;1=yes  [{(<>)}]
        char quote;    		// -1=not defined; 0=no;1=yes  '"`
        char special;  		// -1=not defined; 0=no;1=yes  /~!@#$%^&\*/
        char other;    		// -1=not defined; 0=no;1=yes  .,?|=+-_;:
        int covered;     	// for LEN and HASH;
	u_int leg;       	// -P=pkt num in a given capture stream. -E leg it belongs to. :mysql:->leg_position_id
        u_int tokNum;    	// Token number in a list of Tokens for a given leg. :mysql:->tok_position
        u_int currentTotal;	// Total size of all Tokens so far within a leg. Could be used for other things too.
        struct Tok * prev; 
        struct Tok * next;
};

/*this struct will be used to sort session and pool fitness lists*/
struct fitness_list {
	int position;		//holds the position of the real session/pool (i.e. these are not from 0-n like a sessons/pools)
	double fitness;		//holds the fitness asscoiated with this position
	struct fitness_list * next;
};

/*struct to hold stuff from a .seed file*/
struct seed {
	int position;
	int total_seeds;
	char type;
	int data_len;
	char * data;
	struct seed * next;
	struct seed * prev;
};
