/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com
    Copyright (c) 2007, VDA Labs, LLC
    Jared DeMott, jdemott@vdalabs.com, www.vdalabs.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*
 This is a General Purpose Fuzzing (GPF) suite to fuzz IP networked code.
*/

/*Includes*/
#include "headers.h"
#include "userFuncs.h" //the only globals are here
#include "sig.h"

/*Main code*/
main(int argc, char *argv[])
{
 /*Varible definations*/
 int c;
 int loopedOnce=0; //0=no; 1=yes -  To be sure the fuzzer did something under basic mode

 /*Setup signal handling stuff*/
 signal(SIGINT, Sig);
 signal(SIGQUIT, Quit);
 signal(SIGPIPE, Pipe);

 /*Master switch - Determine the mode we will run in*/
 c = getopt(argc, (char **)argv, "R:FCG:PpEXh:");
 switch (c)
 {
   case 'R':  /*Pure Fuzz*/
 	rand_mode(argc, (char**)argv);
	break; //should never get here
  
   case 'F':  /*FileFuzz Mode*/
 	filefuzz_mode(argc, (char**)argv);
	printf("Fuzzy File dumping complete.\n");
	printf(" Now use the filefuzz.py script to deliver files to the application.\n\n");
	exit(0);
 
   case 'C': /*Convert mode - allows you to convert a libpcap file to a text file*/
	convert_mode(argc, (char**)argv);
	exit (0);

   case 'G': /*GPF*/
	loopedOnce=basic_mode(argc, (char**)argv, optarg);
	break;

   case 'P': /*Pattern GPF*/
	pattern_mode(argc, (char**)argv);
	break; //should never get here

   case 'p': /*Pattern GPF*/
	new_pattern_mode(argc, (char**)argv);
	break; //should never get here

   case 'E': /*Evolve GPF*/
	if( evolve_mode(argc, (char**)argv) == ERROR)
		exit(-1);
	else
		exit (0);

   case 'X': /*Replay the saved pools in a Evolving Mode GPF mysql database*/
	if( replay_evolve_mode(argc, (char**)argv) == ERROR)
		exit(-1);
	else
		exit (0);

   case 'h': /*Print Usage*/
		if ( optarg!=NULL && (strcasecmp("R", optarg))==0 )
			random_usage();
		else if ( optarg!=NULL && (strcasecmp("F", optarg))==0 )
			FileFuzz_usage();
		else if ( optarg!=NULL && (strcasecmp("C", optarg))==0 )
			doctor_usage();
		else if ( optarg!=NULL && (strcasecmp("G", optarg))==0 )
			GPF_usage();
		else if ( optarg!=NULL && (strcasecmp("P", optarg))==0 )
			Pattern_GPF_usage();
		else if ( optarg!=NULL && (strcasecmp("E", optarg))==0 )
			Evolving_GPF_usage();
		else
			main_usage();

   default:  /*The default switch*/
        fprintf(stderr, "{{{Bad args -- Did you choose a mode?}}}\n\n");
        main_usage();
 }
 //from fall out of -G	  
 if(loopedOnce)
    printf("\n\n Successful Finish.  Did we crash your target?\n\n");
 else {
    TIME(); fprintf(stderr,"{{{Didn't loop even once. Check command line params or capture file.}}}\n\n"); }
 exit(0);
}
