/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"


int calculate_fitness(MYSQL * mysql, char * mysql_host, char * mysql_user, char * mysql_passwd, char * dbname,
	u_int gpf_id, int generation, struct pools * pools, int display_level, int stalk_type)
{
	struct pool * pool;
	struct session * session;
	MYSQL_RES * hits;
	char query[1024];
	int query_len = 0;
	int total=0;

	pool = pools->first_pool;
	while ( pool != NULL )
	{
		session = pool->first_session;
		while ( session != NULL )  //on average, what's the total time for one session
		{
			/*count the number of hits for each session*/
			sprintf(query, "SELECT eip FROM hits WHERE gpf_id='%d' \
					AND generation='%d' \
					AND pool='%d' \
					AND session='%d'",
					gpf_id,
					generation,
					pool->position,
					session->position);
			query_len = strlen(query);
			if( _mysql_query(query, query_len, mysql, mysql_host, mysql_user, mysql_passwd, dbname) == NULL)
			{
				printf("In calculate_fitness, _mysql_query returned an error.\n");
				return ERROR;
			}
			if( (hits = mysql_store_result(mysql)) == NULL)
			{
				printf("In calclulate_fitness, mysql_store_result (1) returned an error.\n");
				if (mysql!=NULL){printf("Error %u (%s)\n",mysql_errno(mysql),mysql_error(mysql));}
				return ERROR;
			}
			total = (int)(mysql_num_rows(hits));
			if( stalk_type == FUNCS)
			{
				session->hit_funcs = total;
				session->func_fitness = total;
			}
			else
			{
				session->hit_blocks = total;
				session->block_fitness = total;
			}
			if( display_level==MED || display_level==HIGH)
				printf("Session %d fitness = %d; ", session->position, total);
			session = session->next;
			mysql_free_result(hits);
		}
		/*count the total number of unique hits in the pool, for the pool fitness*/
		sprintf(query, "SELECT DISTINCT eip FROM hits WHERE gpf_id='%d' \
				AND generation='%d' \
				AND pool='%d'",
				gpf_id,
				generation,
				pool->position);
		query_len = strlen(query);
		if( _mysql_query(query, query_len, mysql, mysql_host, mysql_user, mysql_passwd, dbname) == NULL)
		{
			printf("In calculate_fitness, _mysql_query returned an error.\n");
			return ERROR;
		}
		if( (hits = mysql_store_result(mysql)) == NULL)
		{
			printf("In calculate_fitness, mysql_store_result (1) returned an error.\n");
			if (mysql!=NULL){printf("Error %u (%s)\n",mysql_errno(mysql),mysql_error(mysql));}
			return ERROR;
		}
		total = (int)(mysql_num_rows(hits));
		if( stalk_type == FUNCS)
		{
			pool->hit_funcs = total;
			pool->func_fitness = total;
		}
		else
		{	
			pool->hit_blocks = total;
			pool->block_fitness = total;
		}
		if( display_level==MED || display_level==HIGH)
			printf("\nPool %d fitness = %d\n", pool->position, total);
		pool = pool->next;
		mysql_free_result(hits);
	}
	return OK;
}

int calculate_diversity(MYSQL * mysql, char * mysql_host, char * mysql_user, char * mysql_passwd, char * dbname,
	u_int gpf_id, int generation, struct pools * pools, int display_level, int stalk_type)
{
	struct pool * pool;
	struct session * session;
	MYSQL_RES * responce;
	MYSQL_ROW row;
	char query[1024];
	int query_len = 0;
	int best_pool_num=0;
	int best_pool_size=0;
	int best_pool_less_one=0;
	int best_session_num=0;
	int best_session_size=0;
	int best_session_less_one=0;
	int unique=0;

//1. find the most fit pool or session
//2. find unique hits/session when compared to the most fit session
	//Total Fitness = hits + diversity
		//diversity = (UNIQUE/BEST) * (BEST-1)

	if( display_level==MED || display_level==HIGH)
		printf("Calculating diversity boost (if any) for each pool and session:\n");

	//Find the most fit pool
	best_pool_size=0;
	pool = pools->first_pool;
	while ( pool != NULL )
	{
		if( stalk_type == FUNCS)
		{
			if (pool->hit_funcs > best_pool_size)
			{
				best_pool_size = pool->hit_funcs;
				best_pool_num = pool->position;
			}
		}
		else
		{
			if (pool->hit_blocks > best_pool_size)
			{
				best_pool_size = pool->hit_blocks;
				best_pool_num = pool->position;
			}
		}

		pool = pool->next;
	}
	if( best_pool_size == 0 )
	{
		printf("All sessions == zero, so no diversity boosting is possible.\n");
		return OK;
	}
        best_pool_less_one= best_pool_size - 1;
	if( display_level==MED || display_level==HIGH )
		printf("best_pool_num=%d, best_pool_size=%d, best_pool_less_one=%d\n",best_pool_num,
			best_pool_size,best_pool_less_one);
	
	//begin -- for each pool
	pool = pools->first_pool;
	while ( pool != NULL )
	{
		//Find the most fit session
		best_session_size=0;
		session = pool->first_session;
		while ( session != NULL )
		{
			if( stalk_type == FUNCS)
			{
				if (session->hit_funcs > best_session_size)
				{
					best_session_size = session->hit_funcs;
					best_session_num = session->position;
				}
			}
			else
			{
				if (session->hit_blocks > best_session_size)
				{
					best_session_size = session->hit_blocks;
					best_session_num = session->position;
				}
			}
	
			session = session->next;
		}
		best_session_less_one= best_session_size - 1;
		if( best_session_less_one == -1 )
		{
			printf("Best fitness for this pool was zero, so moving to next pool.\n");
			pool = pool->next;
			continue;
		}
			
		if( display_level==MED || display_level==HIGH )
			printf("best_session_num=%d, best_session_size=%d, best_session_less_one=%d\n",best_session_num,
				best_session_size,best_session_less_one);

		//begin -- for each session
		session = pool->first_session;
		while ( session != NULL )
		{
			//Find unique hits via sessions (best vs. current)
			sprintf(query, "SELECT (SELECT COUNT(DISTINCT(eip)) from hits where gpf_id=%d \
					and generation=%d and pool=%d and (session=%d || session=%d)) - \
					(SELECT COUNT(eip) from hits where gpf_id=%d and generation=%d \
					and pool=%d and session=%d)",
					gpf_id,
					generation,
					pool->position,
					best_session_num,
					session->position,
					gpf_id,
					generation,
					pool->position,
					best_session_num);
			query_len = strlen(query);
			if( _mysql_query(query, query_len, mysql, mysql_host, mysql_user, mysql_passwd, dbname) == NULL)
			{
				printf("In calculate_diversity, _mysql_query distinct (1) returned an error.\n");
				return ERROR;
			}
			if( (responce = mysql_store_result(mysql)) == NULL)
			{
				printf("In calculate_diversity, mysql_store_result (3) returned an error.\n");
				if (mysql!=NULL){printf("Error %u (%s)\n",mysql_errno(mysql),mysql_error(mysql));}
				return ERROR;
			}
			row = mysql_fetch_row(responce);
			unique = atoi(row[0]);

			//calculate the new total fitness for sessions
			if( stalk_type == FUNCS)
			{
				session->func_fitness = (double)session->hit_funcs + ( ((double)unique/best_session_size)
					* (double)best_session_less_one );
				if( display_level==MED || display_level==HIGH )
					printf("Session %d: %d + ( (%d/%d) * %d ) = %4.3f\n", session->position,
						session->hit_funcs, unique, best_session_size, best_session_less_one,
						session->func_fitness); 
			}
			else
			{	
				session->block_fitness = (double)session->hit_blocks + ( ((double)unique/best_session_size)
					* (double)best_session_less_one );
				if( display_level==MED || display_level==HIGH )
					printf("Session %d: %d + ( (%d/%d) * %d ) = %4.3f\n", session->position,
						session->hit_blocks, unique, best_session_size, best_session_less_one,
						session->block_fitness); 
			}
			session = session->next;
			mysql_free_result(responce);
		}

		//Find unique hits via pools (best vs. current)
		sprintf(query, "SELECT (SELECT COUNT(DISTINCT(eip)) from hits where gpf_id=%d \
				and generation=%d and (pool=%d || pool=%d)) - \
				(SELECT COUNT(DISTINCT(eip)) from hits where gpf_id=%d and generation=%d \
				and pool=%d)",
				gpf_id,
				generation,
				best_pool_num,
				pool->position,
				gpf_id,
				generation,
				best_pool_num);
		query_len = strlen(query);
		if( _mysql_query(query, query_len, mysql, mysql_host, mysql_user, mysql_passwd, dbname) == NULL)
		{
			printf("In calculate_diversity, _mysql_query (distinct 2) returned an error.\n");
			return ERROR;
		}
		if( (responce = mysql_store_result(mysql)) == NULL)
		{
			printf("In calculate_diversity, mysql_store_result (2) returned an error.\n");
			if (mysql!=NULL){printf("Error %u (%s)\n",mysql_errno(mysql),mysql_error(mysql));}
			return ERROR;
		}
		row = mysql_fetch_row(responce);
		unique = atoi(row[0]);
		//printf("\tPool %d: %d + ( (%d/%d) * %d )\n", pool->position, pool->hit_funcs, unique,
		//	best_pool_size, best_pool_less_one); 
		//calculate the new total fitness for pools
		if( stalk_type == FUNCS)
		{
			pool->func_fitness = (double)pool->hit_funcs + ( ((double)unique/best_pool_size) *
				(double)best_pool_less_one );
			if(display_level==MED || display_level==HIGH )
				printf("Pool %d: %d + ( (%d/%d) * %d ) = %4.3f\n", pool->position, pool->hit_funcs,
					unique, best_pool_size, best_pool_less_one, pool->func_fitness); 
		}
		else
		{	
			pool->block_fitness = (double)pool->hit_blocks + ( ((double)unique/best_pool_size) *
				(double)best_pool_less_one );
			if( display_level==MED || display_level==HIGH )
				printf("Pool %d: %d + ( (%d/%d) * %d ) = %4.3f\n", pool->position, pool->hit_blocks,
					unique, best_pool_size, best_pool_less_one, pool->block_fitness); 
		}
		pool = pool->next;
		mysql_free_result(responce);
	}
	return OK;
}
