/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*Includes*/
#include "headers.h"

int combine_toks_to_leg(struct leg * leg, struct seed * seeds, char * types, int proxy)
{
	struct Tok * tok=NULL;
	struct Tok * prev_tok=NULL;
	int flag = 0;
	int flag2 = 0;
	int new_total=0;
	int prev_total=0;
	int x=0;

	/*This will made the data in a LEN token be correct before delievery.  Need one for HASH too.*/
	if ( Fix_Lens(leg->first_tok) == ERROR)
	{
                TIME(); fprintf(stderr, "{{Fix_Lens in combine_toks_to_leg failed!)}}\n"); return ERROR;
        }

	/*if we've sent this leg before it'll already have mem in it.  Could we use this for compare of some type?
	Anyway, we need to free it before using again or we'll have a memory leak
	*/
	if (leg->data != NULL && leg->data_len == -1)
	{
		printf("Invalid data_len/data combo in combine_toks_to_leg!\n");
		return ERROR;
	}

	if (leg->data != NULL) // || leg->data_len != -1
	{
		free(leg->data);
		leg->data = NULL;
	}

	leg->data_len = 0;

	tok = leg->first_tok;
        /*calc total len by adding up each tok->dataLen and storing in leg->data_len*/
        while(tok != NULL)
        {
		/*any token bigger than our 4M _malloc limit should die*/
		if( tok->dataLen > 4000000 || tok->dataLen < 1)
		{
			printf("Found ugly token, destroying...\n");
			if(tok->prev == NULL)
				flag2=1;
			if( (tok = Destroy_Toks(tok, "current")) == NULL)
			{
				printf("In combine_toks_to_leg, Destroy_Toks returned and error!\n");
				return ERROR;
			}
			leg->total_toks = leg->total_toks - 1;
		}
		else
		{
                	leg->data_len+=tok->dataLen;
			if( leg->data_len > 4000000 )
			{
				printf("Leg size to large! Chopping and just sending as is.\n");
				leg->data_len = 4000000;
				break;
			}
		}
                
		if ((int)tok > 2)
		{
			prev_tok = tok;
			//if the first tok was deleted we've already advanced to the next tok
			if(flag2)
				flag2=0;
			else
				tok=tok->next;
		}
		else
		{
			printf("Leg had only 1 tok, which was too big and got eliminated. Thus, no toks for this leg now.\n");
			printf("Creating a new tok.\n");
			tok = NULL;
                        tok = Create_Next_Tok(tok, leg->position); //style difference is due to backwards compliance
			leg->first_tok = tok;
			if( proxy == YES )
			{
				printf("If we ever get to this code we're in trouble, becuase we don't know type.\r");
				exit(-1);
				if( type_initialize_tok(tok, 0, seeds, types) == ERROR)
				{
					printf("In combine_toks_to_leg, type_init_tok return error!\n");
					return ERROR;
				}
			}
			else
			{
				if( randomly_initialize_tok(tok, 0, seeds) == ERROR)
				{
					printf("In combine_toks_to_leg, randomly_init_tok return error!\n");
					return ERROR;
				}
			}
			leg->total_toks = leg->total_toks + 1;
			leg->data_len = tok->dataLen;
			leg->first_tok = tok;
			flag = 1;
			break;
		}
        }

	if( flag == 0)
	{
		/*search back to possible new head of tok chain and set in leg*/
		while( prev_tok->prev != NULL)
			prev_tok = prev_tok->prev;
		leg->first_tok = prev_tok;
	}

	/*Malloc the memory*/
        if( (leg->data=_malloc(leg->data_len)) == NULL)
	{
		printf("In combine_toks_to_leg, malloc return error!\n");
		return ERROR;
	}

        /*cp data from each tok->data to leg->data*/
	/*If bugs are noticed consider riding the above code and adding a simple realloc in here*/
	tok = leg->first_tok;
        while(tok != NULL)
        {
		prev_total = new_total;
                new_total+=tok->dataLen;
		if(new_total >=4000000)
		{
			x = 4000000 - prev_total;
                	memcpy( (leg->data)+prev_total, tok->data, x);
			printf("Full leg not sent due to size > than 4MB.\n");
			break;
		}
                memcpy( (leg->data)+prev_total, tok->data, tok->dataLen);

                tok=tok->next;
        }

	return OK;
}
