/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

/*For the following:
	If a null ptr is passed in, that indicates the presense of a first, and a new will be created.
	Otherwise a new one is created and linked on.
*/

/* for all of the places you see a -1, that indicates a value has yet to be set.*/

struct pool * create_next_pool(struct pool * pool, int position)
{
	struct pool * new_pool;
        new_pool=_malloc(sizeof(struct pool));
        if(new_pool == NULL) {
                TIME(); fprintf(stderr, "{{{In create_next_pool, bad new new_pool}}}\n"); return(NULL);
        }
        if(position < 0) {
                TIME(); fprintf(stderr, "{{{In create_next_pool, bad position. position=%d}}}\n", position); return(NULL);
        }

	new_pool->position = position;
        new_pool->total_sessions = -1;
	new_pool->hit_funcs = -1;
        new_pool->total_funcs = -1;
        new_pool->func_fitness = -1;
        new_pool->hit_blocks = -1;
        new_pool->total_blocks = -1;
        new_pool->block_fitness = -1;  
	new_pool->next=NULL;

	if(pool == NULL)
                new_pool->prev=NULL;
        else
        {
                //set linked list ptrs
                new_pool->prev=pool;
                pool->next=new_pool;
        }

        return new_pool;
}

struct session * create_next_session(struct session * session, int position)
{
	struct session * new_session;
        new_session=_malloc(sizeof(struct session));
        if(new_session == NULL) {
                TIME(); fprintf(stderr, "{{{In create_next_session, bad new_session}}}\n"); return(NULL);
        }
        if(position < 0) {
                TIME(); fprintf(stderr, "{{{In create_next_session, bad position. position=%d}}}\n", position); return(NULL);
        }

	new_session->position = position;
        new_session->total_legs = -1; 
	new_session->hit_funcs = -1;
        new_session->total_funcs = -1;
        new_session->func_fitness = -1;
        new_session->hit_blocks = -1;
        new_session->total_blocks = -1;
        new_session->block_fitness = -1;  
	new_session->next=NULL;
	new_session->error = NO;	//set to NO for start.  if error in play will be set to YES.

	if(session == NULL)
                new_session->prev=NULL;
        else
        {
                //set linked list ptrs
                new_session->prev=session;
                session->next=new_session;
        }

        return new_session;
}

//also auto initializes the type READ || WRITE
struct leg * create_next_leg(struct leg * leg, int position, int PROXY)
{

	struct leg * new_leg;
        new_leg=_malloc(sizeof(struct leg));
        if(new_leg == NULL) {
                TIME(); fprintf(stderr, "{{{In create_next_leg, bad new new_leg}}}\n"); return(NULL);
        }
        if(position < 0) {
                TIME(); fprintf(stderr, "{{{In create_next_pool, bad position. position=%d}}}\n", position); return(NULL);
        }

	new_leg->position = position;
        new_leg->total_toks = -1;
	if( PROXY == NO )
        	new_leg->type = _rand(2);	//pick 0 || 1 to determine this legs type.  0=read. 1=write
	else
		new_leg->type = 1;
	new_leg->data_len = -1;		//just an initialization, will be set later
	new_leg->data = NULL; 		//just an initialization, will be set later
	new_leg->next=NULL;
	new_leg->error = NONE;		//set to NONE for start.  if error in play will to the error in that leg.

	if(leg == NULL)
                new_leg->prev=NULL;
        else
        {
                //set linked list ptrs
                new_leg->prev=leg;
                leg->next=new_leg;
        }

        return new_leg;
}

//
struct seed * create_next_seed(struct seed * seed, int position)
{

	struct seed * new_seed;
        new_seed=_malloc(sizeof(struct seed));
        if(new_seed == NULL) {
                TIME(); fprintf(stderr, "{{{In create_next_seed, bad new new_seed}}}\n"); return(NULL);
        }
        if(position < 0) {
                TIME(); fprintf(stderr, "{{{In create_next_pool, bad position. position=%d}}}\n", position); return(NULL);
        }

	new_seed->position = position;
        new_seed->type = -1;			// just an init, will be set later
	new_seed->data_len = -1;		// just an initialization, will be set later
	new_seed->data = NULL; 			//just an initialization, will be set later
	new_seed->next=NULL;

	if(seed == NULL)
                new_seed->prev=NULL;
        else
        {
                //set linked list ptrs
                new_seed->prev=seed;
                seed->next=new_seed;
        }

        return new_seed;
}

struct session * insert_session(struct pool * pool, int insert_position)
{
	struct session * tmp_session = NULL;
	struct session * new_session = NULL;
        new_session=_malloc(sizeof(struct session));
        if(new_session == NULL) {
                TIME(); fprintf(stderr, "{{{In insert_session, bad new_session}}}\n"); return(NULL);
        }
        if(insert_position < 0) {
                TIME(); fprintf(stderr, "{{{In insert_session, bad position. position=%d}}}\n", insert_position); return(NULL);
        }
	
	//common
        new_session->total_legs = -1; 
	new_session->hit_funcs = -1;
        new_session->total_funcs = -1;
        new_session->func_fitness = -1;
        new_session->hit_blocks = -1;
        new_session->total_blocks = -1;
        new_session->block_fitness = -1;  
	new_session->error = NO;	//set to NO for start.  if error in play will be set to YES.

	tmp_session = pool->first_session;
	while( tmp_session != NULL)
	{
		if( insert_position == tmp_session->position)
			break;
		tmp_session = tmp_session->next;
	}

	//first session
	if( tmp_session->position == 0)
	{
		pool->first_session = new_session;
		new_session->prev = NULL;
		new_session->next = tmp_session;
		tmp_session->prev = new_session;
	}
	//last or middle
	else
	{
		tmp_session->prev->next = new_session;

		new_session->prev = tmp_session->prev;
		new_session->next = tmp_session;

		tmp_session->prev = new_session;
	}

	//fix position nums
	new_session->position = tmp_session->position;
	while( tmp_session != NULL)
	{
		tmp_session->position = tmp_session->position + 1;
		tmp_session = tmp_session->next;
	}

        return new_session;
}

int delete_session(struct pool * pool, int delete_position)
{
	struct session * tmp_session = NULL;
	struct session * next_session = NULL;
	struct session * prev_session = NULL;
        if(pool==NULL || delete_position < 0)
	{
		printf("{{{In delete_session, bad pool or delete position. position=%d}}}\n", delete_position);
		return ERROR;
        }

	if(pool->total_sessions <= 1)
	{
		printf("Not enough session to safely delete one.\n");
		return ERROR;
	}

	//find the guy to kill	
	tmp_session = pool->first_session;
	while( tmp_session != NULL)
	{
		if( delete_position == tmp_session->position)
			break;
		tmp_session = tmp_session->next;
	}

	next_session = tmp_session->next;
	prev_session = tmp_session->prev;
	//first session
	if( tmp_session->position == 0)
	{
		pool->first_session = tmp_session->next;
		next_session->prev = NULL;
	}
	//last
	else if( tmp_session->next == NULL)
	{
		prev_session->next = NULL;
	}
	//middle
	else
	{
		prev_session->next = next_session;
		next_session->prev = prev_session;
	}

	//fix position nums
	while( next_session != NULL)
	{
		next_session->position--;
		next_session = next_session->next;
	}

	//free the dead guy
	if( destroy_session(tmp_session) == ERROR)
	{
		printf("In delete_session, destory_session returned an error!\n");
		return ERROR;
	}
	
        return OK;
}

int init_session(struct session * session, int legs_fixed_or_max, int max_legs_per_session, int toks_fixed_or_max,
	int max_toks_per_leg, struct seed * seeds)
{
	int k, l;
	struct leg * leg = NULL;
        struct Tok * tok = NULL;
        int first_leg = 1;
        int first_tok = 1;

        if(session == NULL) {
                TIME(); fprintf(stderr, "{{{In init_session, bad new_session}}}\n"); return ERROR;
	}
        if(legs_fixed_or_max < 0 || toks_fixed_or_max || max_legs_per_session < 0 || max_toks_per_leg < 0) {
                TIME(); fprintf(stderr, "{{{In init_session, bad values }}}\n"); return ERROR;
	}

	if(legs_fixed_or_max == FIXED)
		session->total_legs = max_legs_per_session;
	else
		session->total_legs = _rand( max_legs_per_session ) + 1;

        for(k=0; k < (session->total_legs); k++)
        {
                leg = create_next_leg(leg, k, NO);
                if(first_leg)
                {
                        session->first_leg = leg;
                        first_leg = 0;
                }

		if( toks_fixed_or_max == FIXED )
			leg->total_toks = max_toks_per_leg;
		else
                	leg->total_toks = _rand( max_toks_per_leg ) + 1;
                for(l=0; l < (leg->total_toks); l++)
                {
                        tok = Create_Next_Tok(tok, leg->position); //style difference is due to backwards compliance
                        if (first_tok)
                        {
                                leg->first_tok = tok;
                                first_tok = 0;
                        }
                        if( randomly_initialize_tok(tok, (leg->total_toks - l), seeds ) == ERROR)
                        {
                                printf("In init_session, randomly_init_tok return error!\n");
                                return ERROR;
                        }
                }
                first_tok = 1;
                tok = NULL;
        }

	return OK;
}

int init_type_session(struct session * session, char * types, struct seed * seeds)
{
	int l;
	struct leg * leg = NULL;
        struct Tok * tok = NULL;
        int first_tok = 1;
        int toks_per_leg=0;
        int x=0;

        if(session == NULL) {
                TIME(); fprintf(stderr, "{{{In init_type_session, bad new_session}}}\n"); return ERROR;
	}

        //get number of toks ... bad form since this is done in evolve_mode.c, but no harm done.
        while(1)
        {
                if( *(types+x)==0x00 )
                        break;
                toks_per_leg++;
                x = x + 5;
        }

	leg = create_next_leg(leg, 0, YES);
        session->first_leg = leg;

        leg->total_toks = toks_per_leg;
        x=0;
        for(l=0; l < (leg->total_toks); l++)
        {
		tok = Create_Next_Tok(tok, leg->position); //style difference is due to backwards compliance
		if (first_tok)
                {
                        leg->first_tok = tok;
                        first_tok = 0;
                }
                //the seed file gets used a speical way for proxy mode.
                //if a seed matches a token, we'll use the seed data some portion of the time.
                if( type_initialize_tok(tok, (leg->total_toks - l), seeds, types+x ) == ERROR)
                {
                        printf("In init_type_session, types_init_tok returned error!\n");
                        return ERROR;
                }
                x = x + 5;

        }
        first_tok = 1;
        tok = NULL;

	return OK;
}
