/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

struct pools * create_pools(u_int total_pools, u_int sessions_per_pool, u_int legs_per_session,
				u_int toks_per_leg, struct seed * seeds, int sessions_fixed_or_max, int legs_fixed_or_max,
					int toks_fixed_or_max)
{
	int i, j, k, l;
	struct pools * pools = NULL;
	struct pool * pool = NULL;
	struct session * session = NULL;
	struct leg * leg = NULL;
	struct Tok * tok = NULL;
	int first_pool = 1;
	int first_session = 1;
	int first_leg = 1;
	int first_tok = 1;

	pools = _malloc( sizeof(struct pools) );
	pools->total_pools = total_pools;

	for(i=0; i < pools->total_pools; i++)
	{
		pool = create_next_pool(pool, i);
		if(first_pool)
		{
			pools->first_pool = pool;
			first_pool = 0;
		}

		if( sessions_fixed_or_max == FIXED )
			pool->total_sessions = sessions_per_pool;
		else
			pool->total_sessions = _rand(sessions_per_pool) + 1;
		for(j=0; j < (pool->total_sessions); j++)
		{
			session = create_next_session(session, j);
			if(first_session)
			{
				pool->first_session = session;
				first_session = 0;
			}

			if( legs_fixed_or_max == FIXED )	
				session->total_legs = legs_per_session;
			else
				session->total_legs = _rand( legs_per_session ) + 1;
			for(k=0; k < (session->total_legs); k++)
			{
				leg = create_next_leg(leg, k, NO);
				if(first_leg)
				{
					session->first_leg = leg;
					first_leg = 0;
				}

				if( toks_fixed_or_max == FIXED )	
					leg->total_toks = toks_per_leg;
				else
					leg->total_toks = _rand( toks_per_leg ) + 1;
				for(l=0; l < (leg->total_toks); l++)
				{
					tok = Create_Next_Tok(tok, leg->position); //style difference is due to backwards compliance
					if (first_tok)
					{
						leg->first_tok = tok;
						first_tok = 0;
					}
					if( randomly_initialize_tok(tok, (leg->total_toks - l), seeds ) == ERROR)
					{
						printf("In create_pools, randomly_init_tok returned error!\n");
						return NULL;
					}
	
				}
				first_tok = 1;
				tok = NULL;

			}
			first_leg = 1;
			leg = NULL;

		}
		first_session = 1;
		session = NULL;
	}

	return pools;
}
