/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

struct fitness_list * create_fitness_list(struct session * session, int stalk_type)
{
	struct session * tmp_session;
	struct fitness_list * head;
	struct fitness_list * a;
	struct fitness_list * b;
	int first_time = 1;

	if( session == NULL){
		TIME(); fprintf(stderr, "{{{bad params to create_fitness_list}}}\n");return NULL;}

       	if( (head=_malloc(sizeof(struct fitness_list))) == NULL)
	{
               	TIME(); fprintf(stderr, "{{{Bad new a in create_fitness_list: low mem}}\n");
		return NULL;
       	}

	a = head;
	tmp_session = session;
	while (tmp_session != NULL)
	{
		if(first_time == 0)
		{
			if( (a = _malloc(sizeof(struct fitness_list))) == NULL)
			{
				printf("In create_fitness_list: error in malloc!\n");
				return NULL;
			}
			b->next = a;
		}
		first_time = 0;

	        a->position = tmp_session->position;

		if(stalk_type == FUNCS)
        		a->fitness = tmp_session->func_fitness;
	
		else
        		a->fitness = tmp_session->block_fitness;

		b = a;
	
		tmp_session = tmp_session->next;

	}
	
	a->next = NULL;

        return head;
}

int sort_fitness_list(struct fitness_list * fitness_list, int list_size)
{
	int i;
	double best_fit = 0;
	struct fitness_list * best_addr = NULL;
	struct fitness_list * head = NULL;
	struct fitness_list * f_list = NULL;
	int found_better = 0;

	if( fitness_list == NULL || list_size < 1){
		TIME(); fprintf(stderr, "{{{bad params to sort_fitness_list}}}\n");return ERROR;}

	head = fitness_list;
	f_list = head;
	best_fit = f_list->fitness;
	for(i=0; i < list_size; i++)
	{
		while( f_list != NULL )
		{
			if( f_list->fitness > best_fit)
			{
				best_fit = f_list->fitness;
				best_addr = f_list;
				found_better = 1;
			}
			f_list = f_list->next;
		}
		if(found_better)
		{ 
			if (swap(head, best_addr) == ERROR)
			{
				printf("In sort_fitness_list, swap returned error!\n");
				return ERROR;
			}
		}
		found_better = 0;

		/*now inc the head since we've compared that guy to all remaining so far.*/
		head = new_head(fitness_list, i);
		f_list = head;
		best_fit = f_list->fitness;
	}

	return OK;
}

struct fitness_list * new_head(struct fitness_list * fitness_list, int index)
{
	int i;
	struct fitness_list * f_list;

	if( fitness_list == NULL){
		TIME(); fprintf(stderr, "{{{Null fitness_list passed to new_head}}}\n");return NULL;}

	f_list = fitness_list;

	for(i=0; i < index; i++)
	{
		f_list = f_list->next;	
	}	

	return f_list;
}

int swap(struct fitness_list * head, struct fitness_list * best)
{
	if( head == NULL || best == NULL){
		TIME(); fprintf(stderr, "{{{Bad args passed to swap. head=%x and best=%x}}}\n", head, best);return ERROR;}
	int tmp_pos = head->position;
	double tmp_fit = head->fitness;

	head->position = best->position;
	head->fitness = best->fitness;

	best->position = tmp_pos;
	best->fitness = tmp_fit;

	return OK;
}

int print_fitness_list(struct fitness_list * fitness_list)
{
	struct fitness_list * f_list;

	if( fitness_list == NULL){
		TIME(); fprintf(stderr, "{{{Null fitness_list passed to print_fitness_list}}}\n");return ERROR;}
	
	f_list = fitness_list;
	while (f_list != NULL)
	{
		printf("%d=%4.3f; ", f_list->position, f_list->fitness);
		f_list = f_list->next;
	}
	printf("\n"); 
}

int destroy_fitness_list(struct fitness_list * fitness_list)
{
	if( fitness_list == NULL){
		TIME(); fprintf(stderr, "{{{Null fitness_list passed to destroy_fitness_list}}}\n");return ERROR;}

	struct fitness_list * f_list = fitness_list;

	while (fitness_list != NULL)
	{
		f_list = fitness_list->next;
		free(fitness_list);
		fitness_list = f_list;
	}

	return OK;
}

int sessions_crossover(struct pool * pool, struct fitness_list * fitness_list, int display_level,
	u_int max_legs_per_session, u_int max_toks_per_leg, int proxy)
{
	int i=0;
	struct session * a = NULL;
	struct session * b = NULL;
	struct session * kids = NULL;
	struct session * tmp_session = NULL;
	int size = 0;
	int half_size = 0;

	if( pool == NULL || fitness_list == NULL){
		TIME(); fprintf(stderr, "{{{Null pool or fitness_list passed to sessions_crossover}}}\n");return ERROR;}

	/*if there's only one session, don't crossover*/
	size = pool->total_sessions;
	if (size == 1)
	{
		if(display_level == HIGH)
			printf("Pool %d has only 1 session so we didn't perform a crossover.\n", pool->position);
		return OK;
	}
	if (size < 1)
	{
		printf("found pool with no session!!\n");
		return ERROR;
	}

	//calc sizes
	if( size % 2 == 0)
		half_size = size/2;
	else
		half_size = size/2 + 1;

	/*Always protect the first session (Elitism of one)*/
	//find the session with the highest fitness
	tmp_session = pool->first_session;
	if (tmp_session == NULL)
	{
		printf("found pool with a NULL session!!\n");
		return ERROR;
	}
	while(tmp_session != NULL)
	{
		if( tmp_session->position == fitness_list->position )
			break;
		tmp_session = tmp_session->next;
	}
	if (display_level == MED || display_level == HIGH)
		printf("\tBest session = %d. It becomes session 0 of the next gen.\n", tmp_session->position);
	//copy it in as the first session
	if( (kids = create_next_session(kids, 0)) == NULL)
	{
		printf("In session_crossover, create_next_session returned NULL!\n");
		return ERROR;
	}
	dup_session(kids, tmp_session, 0, proxy);

	/*breed to create list of kids*/
	tmp_session = pool->first_session;
	for(i=1; i < size; i+=2)
	{
		//70% of the time we pick from the top half
		if( _rand(100) > 29)
			a = pick_session(tmp_session, fitness_list, half_size);
		//30% of the time we pick from anywhere in the session list
		else
			a = pick_session(tmp_session, fitness_list, size);
		if( _rand(100) > 29)
			b = pick_session(tmp_session, fitness_list, half_size);
		else
			b = pick_session(tmp_session, fitness_list, size);

	    if( proxy == NO)
	    {	
		if( (kids = breed(a, b, kids, i, size-1, max_legs_per_session, max_toks_per_leg, proxy)) == NULL)
		{
			printf("In sessions_crossover (pool number %d), breed returned an error!\n",
				pool->position);
			printf("a=%x, b=%x, kids=%x, i=%d, half_size=%d, size=%d, size-1=%d\n", 
				a, b, kids, i, half_size, size, size-1);
			return ERROR;
		}
	    }
	    else
	    {
		if( (kids = types_breed(a, b, kids, i, size-1, proxy)) == NULL)
		{
			printf("In sessions_crossover (pool number %d), types_breed returned an error!\n",
				pool->position);
			printf("a=%x, b=%x, kids=%x, i=%d, half_size=%d, size=%d, size-1=%d\n", 
				a, b, kids, i, half_size, size, size-1);
			return ERROR;
		}
	    }
	}

	//save for destruction
	tmp_session = pool->first_session;

	while (kids->prev != NULL)
		kids = kids->prev;

	//we now have a totally new pool, except the best session was carried over as is.
	pool->first_session = kids;

	/*We have a new generation, so the parents can RIP*/
	if ( destroy_sessions(tmp_session) == ERROR)
	{
		printf("destroy_sessions returned an error!\n");
		return ERROR;
	}

	return OK;
}

int dup_session(struct session * dup_session, struct session * orig_session, int session_position, int proxy)
{
	if( dup_session == NULL || orig_session == NULL || session_position < 0){
		TIME(); fprintf(stderr, "{{{bad params to dup_session}}}\n");return ERROR;}

	int first_time = 1;

	struct leg * new_leg = NULL;	
	struct leg * orig_leg = NULL;	

	dup_session->position = session_position;
	dup_session->total_legs = orig_session->total_legs;
	dup_session->hit_funcs = orig_session->hit_funcs;
	dup_session->total_funcs = orig_session->total_funcs;
	dup_session->func_fitness = orig_session->func_fitness;
	dup_session->hit_blocks = orig_session->hit_blocks;
	dup_session->total_blocks = orig_session->total_blocks;
	dup_session->block_fitness = orig_session->block_fitness;
	dup_session->error = orig_session->error;

	orig_leg = orig_session->first_leg;

	while(orig_leg != NULL)
	{
		if( (new_leg = create_next_leg(new_leg, orig_leg->position, proxy)) == NULL)
		{
			printf("In dup_session, create_next_leg returned NULL!\n");
			return ERROR;
		}
		if( first_time == 1)
		{
			dup_session->first_leg = new_leg;
			first_time = 0;
		}

		if( dup_leg(new_leg, orig_leg, orig_leg->position) == ERROR)
		{
			printf("In dup_session, dup_leg returned an error!\n");
			return ERROR;
		}
		
		orig_leg = orig_leg->next;
	}

	return OK;
}

struct session * pick_session(struct session * session, struct fitness_list * fitness_list, int size)
{
	struct fitness_list * f_choosen;
	struct session * s_choosen;
	int x, i;

	if( session == NULL || fitness_list == NULL || size < 1){
		TIME(); fprintf(stderr, "{{{bad params to pick_session}}}\n");return NULL;}

	f_choosen = fitness_list;
	x = _rand(size);
	for(i=0; i < x; i++)
		f_choosen = f_choosen->next;

	s_choosen = session;
	while (s_choosen != NULL)
	{
		if ( s_choosen->position == f_choosen->position)
			break;
		s_choosen = s_choosen->next;
	}
	
	return s_choosen;
}

struct session * breed(struct session * a, struct session * b, struct session * kids, int position,
			int size, u_int max_legs_per_session, u_int max_toks_per_leg, int proxy)
{
	int i;
	int cp_a_leg;
	int cp_b_leg;
	struct leg * leg;
	struct leg * other_leg;
	struct leg * kid_leg;
	int cp_a_tok;
	int cp_b_tok;

	if( a == NULL){
		TIME(); fprintf(stderr, "{{{passed to breed: a == NULL}}}\n");return NULL;}
	if( b == NULL){
		TIME(); fprintf(stderr, "{{{passed to breed: b == NULL}}}\n");return NULL;}
	if( position < 0 || position > 9999999){
		TIME(); fprintf(stderr, "{{{passed to breed: position < 0 or greater than 9999999}}}\n");return NULL;}
	if( size < 1){
		TIME(); fprintf(stderr, "{{{passed to breed: size < 1}}}\n");return NULL;}

	/*pick a crossover position in some leg in session a and b*/
	cp_a_leg = _rand(a->total_legs);
	cp_b_leg = _rand(b->total_legs);

	/*Now pick a crossover position in some tok in that leg*/
	leg = a->first_leg;
	while ( leg != NULL)
	{
		if (leg->position == cp_a_leg)
		{
			cp_a_tok = _rand(leg->total_toks);
			break;
		}
		leg = leg->next;
	}
	other_leg = b->first_leg;
	while ( other_leg != NULL)
	{
		if (other_leg->position == cp_b_leg)
		{
			cp_b_tok = _rand(other_leg->total_toks);
			break;
		}
		other_leg = other_leg->next;
	}

	/*mama, meet papa. ;)*/
	if( (kids = create_next_session(kids, position)) == NULL)
	{
		printf("In breed, create_next_session returned NULL!\n");
		return NULL;
	}
	kids->position = position;
	kids->total_legs = 0;
	kid_leg = NULL;
	leg = a->first_leg;
	//get the 'a' leg to the leg with a crossover point
	//a new leg will be ready for tokens
	while(leg != NULL)
	{
		if( (kid_leg = create_next_leg(kid_leg, leg->position, proxy)) == NULL)
		{
			printf("in breed, create_next_leg returned NULL!\n");
			return NULL;
		}

		if( leg->position == cp_a_leg)
			break;

		if( dup_leg(kid_leg, leg, leg->position) == ERROR)
		{
			printf("In breed, dup_leg (1) returned an error!\n");
			return NULL;
		}
		
		leg = leg->next;
		kids->total_legs = kids->total_legs + 1;
	}

	kids->total_legs = kids->total_legs + 1;
	//now we copy in tokens UP TO the cp
	if (dup_leg_to(kid_leg, leg, cp_a_tok) == ERROR)
	{
		printf("In breed, dup_leg_to returned an error!\n");
		return NULL;
	}
	
	//index to the crossover point in the other session
	other_leg = b->first_leg;
	while(other_leg != NULL)
	{
		if( other_leg->position == cp_b_leg)
			break;
		other_leg = other_leg->next;
	}

	//now we copy in tokens STARTING FROM the cp
	if (dup_leg_from(kid_leg, cp_a_tok, other_leg, leg->position, cp_b_tok, max_toks_per_leg) == ERROR)
	{
		printf("In breed, dup_leg_from returned an error!\n");
		return NULL;
	}

	//and now we finish coping the rest of the complete legs from session b
	i = leg->position + 1;
	other_leg = other_leg->next;
	while(other_leg!=NULL && i<max_legs_per_session)
	{
		if( (kid_leg = create_next_leg(kid_leg, i, proxy)) == NULL)
		{
			printf("in breed, create_next_leg (2) returned NULL!\n");
			return NULL;
		}

		if( dup_leg(kid_leg, other_leg, i) == ERROR)
		{
			printf("In breed, dup_leg (2) returned an error!\n");
			printf("kids_leg=%x, other_leg=%x, i=%d\n", kid_leg, other_leg, i);
			print_leg(kid_leg);
			print_leg(other_leg);
			return NULL;
		}
		
		other_leg = other_leg->next;
		i++;
		kids->total_legs = kids->total_legs + 1; //counts the rest of the legs for this entirely new session.
	}
	
	/*Walk backwards and set the first leg in the session struct*/
	while (kid_leg->prev != NULL)
		kid_leg = kid_leg->prev;
	kids->first_leg = kid_leg;
	
	/***/
	/*bail out, we have enough kids*/
	if(position == size)
		return kids;
	/***/

	/*kid number 2*/
	/*do the same as above but in opposite manner*/
	if( (kids = create_next_session(kids, position+1)) == NULL)
	{
		printf("In breed, create_next_session (2) returned NULL!\n");
		return NULL;
	}
	kids->position = position + 1;
	kids->total_legs = 0;
	kid_leg = NULL;
	leg = b->first_leg;
	//get the 'b' leg to the leg with a crossover point
	//a new leg will be ready for tokens
	while(leg != NULL)
	{
		if( (kid_leg = create_next_leg(kid_leg, leg->position, proxy)) == NULL)
		{
			printf("in breed, create_next_leg (3) returned NULL!\n");
			return NULL;
		}

		if( leg->position == cp_b_leg)
			break;

		if( dup_leg(kid_leg, leg, leg->position) == ERROR)
		{
			printf("In breed, dup_leg (4) returned an error!\n");
			return NULL;
		}
		
		leg = leg->next;
		kids->total_legs = kids->total_legs + 1; //counts some 
	}

	kids->total_legs = kids->total_legs + 1; //counts the cp leg
	//now we copy in tokens UP TO the cp
	if (dup_leg_to(kid_leg, leg, cp_b_tok) == ERROR)
	{
		printf("In breed, dup_leg_to (2) returned an error!\n");
		return NULL;
	}
	
	//index to the crossover point in the other session
	other_leg = a->first_leg;
	while(other_leg != NULL)
	{
		if( other_leg->position == cp_a_leg)
			break;
		other_leg = other_leg->next;
	}

	//now we copy in tokens STARTING FROM the cp
	if (dup_leg_from(kid_leg, cp_b_tok, other_leg, leg->position, cp_a_tok, max_toks_per_leg) == ERROR)
	{
		printf("In breed, dup_leg_from (2) returned an error!\n");
		return NULL;
	}

	//and now we finish coping the rest of the complete legs from session a
	i = leg->position + 1;
	other_leg = other_leg->next;
	while(other_leg != NULL && i<max_legs_per_session)
	{
		if( (kid_leg = create_next_leg(kid_leg, i, proxy)) == NULL)
		{
			printf("in breed, create_next_leg (4) returned NULL!\n");
			return NULL;
		}

		if( dup_leg(kid_leg, other_leg, i) == ERROR)
		{
			printf("In breed, dup_leg (4) returned an error!\n");
			printf("kids_leg=%x, other_leg=%x, i=%d\n", kid_leg, other_leg, i);
			print_leg(kid_leg);
			print_leg(other_leg);
			return NULL;
		}
		
		other_leg = other_leg->next;
		i++;
		kids->total_legs = kids->total_legs + 1; //counts the rest of the legs for this entirely new session.
	}
	
	/*Walk backwards and set the first leg in the session struct*/
	while (kid_leg->prev != NULL)
		kid_leg = kid_leg->prev;
	kids->first_leg = kid_leg;

	return kids;
}

int dup_leg(struct leg * dup_leg, struct leg * orig_leg, int leg_position)
{
	int first_time = 1;
	struct Tok * tok;
	struct Tok * new_tok = NULL;

	if( dup_leg == NULL || orig_leg == NULL || leg_position < 0){
		TIME(); fprintf(stderr, "{{{bad params to dup_leg}}}\n");return ERROR;}

	dup_leg->position = leg_position;
	dup_leg->total_toks = orig_leg->total_toks;
	dup_leg->error = orig_leg->error;	
	dup_leg->type = orig_leg->type;	
	dup_leg->data_len = orig_leg->data_len;

	if( (orig_leg->data != NULL) && (dup_leg->data_len != -1) )
	{
		if ( (dup_leg->data = _malloc(dup_leg->data_len)) == NULL)
		{
			printf("In dup leg, error in malloc!\n");
			return ERROR;
		}
		memcpy(dup_leg->data, orig_leg->data, dup_leg->data_len);
	}

	tok = orig_leg->first_tok;
	while ( tok != NULL)
	{
		if( (new_tok = Create_Next_Tok(new_tok, dup_leg->position)) == NULL)
		{
			printf("In dup_leg, Create_Next_Tok returned NULL!\n");
			return ERROR;
		}
		if (first_time == 1)
		{
			dup_leg->first_tok = new_tok;
			first_time = 0;
		}
		if( limited_copy_tok(new_tok, tok) == ERROR)
		{
			printf("in dup_leg, copy_tok returned an error!\n");
			return ERROR;
		}
			
		tok = tok->next;
	}

	return OK;
}

int dup_leg_to(struct leg * dup_leg, struct leg * orig_leg, int stop_position)
{
	int first_time = 1;
	struct Tok * tok;
	struct Tok * new_tok = NULL;

	if( dup_leg == NULL ){
		TIME(); fprintf(stderr, "{{{bad params to dup_leg_to: dup_leg=%x}}}\n", dup_leg); return ERROR;}
	if( orig_leg == NULL){
		TIME(); fprintf(stderr, "{{{bad params to dup_leg_to: orig_leg=%x}}}\n", orig_leg); return ERROR;}
	if( stop_position < 0){
		TIME(); fprintf(stderr, "{{{bad params to dup_leg_to: stop_position=%d}}}\n",stop_position); return ERROR;}

	dup_leg->position = orig_leg->position;	
	dup_leg->type = orig_leg->type;	
	dup_leg->total_toks = 0;
	
	tok = orig_leg->first_tok;
	while ( tok != NULL)
	{
		if( (new_tok = Create_Next_Tok(new_tok, dup_leg->position)) == NULL)
		{
			printf("In dup_leg_to, Create_Next_Tok returned NULL!\n");
			return ERROR;
		}
		if (first_time == 1)
		{
			dup_leg->first_tok = new_tok;
			first_time = 0;
		}
		if( limited_copy_tok(new_tok, tok) == ERROR)
		{
			printf("In dup_leg_to, copy_tok returned an error\n");
			return ERROR;
		}
		
		if(tok->tokNum == stop_position)
			break;

		tok = tok->next;
		dup_leg->total_toks = dup_leg->total_toks + 1; //this will only hold some
	}

	return OK;
}

int dup_leg_from(struct leg * dup_leg, int left_off, struct leg * orig_leg, int leg_position,
			int start_position, u_int max_toks_per_leg)
{
	struct Tok * tok;
	struct Tok * new_tok;

	if( dup_leg == NULL){
		TIME(); fprintf(stderr, "{{{bad params to dup_leg_from: dup_leg=%x}}}\n", dup_leg); return ERROR;}
	if( left_off < 0){
		TIME(); fprintf(stderr, "{{{bad params to dup_leg_from: left_off=%d}}}\n", left_off); return ERROR;}
	if( orig_leg == NULL){
		TIME(); fprintf(stderr, "{{{bad params to dup_leg_from: orig_leg=%x}}}\n", orig_leg); return ERROR;}
	if( leg_position < 0){
		TIME(); fprintf(stderr, "{{{bad params to dup_leg_from: leg_position=%d}}}\n", leg_position); return ERROR;}
	if( start_position < 0){
		TIME(); fprintf(stderr, "{{{bad params to dup_leg_from: start_position=%d}}}\n", start_position); return ERROR;}

	/*index the kid_leg to where we left off coping*/
	new_tok = dup_leg->first_tok;
	while ( new_tok != NULL)
	{
		if(new_tok->tokNum == left_off)
                        break;
		new_tok = new_tok->next;
	}

	tok = orig_leg->first_tok;
	while ( tok!=NULL && dup_leg->total_toks < max_toks_per_leg)
	{
		if(tok->tokNum >= start_position)
		{
			if( (new_tok = Create_Next_Tok(new_tok, leg_position)) == NULL)
			{
				printf("In dup_leg_from, Create_Next_Tok returned NULL!\n");
				return ERROR;
			}

			if( limited_copy_tok(new_tok, tok) == ERROR)
			{
				printf("in dup_leg_from, copy_tok returned an error!\n");
				return ERROR;
			}
		}

		tok = tok->next;
		dup_leg->total_toks = dup_leg->total_toks + 1; //now we count the rest of the toks in this leg
	}

	return OK;
}

int limited_copy_tok(struct Tok * new_tok, struct Tok * tok)
{
        if(new_tok == NULL || tok == NULL) {
                TIME(); fprintf(stderr, "{{{Null tok passed to copy_tok}}}\n"); return(ERROR);}
	
	new_tok->dataLen = tok->dataLen;
	if( new_tok->dataLen < 1)
		new_tok->data=NULL;
	else
	{
		if( (new_tok->data = _malloc(new_tok->dataLen)) == NULL)
		{
			printf("In copy_tok: error in _malloc!\n");
			return ERROR;
		}
		memcpy(new_tok->data, tok->data, new_tok->dataLen);
	}

	new_tok->type = tok->type;
	new_tok->numeric = tok->numeric;
	new_tok->bracket = tok->bracket;
	new_tok->quote = tok->quote;
	new_tok->special = tok->special;
	new_tok->other = tok->other;
	new_tok->covered = tok->covered;
	new_tok->currentTotal = tok->currentTotal;

	return OK;
}

/*frees every peice of data in a pool execpt the pool data*/
int destroy_sessions(struct session * session)
{
	struct session * tmp_session;
	struct leg * leg;
	struct leg * tmp_leg;
	struct Tok * tok;

        if(session == NULL) {
                TIME(); fprintf(stderr, "{{{bad args in destory_sessions}}}\n"); return(ERROR);}

	while (session != NULL)
	{
		tmp_session = session->next;

		leg = session->first_leg;
		while (leg != NULL)
		{
			tmp_leg = leg->next;

			tok = leg->first_tok;
			if ( Destroy_Toks(tok, "all") == NULL)
			{
				printf("in destroy_session, Destory_Toks returned an error!\n");
				return ERROR;
			}

			//free leg->data and leg
			if (leg->data != NULL && leg->data_len != 0)
				free(leg->data);
			free(leg);

			leg = tmp_leg;
		}
		
		//free session
		free(session);

		session = tmp_session;
	}	

	return OK;
}

/*frees just this session and it's legs/toks*/
int destroy_session(struct session * session)
{
	struct leg * leg;
	struct leg * tmp_leg;
	struct Tok * tok;

        if(session == NULL) {
                TIME(); fprintf(stderr, "{{{bad args in destory_sessions}}}\n"); return(ERROR);}

	leg = session->first_leg;
	while (leg != NULL)
	{
		tmp_leg = leg->next;

		tok = leg->first_tok;
		if ( Destroy_Toks(tok, "all") == NULL)
		{
			printf("in destroy_session, Destory_Toks returned an error!\n");
			return ERROR;
		}

		//free leg->data and leg
		if (leg->data != NULL)
			free(leg->data);
		free(leg);

		leg = tmp_leg;
	}
	
	//free session
	free(session);

	return OK;
}

int tok_type_mutation(struct leg * leg)
{
	int x;
	struct Tok * tok;

        if(leg == NULL) {
                TIME(); fprintf(stderr, "{{{bad args in tok_type_mutation}}}\n"); return(ERROR);}

	x = _rand(leg->total_toks);

	tok = leg->first_tok;	
	while( tok != NULL)
	{
		if(tok->tokNum == x)
		{
			/*randomly pick a type*/
			tok->type = _rand(14);
			break;
		}
		
		tok = tok->next;	
	}

	return OK;
}

int tok_data_mutation(struct leg * leg, struct seed * seeds)
{
	int x;
	struct Tok * tok;

        if(leg == NULL) {
                TIME(); fprintf(stderr, "{{{bad args in tok_data_mutation}}}\n"); return(ERROR);}

	x = _rand(leg->total_toks);

	tok = leg->first_tok;	
	while( tok != NULL)
	{
		if(tok->tokNum == x)
		{
			/*be sure these are up to data*/
			if(Is_Numeric(tok))
				tok->numeric=1;
			if(Is_Bracket(tok))
				tok->bracket=1;
			if(Is_Quote(tok))
				tok->quote=1;
			if(Is_Special(tok))
				tok->special=1;
			if(Is_Other(tok))
				tok->other=1;

			/*Since fuzzing classes already exists, why not use those?*/
			if( fuzz_to_mutate(tok, seeds) == ERROR)
			{
				printf("fuzz_to_mutate returned an error!\n");
				return ERROR;
			}
		
			break;
		}
		
		tok = tok->next;	
	}

	return OK;
}

struct pools * pool_crossover(struct pools * pools, struct fitness_list * fitness_list,
			int display_level, u_int max_sessions_per_pool, int sessions_fixed_or_max, 
			u_int max_legs_per_session, int legs_fixed_or_max, u_int max_toks_per_leg, 
			int toks_fixed_or_max, struct seed * seeds, char * types, int proxy)
{
	struct pools * Pools_new=NULL;
	struct pool * new_pool=NULL;
	struct pool * old_pool=NULL;
	struct pool * a=NULL;
	struct pool * b=NULL;
	int i=0;
	int size=0;
	int half_size=0;

	if( pools == NULL || fitness_list == NULL){
		TIME(); fprintf(stderr, "{{{Null pool or fitness_list passed to pool_crossover}}}\n");return NULL;}

	/*error checking and size establishment*/
	size = pools->total_pools;
	if (size - 1 == 1)
	{
		if(display_level == HIGH)
			printf("Only 2 pool so we didn't perform a crossover.\n");
		return pools;
	}

	/*calculate the top half*/
	if( size % 2 == 0)
		half_size = size/2;
	else
		half_size = size/2 + 1;

	/*create new pools*/
	Pools_new = _malloc( sizeof(struct pools) );
        Pools_new->total_pools = size;

	/*Find the most fit pool*/
	old_pool = pools->first_pool;
	if (old_pool == NULL)
	{
		printf("In pool_crossover, found NULL pool!!\n");
		return NULL;
	}
	while(old_pool != NULL)
	{
		if( old_pool->position == fitness_list->position )
			break;
		old_pool = old_pool->next;
	}
	/*preserve first pool - Elitism of 1*/
	if( (new_pool = create_next_pool(new_pool, 0)) == NULL)
	{
		printf("In pool_crossover, create_next_pool returned NULL!\n");
		return NULL;
	}
	dup_pool(new_pool, old_pool, 0, proxy);
	Pools_new->first_pool = new_pool;

	/*breed to create list of "kid" pools*/
	old_pool = pools->first_pool;
	for(i=1; i<size; i+=2)
	{
		//70% of the time we pick from the top half
		if( _rand(100) > 29)
			a = pick_pool(old_pool, fitness_list, half_size);
		//30% of the time we pick from anywhere in the session list
		else
			a = pick_pool(old_pool, fitness_list, size);
		if( _rand(100) > 29)
			b = pick_pool(old_pool, fitness_list, half_size);
		else
			b = pick_pool(old_pool, fitness_list, size);
	
		if( (new_pool = breed_pools(a, b, new_pool, i, size-1, max_sessions_per_pool, sessions_fixed_or_max,
				max_legs_per_session, legs_fixed_or_max, max_toks_per_leg, toks_fixed_or_max,
				seeds, types, proxy)) == NULL)
		{
			printf("a=%x, b=%x, new_pool=%x, i=%d, half_size=%d, size=%d, size-1=%d\n", 
				a, b, new_pool, i, half_size, size, size-1);
			return NULL;
		}
	}

	/*We have a new generation of pools, so the parents can RIP*/
	if ( destroy_pools(pools) == ERROR)
	{
		printf("destroy_pools returned an error!\n");
		return NULL;
	}

	return Pools_new;
}

int dup_pool(struct pool * dup_pool, struct pool * orig_pool, int pool_position, int proxy)
{
	if( dup_pool == NULL || orig_pool == NULL || pool_position < 0){
		TIME(); fprintf(stderr, "{{{bad params to dup_pool}}}\n");return ERROR;}

	int first_time = 1;

	struct session * new_session = NULL;	
	struct session * orig_session = NULL;	

	dup_pool->position = pool_position;
	dup_pool->total_sessions = orig_pool->total_sessions;
	dup_pool->hit_funcs = orig_pool->hit_funcs;
	dup_pool->total_funcs = orig_pool->total_funcs;
	dup_pool->func_fitness = orig_pool->func_fitness;
	dup_pool->hit_blocks = orig_pool->hit_blocks;
	dup_pool->total_blocks = orig_pool->total_blocks;
	dup_pool->block_fitness = orig_pool->block_fitness;

	orig_session = orig_pool->first_session;

	while(orig_session != NULL)
	{
		if( (new_session = create_next_session(new_session, orig_session->position)) == NULL)
		{
			printf("In dup_pool, create_next_session returned NULL!\n");
			return ERROR;
		}
		if( first_time == 1)
		{
			dup_pool->first_session = new_session;
			first_time = 0;
		}

		if( dup_session(new_session, orig_session, orig_session->position, proxy) == ERROR)
		{
			printf("In dup_pool, dup_session returned an error!\n");
			return ERROR;
		}
		
		orig_session = orig_session->next;
	}

	return OK;
}

struct pool * pick_pool(struct pool * pool, struct fitness_list * fitness_list, int size)
{
	struct fitness_list * f_choosen;
	struct pool * s_choosen;
	int x, i;

	if( pool == NULL || fitness_list == NULL || size < 1){
		TIME(); fprintf(stderr, "{{{bad params to pick_pool}}}\n");return NULL;}

	f_choosen = fitness_list;
	x = _rand(size);
	for(i=0; i < x; i++)
		f_choosen = f_choosen->next;

	s_choosen = pool;
	while (s_choosen != NULL)
	{
		if ( s_choosen->position == f_choosen->position)
			break;
		s_choosen = s_choosen->next;
	}
	
	return s_choosen;
}

struct pool * breed_pools(struct pool * a, struct pool * b, struct pool * new_pool, int position,
			int size, u_int max_sessions_per_pool, int sessions_fixed_or_max, u_int max_legs_per_session,
			int legs_fixed_or_max, u_int max_toks_per_leg, int toks_fixed_or_max, struct seed * seeds,
			char * types, int proxy)
{
	int i=0;
	int cp_a_session;
	int cp_b_session;
	struct session * session;
	struct session * other_session;
	struct session * kid_session;

	if( a == NULL){
		TIME(); fprintf(stderr, "{{{passed to breed_pool_pool: a == NULL}}}\n");return NULL;}
	if( b == NULL){
		TIME(); fprintf(stderr, "{{{passed to breed_pool_pool: b == NULL}}}\n");return NULL;}
	if( position < 0 || position > 9999999){
		TIME(); fprintf(stderr, "{{{passed to breed_pool_pool: position < 0 or greater than 9999999}}}\n");return NULL;}
	if( size < 1){
		TIME(); fprintf(stderr, "{{{passed to breed_pool_pool: size < 1}}}\n");return NULL;}

	/*pick a crossover position at some session*/
	cp_a_session = _rand(a->total_sessions);
	cp_b_session = _rand(b->total_sessions);

	if( (new_pool = create_next_pool(new_pool, position)) == NULL)
	{
		printf("In breed_pool_pool, create_next_pool returned NULL!\n");
		return NULL;
	}
	new_pool->total_sessions = 0;
	kid_session = NULL;
	session = a->first_session;
	//get the 'a' session to the session with a crossover point
	i=0;
	while(session != NULL)
	{
		//copy including crosspoint
		if( (kid_session = create_next_session(kid_session, i)) == NULL)
		{
			printf("in breed_pool, create_next_session returned NULL!\n");
			return NULL;
		}

		if( dup_session(kid_session, session, i, proxy) == ERROR)
		{
			printf("In breed_pool, dup_session (1) returned an error!\n");
			return NULL;
		}
		new_pool->total_sessions = new_pool->total_sessions + 1;
		i++;
		if( session->position == cp_a_session)
			break;
		
		session = session->next;
	}

	//index to the crossover point in the other session
	other_session = b->first_session;
	while(other_session != NULL)
	{
		if( other_session->position == cp_b_session)
			break;
		other_session = other_session->next;
	}

	//skip over the one at the crosspoint 
	other_session = other_session->next;
	//and now we finish coping the rest of the complete sessions from session b
	while(other_session!=NULL && i<max_sessions_per_pool)
	{
		if( (kid_session = create_next_session(kid_session, i)) == NULL)
		{
			printf("in breed_pool, create_next_session (2) returned NULL!\n");
			return NULL;
		}

		if( dup_session(kid_session, other_session, i, proxy) == ERROR)
		{
			printf("In breed_pool, dup_session (2) returned an error!\n");
			printf("kid_session=%x, other_session=%x, i=%d\n", kid_session, other_session, i);
			print_session(kid_session);
			print_session(other_session);
			return NULL;
		}
		new_pool->total_sessions = new_pool->total_sessions + 1;
		i++;

		other_session = other_session->next;
	}

	/*keep the same number of sessions/pool fixed.
	  Accomplish this by generating totally new session.
	  This will also keep some devirsity in the mix.
	*/
	if (sessions_fixed_or_max == FIXED)
	{
		while( new_pool->total_sessions < max_sessions_per_pool )
		{
			if( (kid_session = create_next_session(kid_session, i)) == NULL)
			{
				printf("In breed_pools, create_next_session (3) returned NULL!\n");
				return NULL;
			}

			if( proxy == NO)
			{
				if( init_session(kid_session, legs_fixed_or_max, max_legs_per_session, toks_fixed_or_max,
					max_toks_per_leg, seeds)==ERROR)
				{
					printf("In breed_pools, init_session return an error!\n");
					return NULL;
				}
			}
			else
			{
				if( init_type_session(kid_session, types, seeds)==ERROR)
				{
					printf("In breed_pools, init_type_session return an error!\n");
					return NULL;
				}
			}

			new_pool->total_sessions = new_pool->total_sessions + 1;
			i++;
		}
	}
	
	/*Walk backwards and set the first session in the session struct*/
	while (kid_session->prev != NULL)
		kid_session = kid_session->prev;
	new_pool->first_session = kid_session;
	
	/***/
	/*bail out, we have enough pools in new_pool*/
	if(position == size)
		return new_pool;
	/***/

	/*kid number 2*/
	/*do the same as above but in opposite manner*/
	if( (new_pool = create_next_pool(new_pool, position+1)) == NULL)
	{
		printf("In breed_pool, create_next_session (2) returned NULL!\n");
		return NULL;
	}
	new_pool->total_sessions = 0;
	kid_session = NULL;
	session = b->first_session;
	//get the 'b' session to the session with a crossover point
	i=0;
	while(session != NULL)
	{
		//copy including crosspoint
		if( (kid_session = create_next_session(kid_session, i)) == NULL)
		{
			printf("in breed_pool, create_next_session (3) returned NULL!\n");
			return NULL;
		}

		if( dup_session(kid_session, session, i, proxy) == ERROR)
		{
			printf("In breed_pool, dup_session (4) returned an error!\n");
			return NULL;
		}

		new_pool->total_sessions = new_pool->total_sessions + 1;
		i++;
		if( session->position == cp_b_session)
			break;
		
		session = session->next;
	}

	//index to the crossover point in the other session
	other_session = a->first_session;
	while(other_session != NULL)
	{
		if( other_session->position == cp_a_session)
			break;
		other_session = other_session->next;
	}

	//skip the crosspoint
	other_session = other_session->next;
	//and now we finish coping the rest of the complete sessions from session a
	while(other_session != NULL && i<max_sessions_per_pool)
	{
		if( (kid_session = create_next_session(kid_session, i)) == NULL)
		{
			printf("in breed_pool, create_next_session (4) returned NULL!\n");
			return NULL;
		}

		if( dup_session(kid_session, other_session, i, proxy) == ERROR)
		{
			printf("In breed_pool, dup_session (4) returned an error!\n");
			printf("new_pool_session=%x, other_session=%x, i=%d\n", kid_session, other_session, i);
			print_session(kid_session);
			print_session(other_session);
			return NULL;
		}
		new_pool->total_sessions = new_pool->total_sessions + 1;
		i++;

		other_session = other_session->next;
	}

	/*keep the same number of sessions/pool fixed.
	  Accomplish this by generating totally new session.
	  This will also keep some devirsity in the mix.
	*/
	if (sessions_fixed_or_max == FIXED)
	{
		while( new_pool->total_sessions < max_sessions_per_pool )
		{
			if( (kid_session = create_next_session(kid_session, i)) == NULL)
			{
				printf("in breed_pool, create_next_session (3) returned NULL!\n");
				return NULL;
			}

			if( proxy == NO)
			{
				if( init_session(kid_session, legs_fixed_or_max, max_legs_per_session, toks_fixed_or_max,
					max_toks_per_leg, seeds)==ERROR)
				{
					printf("In breed_pools, init_session return an error!\n");
					return NULL;
				}
			}
			else
			{
				if( init_type_session(kid_session, types, seeds)==ERROR)
				{
					printf("In breed_pools, init_type_session return an error!\n");
					return NULL;
				}
			}

			new_pool->total_sessions = new_pool->total_sessions + 1;
			i++;
		}
	}
	
	/*Walk backwards and set the first session in the session struct*/
	while (kid_session->prev != NULL)
		kid_session = kid_session->prev;
	new_pool->first_session = kid_session;

	return new_pool;
}

/*frees every peice of data in all pools*/
int destroy_pools(struct pools * pools)
{
	struct pool * pool;
	struct pool * tmp_pool;
	struct session * session;

        if(pools == NULL) {
                TIME(); fprintf(stderr, "{{{bad args in destory_pools}}}\n"); return(ERROR);}

	pool = pools->first_pool;
	while (pool != NULL)
	{
		//destory all of the toks, legs, and sessions of this pool
		session = pool->first_session;
		destroy_sessions(session);
	
		//free pool	
		tmp_pool = pool->next;
		free(pool);
		pool = tmp_pool;
	}	

	//free the main Pools ptr
	free(pools);

	return OK;
}

int add_random_session(struct pool * pool, u_int max_session_per_pool, int sessions_fixed_or_max, u_int max_legs_per_session,
	int legs_fixed_or_max, u_int max_toks_per_leg, int toks_fixed_or_max, struct seed * seeds)
{
	struct session * session = NULL;
	int insert_position = 0;

        if(pool == NULL) {
                TIME(); fprintf(stderr, "{{{bad args in add_random_session}}}\n"); return(ERROR);}
	
	//be sure we won't go over max
	if( pool->total_sessions >= max_session_per_pool )
	{
		return OK;
	}

	//if there's only one session we allow for the insertion as a first session
	if( pool->total_sessions == 1)
	{
		insert_position = 0;
	}
	else
	{
		insert_position = _rand(pool->total_sessions);
		//leave 1st session alone
		if( insert_position == 0)
			insert_position = 1;
	}

	if( (session = insert_session(pool, insert_position)) == NULL)
	{
		printf("In add_random_session, insert_session return an error!\n");
		return ERROR;
	}

	if( init_session(session, legs_fixed_or_max, max_legs_per_session, toks_fixed_or_max,
		max_toks_per_leg, seeds) == ERROR)
	{
		printf("In add_random_session, init_session return an error!\n");
		return ERROR;
	}
	
	pool->total_sessions = pool->total_sessions + 1;
	return OK;
}

int remove_random_session(struct pool * pool)
{
	int deletion_point;

        if(pool == NULL) {
                TIME(); fprintf(stderr, "{{{bad args in remove_random_session}}}\n"); return(ERROR);}

	if( pool->total_sessions <=1)
		return OK;

	if( pool->total_sessions == 2)
	{
		deletion_point = 1;
	}
	else
	{
		deletion_point = _rand(pool->total_sessions);
		//leave 1st session alone
		if( deletion_point == 0)
		{
			deletion_point = pool->total_sessions - 1;
		}
	}

	if( delete_session(pool, deletion_point) == ERROR)
	{
		printf("In remove_random_session, insert_session return an error!\n");
		return ERROR;
	}

	pool->total_sessions = pool->total_sessions - 1;
	return OK;
}

int add_type_session(struct pool * pool, u_int max_session_per_pool, int sessions_fixed_or_max,
	char * types, struct seed * seeds)
{
	struct session * session = NULL;
	int insert_position = 0;

        if(pool == NULL) {
                TIME(); fprintf(stderr, "{{{bad args in add_random_session}}}\n"); return(ERROR);}
	
	//be sure we won't go over max
	if( pool->total_sessions >= max_session_per_pool )
	{
		return OK;
	}

	//if there's only one session we allow for the insertion as a first session
	if( pool->total_sessions == 1)
	{
		insert_position = 0;
	}
	else
	{
		insert_position = _rand(pool->total_sessions);
		//leave 1st session alone
		if( insert_position == 0)
			insert_position = 1;
	}

	if( (session = insert_session(pool, insert_position)) == NULL)
	{
		printf("In add_random_session, insert_session return an error!\n");
		return ERROR;
	}

	if( init_type_session(session, types, seeds) == ERROR)
	{
		printf("In add_random_session, init_session return an error!\n");
		return ERROR;
	}
	
	pool->total_sessions = pool->total_sessions + 1;
	return OK;
}

struct session * types_breed(struct session * a, struct session * b, struct session * kids, int position, int size, int proxy)
{
	int x;
	char * ptr;
	struct leg * leg;
	struct Tok * token;
	struct Tok * token_other;
	int swap_token;
	int swap_point;

	if( a == NULL){
		TIME(); fprintf(stderr, "{{{passed to types_breed: a == NULL}}}\n");return NULL;}
	if( b == NULL){
		TIME(); fprintf(stderr, "{{{passed to types_breed: b == NULL}}}\n");return NULL;}
	if( position < 0 || position > 9999999){
		TIME(); fprintf(stderr, "{{{passed to types_breed: position < 0 or greater than 9999999}}}\n");return NULL;}
	if( size < 1){
		TIME(); fprintf(stderr, "{{{passed to types_breed: size < 1}}}\n");return NULL;}

	/*pick a token to perform crossover in*/
	leg = a->first_leg;
	swap_token = _rand(leg->total_toks);

	/*now pick a byte position to slice on in that token*/
	token = leg->first_tok;
	while ( token != NULL)
	{
		if (token->tokNum == swap_token)
		{
			swap_point = _rand(token->dataLen);
			break;
		}
		token = token->next;
	}

	/*both sessions remain untouched, other than the token in which the swap occurs*/
	/*in that token keep original data up to swap point, than copy from the other for the rest. via-versa.*/
	if( (kids = create_next_session(kids, position)) == NULL)
	{
		printf("In session_crossover, create_next_session returned NULL!\n");
		return NULL;
	}
	dup_session(kids, a, position, proxy);
	//copy data after swap point into new session a
	token = kids->first_leg->first_tok;
	token_other = b->first_leg->first_tok;
	while ( token != NULL)
	{
		if (token->tokNum == swap_token)
		{
			break;
		}
		token = token->next;
		token_other = token_other->next;
	}
	
	//was having problems with swap_point being greater than dataLen
	if (swap_point > token->dataLen)
	{
		printf("ERROR in types breed: swap_point > token->dataLen\n");
		swap_point = _rand(token->dataLen);
		printf("Re-choosing swap_point.\n");
		
	}

	if( token->dataLen == token_other->dataLen)
		memcpy(token->data+swap_point, token_other->data+swap_point, (token->dataLen) - swap_point);
	else
	{
		//33% of the time we extend both lens to largest len
		if( (x=_rand(100)) < 33)
		{
			if( token->dataLen > token_other->dataLen)
			{
				token_other->data = realloc(token_other->data, token->dataLen);
				token_other->dataLen = token->dataLen;
			}
			else
			{
				token->data = realloc(token->data, token_other->dataLen);
				token->dataLen = token_other->dataLen;
			}
			if( swap_point < token->dataLen)
				memcpy(token->data+swap_point, token_other->data+swap_point, (token->dataLen) - swap_point);
		}
		//33% of the time we trunc both lens to the shortest len
		else if (x < 66)
		{
			if( token->dataLen < token_other->dataLen)
			{
				token_other->data = realloc(token_other->data, token->dataLen);
				token_other->dataLen = token->dataLen;
			}
			else
			{
				token->data = realloc(token->data, token_other->dataLen);
				token->dataLen = token_other->dataLen;
			}
			if( swap_point < token->dataLen)
				memcpy(token->data+swap_point, token_other->data+swap_point, (token->dataLen) - swap_point);
		}
		//33% of the time we string swap
		else
		{
			x = token->dataLen;
			ptr = _malloc(x);
			memcpy(ptr, token->data, x);

			token->data = realloc(token->data, token_other->dataLen);
			token->dataLen = token_other->dataLen;
			memcpy(token->data, token_other->data, token_other->dataLen);

			token_other->data = realloc(token_other->data, x);
			token_other->dataLen = x;
			memcpy(token_other->data, ptr, x);
			free(ptr);
		}
	}

	/*bail out, we have enough kids*/
	if(position == size)
		return kids;

	//else carry on	
	if( (kids = create_next_session(kids, position+1)) == NULL)
	{
		printf("In types_breed, create_next_session returned NULL!\n");
		return NULL;
	}
	dup_session(kids, b, position+1, proxy);
	//copy data after swap point into new session a
	token = kids->first_leg->first_tok;
	token_other = a->first_leg->first_tok;
	while ( token != NULL)
	{
		if (token->tokNum == swap_token)
		{
			break;
		}
		token = token->next;
		token_other = token_other->next;
	}
	if( token->dataLen == token_other->dataLen)
		memcpy(token->data+swap_point, token_other->data+swap_point, (token->dataLen) - swap_point);
	else
	{
		//33% of the time we extend both lens to largest len
		if( (x=_rand(100)) < 33)
		{
			if( token->dataLen > token_other->dataLen)
			{
				token_other->data = realloc(token_other->data, token->dataLen);
				token_other->dataLen = token->dataLen;
			}
			else
			{
				token->data = realloc(token->data, token_other->dataLen);
				token->dataLen = token_other->dataLen;
			}
			if( swap_point < token->dataLen)
				memcpy(token->data+swap_point, token_other->data+swap_point, (token->dataLen) - swap_point);
		}
		//33% of the time we trunc both lens to the shortest len
		else if (x < 66)
		{
			if( token->dataLen < token_other->dataLen)
			{
				token_other->data = realloc(token_other->data, token->dataLen);
				token_other->dataLen = token->dataLen;
			}
			else
			{
				token->data = realloc(token->data, token_other->dataLen);
				token->dataLen = token_other->dataLen;
			}
			if( swap_point < token->dataLen)
				memcpy(token->data+swap_point, token_other->data+swap_point, (token->dataLen) - swap_point);
		}
		//33% of the time we string swap
		else
		{
			x = token->dataLen;
			ptr = _malloc(x);
			memcpy(ptr, token->data, x);

			token->data = realloc(token->data, token_other->dataLen);
			token->dataLen = token_other->dataLen;
			memcpy(token->data, token_other->data, token_other->dataLen);

			token_other->data = realloc(token_other->data, x);
			token_other->dataLen = x;
			memcpy(token_other->data, ptr, x);
			free(ptr);
		}
	}

	return kids;
}
