/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

int fuzz_to_mutate(struct Tok * tok, struct seed * seeds)
{
        int ret=0;
	int fuzz=0;
        u_int x=0;
	char * ptr=NULL;
	char * ptr2=NULL;
	int size = 0;
	int half_size = 0;
	struct Tok * tmp_tok=NULL;

	/*This one's getting to big: chop it back or randomly create a new one*/
	if( tok->dataLen > 100000 )
	{
		if ( _rand(100) > 40 )
		{
			x = _rand(40000) + 10;
			ptr = _malloc(x);
			memcpy(ptr, tok->data, x);
			free(tok->data);
			tok->data = ptr;
		}
		else
		{
			tmp_tok = tok;
			while(tmp_tok->next != NULL)
                		tmp_tok=tmp_tok->next;

			x = tmp_tok->tokNum - tok->tokNum;
			free(tok->data);
			randomly_initialize_tok(tok, x, seeds);
		}
		
		return OK;
	}
	/*This one somehow went to  zero. randomly create a new one*/
	if( tok->dataLen < 1)
	{
		tmp_tok = tok;
		while(tmp_tok->next != NULL)
               		tmp_tok=tmp_tok->next;

		x = tmp_tok->tokNum - tok->tokNum;
		if(tok->data != NULL)
			free(tok->data);
		randomly_initialize_tok(tok, x, seeds);
		
		return OK;
	}

	/*Again we add one more class that will help keep the token population from growing to quickly*/
	if( _rand(100) > 85 && tok->data !=NULL && tok->dataLen > 11)
	{
		size = tok->dataLen;
		half_size = size / 2;
		ptr = tok->data + _rand(half_size);
		ptr2 = tok->data + _rand(half_size) + half_size;

		/*use only the middle*/
		if( _rand(100) < 50)
		{
			size = ptr2 - ptr;
			ptr2 = _malloc(size);
			memcpy(ptr2, ptr, size);
			free(tok->data);
			tok->dataLen = size;
			tok->data = ptr2;	

		}
		/*chop out the middle*/
		else
		{
			half_size = ptr - tok->data;
			size = (tok->data + size) - ptr2;

			ptr = _malloc(half_size + size);

			memcpy(ptr, tok->data, half_size);
			memcpy(ptr+half_size, ptr2, size);

			free(tok->data);
			tok->dataLen = half_size + size;
			tok->data = ptr;

		}

		return OK;
	}

	if( tok->type==STRING || tok->type==MIXED || tok->type==ASCII || tok->type==ASCII_CMD || tok->type==ASCII_CMDVAR || tok->type==WIDESTRING_2 || tok->type==WIDESTRING_4 )
	{
		x=_rand(100);
                if( tok->numeric==1 && (x < 25) ) //25% to fuzz if num
		{
                        if( (ret+=Fuzz_Numeric(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Numeric returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
                else if (tok->numeric==0 && (x < 3) ) //3% to fuzz as num even if not
		{
                        if( (ret+=Fuzz_Numeric(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Numeric2 returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
                x=_rand(100);
                if( (tok->bracket==1 || tok->quote==1 || tok->special==1 || tok->other==1) && (x < 25) ) //25% if
		{
                        if( (ret+=Fuzz_Parsers(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Parsers returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
                else if( (x < 3) )
		{
                        if( (ret+=Fuzz_Parsers(tok)) < 0){ //3% if not
				TIME(); fprintf(stderr, "{{Fuzz_Parsers2 returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
		//for sure do one of these
                x=_rand(100);
                if( x < 25 )
		{
                        if( (ret+=Fuzz_BitFlip(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_BitFlip returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
                if( x < 80 )
		{
                        if( (ret+=Fuzz_Buffer(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Buffer returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
                if( x < 100 )
		{
                        if( (ret+=Fuzz_Format(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Format returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
	}

	else if ( tok->type == ASCII_SPACE )
	{
		if( (ret+=Fuzz_ASCII_SPACE(tok)) < 0){
			TIME(); fprintf(stderr, "{{Fuzz_ASCII_SPACE returned -1}}\n"); Print_Toks(tok, "1", "long");}
	}

	else if( tok->type==ASCII_END || tok->type==BINARY_END )
	{
		if( (ret+=Fuzz_END(tok)) < 0){
			TIME(); fprintf(stderr, "{{Fuzz_ASCII_END returned -1}}\n"); Print_Toks(tok, "1", "long");}
	}

	if( tok->type==SESSION_ID || tok->type==FLAGS || tok->type==PROTO_ID || tok->type==LEN
		|| tok->type==HASH || tok->type==BINARY )
	{
		x=_rand(100);
                if( tok->numeric==1 && (x < 25) ) //25% to fuzz if num
		{
                        if( (ret+=Fuzz_Numeric(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Numeric returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
                else if (tok->numeric==0 && (x < 3) ) //3% to fuzz as num even if not
		{
                        if( (ret+=Fuzz_Numeric(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Numeric2 returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
                x=_rand(100);
                if( (tok->bracket==1 || tok->quote==1 || tok->special==1 || tok->other==1) && (x < 25) ) //25% if
		{
                        if( (ret+=Fuzz_Parsers(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Parsers returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
                else if( (x < 3) )
		{
                        if( (ret+=Fuzz_Parsers(tok)) < 0){ //3% if not
				TIME(); fprintf(stderr, "{{Fuzz_Parsers2 returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
		//do one of these for sure
                x=_rand(100);
                if( x < 50 )
		{
                        if( (ret+=Fuzz_BitFlip(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_BitFlip returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
                else if( x < 60 )
		{
                        if( (ret+=Fuzz_Buffer(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Buffer returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
                else if( x < 70 )
		{
                        if( (ret+=Fuzz_Format(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Format returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
		//special case for bin fuzzing.  pass directly to Fuzz_Integer
                else if( (x < 100) && (tok->data != NULL) && (tok->dataLen > 0) )
		{
			u_int16_t size8=0;
			u_int16_t size16=0;
			u_int16_t size32=0;
			if( tok->dataLen == 1)
			{
				size8=Fuzz_Integer( _ndata_to_size8(tok->data) );
				_size8_to_ndata(tok->data, size8);
			}
			else if( tok->dataLen == 2)
			{
				size16=Fuzz_Integer( _ndata_to_size16(tok->data) );
				_size16_to_ndata(tok->data, size16);
			}
			else if( tok->dataLen == 4)
			{
				size32=Fuzz_Integer( _ndata_to_size32(tok->data) );
				_size32_to_ndata(tok->data, size32);
			}
			else
				*((int*)(tok->data))=Fuzz_Integer( *((int*)(tok->data)) );
		}
                else 
		{
                        if( (ret+=Fuzz_BitFlip(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_BitFlip returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}

	}

	if( tok->type == UNDEFINED ){
		TIME(); fprintf(stderr, "{{Undefined type in Fuzz}}\n"); Print_Toks(tok, "1", "long"); return ERROR;}

	//each function will need to report on it's own error
	if(ret < 0 ) {
		TIME(); fprintf(stderr, "{{ret=%d}}\n",ret); return ERROR;}
	else	
		return OK;
}

