/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

int paimei_send_target_tag(int generation, int pool, int session, int paimei_fd, int display_level)
{
	char target_tag[40];
	memset(target_tag, 0x00, 40);

	sprintf(target_tag, "GPF %d_%d:%d", generation, pool, session);
	if( write(paimei_fd, target_tag, strlen(target_tag)) == ERROR)
	{
		printf("in PaiMei_send_target_tag, write returned an error!\n");
		return ERROR;
	}
	if (display_level == HIGH)
		printf("\n-> %s\r\n", target_tag);
	return OK;
}

int is_paimei_ready(int paimei_fd, int display_level)
{
	int x;
	char paimei_ready[30];
	memset(paimei_ready, 0x00, 30);

	if( (x = read(paimei_fd, paimei_ready, 30)) == ERROR)
	{
		printf("In is_paimei_ready, read returned an error!\n");
		return ERROR;
	}
	if (x == 10 || (strcmp(paimei_ready, "Stalking\r\n") == 0) )
	{
		if (display_level == HIGH) {printf(" <- %s", paimei_ready);}
		return OK;
	}
	else
	{
		printf("PaiMei Problem! : %s\n", paimei_ready);
		return ERROR;
	}
}

int paimei_send_done(int paimei_fd, int display_level, int condition)
{
	char cc_option[9];
	cc_option[8] = 0x00;

	if(condition == OK)
		memcpy(cc_option, "Done:OK ", 8);
	else if( condition == ERROR)
		memcpy(cc_option, "Done:ERR", 8);
	else
	{
		printf("in PaiMei_send_done, bad condition!\n");
		return ERROR;
	}
	
	if( write(paimei_fd, cc_option, 8) == ERROR)
	{
		printf("in PaiMei_send_done, write returned an error!\n");
		return ERROR;
	}
	if (display_level == HIGH)
		printf("-> %s\r\n", cc_option);
	return OK;
}

int paimei_send_aOK(int paimei_fd, int display_level)
{
	char cc_option[4];
	cc_option[3] = 0x00;

	memcpy(cc_option, "aOK", 3);
	
	if( write(paimei_fd, cc_option, 3) == ERROR)
	{
		printf("in PaiMei_send_aOK, write returned an error!\n");
		return ERROR;
	}
	if (display_level == HIGH)
		printf("-> %s\r\n", cc_option);
	return OK;
}

int paimei_send_app_err(int paimei_fd, int display_level)
{
	char cc_option[4];
	cc_option[3] = 0x00;

	memcpy(cc_option, "err", 3);
	
	if( write(paimei_fd, cc_option, 3) == ERROR)
	{
		printf("in PaiMei_send_cc_option, write returned an error!\n");
		return ERROR;
	}
	if (display_level == HIGH)
		printf("-> %s\r\n", cc_option);
	return OK;
}

int paimei_send_cc_option(int paimei_fd, int completed_sessions, int sessions_per_pool, int display_level)
{
	char cc_option[4];
	cc_option[3] = 0x00;

	if(completed_sessions == (sessions_per_pool - 1) )
		memcpy(cc_option, "all", 3);
	else
		memcpy(cc_option, "one", 3);
	
	if( write(paimei_fd, cc_option, 3) == ERROR)
	{
		printf("in PaiMei_send_cc_option, write returned an error!\n");
		return ERROR;
	}
	if (display_level == HIGH)
		printf("-> %s\r\n", cc_option);
	return OK;
}

/*
int paimei_read_cc_session(struct session * session, int paimei_fd, int display_level, int stalk_type)
{
	char _cc_data[50];
	char * cc_data;
	char * index;
	int size_index;
	char sub_str[50];
	int cc_data_size;
	
	memset(_cc_data, 0x00, 50);
	cc_data = _cc_data;

	if( (cc_data_size = read(paimei_fd, cc_data, 50)) == ERROR)
	{
		printf("In paimei_read_cc_session, read returned an error!\n");
		return ERROR;
	}

	//look for error in string
	if( strchr(cc_data, '!') != NULL)
	{
		if (display_level==HIGH || display_level==MED)
		{
			printf(" <- %s/", cc_data);
		}
		return ERROR;
	}

	memset(sub_str, 0x00, 50);
	if( (index = strchr(cc_data, '/')) == NULL)
	{
		printf("strchr returned an error in paimei_read_cc_session!\n");
		printf("_cc_data=%s\n", _cc_data);
		printf("cc_data=%s\n", cc_data);
		return ERROR;
	}
	size_index = index - cc_data;
	memcpy(sub_str, cc_data, size_index);
	if (display_level == HIGH) {printf(" <- %s/", sub_str);}
	if(stalk_type == FUNCS)
		session->hit_funcs = atoi(sub_str);
	else if(stalk_type == BASIC_BLOCKS)
		session->hit_blocks = atoi(sub_str);
	else
	{
		printf("Bad stalk type!");
		return ERROR;
	}

	memset(sub_str, 0x00, 50);
	cc_data = index + 1;
	if( (index = strchr(cc_data, ',')) == NULL)
	{
		printf("strchr returned an error in paimei_read_cc_session!\n");
		printf("_cc_data=%s\n", _cc_data);
		printf("cc_data=%s\n", cc_data);
		return ERROR;
	}
	size_index = index - cc_data;
	memcpy(sub_str, cc_data, size_index);
	if (display_level == HIGH) {printf("%s", sub_str);}
	if(stalk_type == FUNCS)
		session->total_funcs = atoi(sub_str);
	else
		session->total_blocks = atoi(sub_str);
	
	memset(sub_str, 0x00, 50);
	//find if it crashed or not	
	cc_data = index + 2;
	if( (index = strchr(cc_data, '\r')) == NULL)
	{
		printf("strchr returned an error in paimei_read_cc_session!\n");
		printf("_cc_data=%s\n", _cc_data);
		printf("cc_data=%s\n", cc_data);
		return ERROR;
	}
	size_index = index - cc_data;
	memcpy(sub_str, cc_data, size_index);
	if (display_level == HIGH) {printf(", %s\r\n", sub_str);}
	session->crashed = atoi(sub_str);

	if(stalk_type == FUNCS)
	{
		session->func_fitness = (double)session->hit_funcs/session->total_funcs;
		printf("Session %d= %d hits (%.3f%)\t", session->position, session->hit_funcs, session->func_fitness * 100);
	}
	else
	{
		session->block_fitness = (double)session->hit_blocks/session->total_blocks;
		printf("Session %d= %d hits (%.3f%)\t", session->position, session->hit_blocks, session->block_fitness * 100);
	}

	if(display_level==MED || display_level==HIGH)
		printf("\r\n");
	fflush(stdout);

	return OK;	
}

int paimei_read_cc_pool(struct pool * pool, int paimei_fd, int display_level, int stalk_type)
{
	char _cc_data[50];
	char * cc_data;
	char * index;
	int size_index;
	char sub_str[50];
	int cc_data_size;
	
	memset(_cc_data, 0x00, 50);
	cc_data = _cc_data;

	if( (cc_data_size = read(paimei_fd, cc_data, 50)) == ERROR)
	{
		printf("In paimei_read_cc_pool, read returned an error!\n");
		return ERROR;
	}

	//look for error in string
	if( strchr(cc_data, '!') != NULL)
	{
		if (display_level==HIGH || display_level==MED)
		{
			printf(" <- %s/", cc_data);
		}
		return ERROR;
	}

	memset(sub_str, 0x00, 50);
	if( (index = strchr(cc_data, '/')) == NULL)
	{
		printf("strchr returned an error in paimei_read_cc_pool!\n");
		printf("_cc_data=%s\n", _cc_data);
		printf("cc_data=%s\n", cc_data);
		return ERROR;
	}
	size_index = index - cc_data;
	memcpy(sub_str, cc_data, size_index);
	if (display_level == HIGH) {printf(" <- %s/", sub_str);}
	pool->hit_funcs = atoi(sub_str);

	memset(sub_str, 0x00, 50);
	cc_data = index + 1;
	if( (index = strchr(cc_data, ',')) == NULL)
	{
		printf("strchr returned an error in paimei_read_cc_pool!\n");
		printf("_cc_data=%s\n", _cc_data);
		printf("cc_data=%s\n", cc_data);
		return ERROR;
	}
	size_index = index - cc_data;
	memcpy(sub_str, cc_data, size_index);
	if (display_level == HIGH) {printf("%s, ", sub_str);}
	pool->total_funcs = atoi(sub_str);
	memset(sub_str, 0x00, 50);

	cc_data = index + 1;
	if( (index = strchr(cc_data, '/')) == NULL)
	{
		printf("strchr returned an error in paimei_read_cc_pool!\n");
		printf("_cc_data=%s\n", _cc_data);
		printf("cc_data=%s\n", cc_data);
		return ERROR;
	}
	size_index = index - cc_data;
	memcpy(sub_str, cc_data, size_index);
	if (display_level == HIGH) {printf("%s/", sub_str);}
	pool->hit_blocks = atoi(sub_str);
	memset(sub_str, 0x00, 50);

	cc_data = index + 1;
	if( (index = strchr(cc_data, ',')) == NULL)
	{
		printf("strchr returned an error in paimei_read_cc_pool!\n");
		printf("_cc_data=%s\n", _cc_data);
		printf("cc_data=%s\n", cc_data);
		return ERROR;
	}
	size_index = index - cc_data;
	memcpy(sub_str, cc_data, size_index);
	if (display_level == HIGH) {printf("%s", sub_str);}
	pool->total_blocks = atoi(sub_str);
	memset(sub_str, 0x00, 50);

	//find if it crashed or not	
	cc_data = index + 2;
	if( (index = strchr(cc_data, '\r')) == NULL)
	{
		printf("strchr returned an error in paimei_read_cc_pool!\n");
		printf("_cc_data=%s\n", _cc_data);
		printf("cc_data=%s\n", cc_data);
		return ERROR;
	}
	size_index = index - cc_data;
	memcpy(sub_str, cc_data, size_index);
	if (display_level == HIGH) {printf(", %s\r\n", sub_str);}
	pool->crashed = atoi(sub_str);

	pool->func_fitness = (double)pool->hit_funcs/pool->total_funcs;
	pool->block_fitness = (double)pool->hit_blocks/pool->total_blocks;
	if(stalk_type == FUNCS)
		printf("\nTotal fitness for Pool %d: %d hits (%.4f%)\n\n", pool->position, pool->hit_funcs, pool->func_fitness * 100);
	else if(stalk_type == BASIC_BLOCKS)
		printf("\nTotal fitness for Pool %d: %d hits (%.4f%)\n\n", pool->position, pool->hit_blocks, pool->block_fitness * 100);
	else
	{
		printf("Bad stalk type!");
		return ERROR;
	}

	return OK;	
}
*/
