/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

int play_session(struct session * session, int play_mode, char * host, char * port, char * sport, int protocol, int delay,
		int wait, int display_level, int print_choice, int user_func, char * user_func_name, struct seed * seeds,
		char * types, int proxy)
{
	/*vars*/
        int err, x, y;
	int listenfd, connfd;
	struct addrinfo hints, *res;
	int flags = 0; //used to store fd flags
        int UDP_SERV=0; //Indicates if we're a UDP server or not; 0=no, 1=yes
        int UDP_SERV_READ=0; //Indicates if we've read once with recvfrom to set udpaddr; 0=no, 1=yes
        int TFTP=0; //for tftp hack 0=no need for it, 1=need it, 2=used it;
        int SRCPORT=ERROR; //created becuase tftp need to know the srcport.  Now allow user to choose.
        struct sockaddr_in udpaddr;  //added for udp
        socklen_t structLen;          //added for udp
	char buf[4];
	char * ptr = NULL;
	int current_size = 0;
	struct leg * leg;
	struct Tok * tok;

	/*Prepare Networking info*/
	bzero(&hints, sizeof(hints));
	hints.ai_socktype = protocol;

        if( (err = getaddrinfo(host, port, &hints, &res)) !=0) {
                TIME(); fprintf(stderr, "{{{%s}}}\n", gai_strerror(err)); return ERROR; }

        /*Bind to server IP for TCP */
        if( play_mode == SERVER && hints.ai_socktype == SOCK_STREAM)
        {
                if( (listenfd = setup_server(res)) == ERROR) {
                        TIME(); fprintf(stderr, "{{{Setup_server failed}}}\n\n"); return ERROR; }
        }

        /*set tftp hack*/
        if( (play_mode == CLIENT && atoi(port) == 69) && (hints.ai_socktype == SOCK_DGRAM) )
                TFTP=1;

        /*We're a UDP server*/ //also includes tftp hack
        if( play_mode == SERVER || TFTP==1 )
        {
                if(play_mode == SERVER)
                        UDP_SERV=1;
                structLen=sizeof(udpaddr);
                bzero(&udpaddr, structLen);
        }

	if(display_level==MED || display_level==HIGH) {printf("Playing session %d\n", session->position);}
	/*Connect to target if client mode*/
	if( play_mode == CLIENT)
	{
		if( sport == NULL)
		{
			if( (connfd = setup_client (res)) == ERROR) {
				TIME(); fprintf (stderr, "{{{Setup_client failed!}}}\n\n"); return APP_ERR; }
		}
		else
		{
			SRCPORT = atoi(sport);
			if( (connfd = setup_client_srcport (res, SRCPORT)) == ERROR) {
			TIME(); fprintf (stderr, "{{{Setup_client failed!}}}\n\n"); return APP_ERR; }
		}
	}

	/*Accept incoming connections if TCP server mode*/
	if( play_mode==SERVER && hints.ai_socktype==SOCK_STREAM)
	{
		if( (connfd = accept(listenfd, (struct sockaddr *)NULL, NULL)) < 0) {
			TIME(); fprintf(stderr, "{{{TCP server setup failed}}}\n\n"); return FATAL; }

		//should we timeout if no connections come?
	}

	/*Bind to server IP for UDP*/
	if( play_mode == SERVER && hints.ai_socktype == SOCK_DGRAM)
	{
		if( (connfd = setup_server(res)) == ERROR) {
			TIME(); fprintf(stderr, "{{{UDP server setup failed}}}\n\n"); return FATAL; }
	}

	/*for tftp hack*/
	if(TFTP == 2)
	{
		TFTP=1;
		UDP_SERV=0;
	}

	/*GET fd flags*/
	_flag(connfd, &flags);

	if( session == NULL)
	{
		printf("\nIn play_session, NULL session!\n");
		return ERROR;
	}
	leg = session->first_leg;
	if( leg == NULL)
	{
		printf("\nIn play_session, NULL leg! : \n");
		print_session(session);
		return ERROR;
	}

	/*where the magic happens*/
	while(leg != NULL)
	{
		if( leg->type == WRITE )
		{
			usleep(delay);	
			WRITEAGAIN:

			if( combine_toks_to_leg(leg, seeds, types, proxy) == ERROR)
			{
				printf("In play_session, combine_toks_to_leg returned an error.\n");
				printf("This session is messed up.\n");
				leg->error = COMBINE_ERROR;
				break;
			}

			/*process send data - i've not tested USERFUNCs with new -E mode*/
			if(user_func)
				USERFUNC(user_func_name, 0, leg->position, 0, 0, leg->data, leg->data_len, connfd, res);
			if (display_level==HIGH || (display_level==MED && session->position==0) )
			{
				printf("\t[%d][%d->] ",leg->position, leg->data_len);
				if(leg->data_len < 100)
				{
					_print(leg->data_len, print_choice, leg->data);
						printf("\n");
				}
				else
				{
					_print(100, print_choice, leg->data);
						printf("...\n");
				}
			}
		
			if (proxy == YES)
			{
				/*don't put in network byte order*/
				*((u_int32_t *)buf) = leg->data_len;
				/*first send total len*/
				if( write(connfd, buf , 4) < 1 )
				{
					TIME();fprintf(stderr, "{{{PROXY ERROR}}}\n"); return ERROR;
				}
				if( (ptr = _malloc( (leg->data_len) + (5 * leg->total_toks)  )) == NULL) {
					TIME(); fprintf(stderr, "{{{malloc error on play session}}}\n\n"); return FATAL;
				}
				current_size = 0;
				tok = leg->first_tok;
				while( tok != NULL)
				{
					memcpy(ptr+current_size, (char *)&(tok->type), 1);
					current_size += 1;
					memcpy(ptr+current_size, (char *)&(tok->dataLen), 4);
					current_size += 4;
					memcpy(ptr+current_size, tok->data, tok->dataLen);
					current_size += tok->dataLen;
					tok = tok->next;
				}
				if( write(connfd, ptr, leg->data_len) < 1 )
				{
					TIME();fprintf(stderr, "{{{PROXY ERROR}}}\n"); return ERROR;
				}
				free(ptr);
			}
			else
			{
				/*send data*/
				x = evolve_write(connfd, leg->data, leg->data_len, &flags, &UDP_SERV, &udpaddr, display_level, wait);
				if(x < 1 )
				{
					leg->error = WRITE_ERROR;
					if(display_level==HIGH){
					TIME(); perror("\t{{{Attack write error}}}");}
				}
			}
	
			/*tftp hack*/
			if(TFTP == 1)
				connfd=TftpHack(&TFTP, &UDP_SERV, &SRCPORT);
		}	
		else if( leg->type == READ )
		{
			usleep(delay);	
			READAGAIN:
			/*read*/
			x = evolve_read(connfd, &flags, &UDP_SERV, &UDP_SERV_READ, &TFTP, leg,
					wait, &udpaddr, print_choice, display_level, session->position);
			if( x == -2 )
			{
				leg->error = READ_ERROR_TIMEOUT;
				if(display_level==HIGH)
					perror("\t{{No data sent before timeout (R -2)}}");
			}
			else if(x == -1 && (display_level==HIGH) )
			{
				leg->error = READ_ERROR_DEADSOCKET;
				if(display_level==HIGH){
					TIME(); perror("\t{{Dead Socket (R -1)}}");}
			}

			/*again, i've not tested user_funcs yet*/
			if(user_func==1 && x!=-1)
			{
				y=USERFUNC(user_func_name, x, leg->position, 1, 0, leg->data, leg->data_len, connfd, res);
				if( y==-1)
				{
					leg->error = YES;
					if(display_level==HIGH)
						fprintf(stderr, "\t{{Didn't like what we got...Let's read again.}}\n");
					goto READAGAIN;
				}
				if( y==1)
				{
					leg->error = YES;
					if(display_level==HIGH)
						fprintf(stderr, "\t{{Didn't like what we got...Let's write again.}}\n");
					leg = leg->prev;
					goto WRITEAGAIN;
				}
			}
		}
		else
		{
			TIME();fprintf(stderr, "{{{Bad READ|WRITE type}}}\n"); return ERROR;
		}

		if (leg->error == READ_ERROR_DEADSOCKET)  //what about write errors?
			break;
		else
			leg = leg->next;
	}
	close(connfd);

	return OK;	
}
