/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

int print_pools(struct pools * pools)
{

	struct pool * tmp_pool = NULL;
	struct session * tmp_session = NULL;
	struct leg * tmp_leg = NULL;
	struct Tok * tmp_tok = NULL;

	tmp_pool = pools->first_pool;
	while ( tmp_pool != NULL)
	{	
		printf("\n<pool number: %d; sessions in pool = %d>",
			tmp_pool->position, tmp_pool->total_sessions);
		
		tmp_session = tmp_pool->first_session;
		while ( tmp_session != NULL)
		{
			printf("\n<<session number: %d; legs in session = %d>>\n",
				tmp_session->position, tmp_session->total_legs);

			tmp_leg = tmp_session->first_leg;
			while ( tmp_leg != NULL)
			{
				
				if( tmp_leg->type == WRITE )
					printf("<<<leg number: %d; toks in leg=%d; type=WRITE; error=%d; total data_len=%d>>>\n",
						tmp_leg->position, tmp_leg->total_toks, tmp_leg->error, tmp_leg->data_len);
				else
					printf("<<<leg number: %d; toks in leg=%d; type=READ; error=%d; total data_len=%d>>>\n",
						tmp_leg->position, tmp_leg->total_toks, tmp_leg->error, tmp_leg->data_len);

				if( tmp_leg->type == WRITE )
				{
					tmp_tok = tmp_leg->first_tok;
					Print_Toks(tmp_tok, "all", "short");
				}

				tmp_leg = tmp_leg->next;
			}

			tmp_session = tmp_session->next;
		}

		tmp_pool = tmp_pool->next;
	}
	printf("\n");
}

int print_pool(struct pool * pool)
{

	struct session * tmp_session = NULL;
	struct leg * tmp_leg = NULL;
	struct Tok * tmp_tok = NULL;

	printf("\n<pool number: %d; sessions in pool = %d>",
		pool->position, pool->total_sessions);

	tmp_session = pool->first_session;
	while ( tmp_session != NULL)
	{
		printf("\n<<session number: %d; legs in session = %d>>\n",
			tmp_session->position, tmp_session->total_legs);

		tmp_leg = tmp_session->first_leg;
		while ( tmp_leg != NULL)
		{
			if( tmp_leg->type == WRITE )
				printf("<<<leg number: %d; toks in leg=%d; type=WRITE; error=%d; total data_len=%d>>>\n",
					tmp_leg->position, tmp_leg->total_toks, tmp_leg->error, tmp_leg->data_len);
			else
				printf("<<<leg number: %d; toks in leg=%d; type=READ; error=%d; total data_len=%d>>>\n",
					tmp_leg->position, tmp_leg->total_toks, tmp_leg->error, tmp_leg->data_len);

			if( tmp_leg->type == WRITE )
			{
				tmp_tok = tmp_leg->first_tok;
				Print_Toks(tmp_tok, "all", "short");
			}

			tmp_leg = tmp_leg->next;
		}

		tmp_session = tmp_session->next;
	}
	printf("\n");
}

int print_session(struct session * session)
{

	struct leg * tmp_leg = NULL;
	struct Tok * tmp_tok = NULL;

	printf("\n<<session number: %d; legs in session = %d>>\n",
		session->position, session->total_legs);

	tmp_leg = session->first_leg;
	while ( tmp_leg != NULL)
	{
		if( tmp_leg->type == WRITE )
			printf("<<<leg number: %d; toks in leg=%d; type=WRITE; error=%d; total data_len=%d>>>\n",
				tmp_leg->position, tmp_leg->total_toks, tmp_leg->error, tmp_leg->data_len);
		else
			printf("<<<leg number: %d; toks in leg=%d; type=READ; error=%d; total data_len=%d>>>\n",
				tmp_leg->position, tmp_leg->total_toks, tmp_leg->error, tmp_leg->data_len);

		if( tmp_leg->type == WRITE )
		{
			tmp_tok = tmp_leg->first_tok;
			Print_Toks(tmp_tok, "all", "short");
		}

		tmp_leg = tmp_leg->next;
	}
	printf("\n");
}

int print_leg(struct leg * leg)
{

	struct Tok * tmp_tok = NULL;

	printf("<<<leg number: %d; toks in leg=%d; type=%d; error=%d; total data_len=%d>>>\n",
		leg->position, leg->total_toks, leg->type, leg->error, leg->data_len);

	Print_Toks(leg->first_tok, "all", "short");
	printf("\n");
}

int print_evolution(struct pools * pools, int print_choice)
{

	struct pool * tmp_pool = NULL;
	struct session * tmp_session = NULL;
	struct leg * tmp_leg = NULL;
	struct Tok * tmp_tok = NULL;

	tmp_pool = pools->first_pool;
	while ( tmp_pool != NULL)
	{	
		tmp_session = tmp_pool->first_session;

		printf("\nPool %d, session %d with %d total legs:\n",
			tmp_pool->position, tmp_pool->total_sessions, tmp_session->total_legs);

			tmp_leg = tmp_session->first_leg;
			while ( tmp_leg != NULL)
			{
				if( tmp_leg->type == WRITE)
				{
					if( combine_toks_to_leg(tmp_leg, NULL, NULL, NO) == ERROR)
		                        {
		                                printf("In print_evolution, combine_toks_to_leg returned an error.\n");
		                                printf("This session is messed up.\n");
		                                tmp_leg->error = COMBINE_ERROR;
		                                break;
		                        }
					printf("[%d][%d->] ",tmp_leg->position, tmp_leg->data_len);
	                                if(tmp_leg->data_len<150)
	                                {
	                                        _print(tmp_leg->data_len, print_choice, tmp_leg->data);
                                                printf("\n");
	                                }
	                                else
	                                {
	                                        _print(150, print_choice, tmp_leg->data);
                                                printf("...TRUNCATED...\n");
	                                }

				}
				else
				{
					if(tmp_leg->data == NULL)
						printf("  [<--][%d]\n", tmp_leg->position);
					else
					{
						printf("  [<-prev read in-][%d]\n", tmp_leg->position);
		                                if(tmp_leg->data_len<150)
		                                {
		                                        _print(tmp_leg->data_len, print_choice, tmp_leg->data);
							 printf("\n");
		                                }
		                                else
		                                {
		                                        _print(150, print_choice, tmp_leg->data);
							printf("...TRUNCATED...\n");
		                                }

					}
				}

				tmp_leg = tmp_leg->next;
			}

		tmp_pool = tmp_pool->next;
	}
	printf("\n");
}

