/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

struct pools * process_generation(struct pools * pools, int generations, int stalk_type, int session_mutation_rate,
	int pool_crossover_rate, int pool_mutation_rate, int display_level, struct seed * seeds,
	u_int max_sessions_per_pool, u_int max_legs_per_session, u_int max_toks_per_leg, int sessions_fixed_or_max,
	int legs_fixed_or_max, int toks_fixed_or_max, char * types, int proxy)
{

	struct pool * tmp_pool = NULL;
        struct session * tmp_session = NULL;
        struct leg * tmp_leg = NULL;
        struct Tok * tmp_tok = NULL;
	struct fitness_list * fitness_list;
	int type_num;	
	int data_num;

	/*
	   This will get done in each generation. Primarily use the best sessions in each pool to
	   create new "kid" sessions. The new generation completely replaces the old generation (parents).
		-Leave the first session OF THE SORTED LIST alone. Elitism of 1.
	*/
        tmp_pool = pools->first_pool;
        while ( tmp_pool != NULL)
        {
		if( display_level==MED || display_level==HIGH )
			printf("\tCrossing sessions in Pool %d\n", tmp_pool->position);
		fitness_list = NULL;
		if( (fitness_list = create_fitness_list(tmp_pool->first_session, stalk_type)) == NULL)
		{
			printf("In process_generation create_fitness_list returned an error!\n");
			printf("Pool %d\n", tmp_pool->position);
			return NULL;
		}
		if( sort_fitness_list(fitness_list, tmp_pool->total_sessions) == ERROR)
		{
			printf("In process_generation sort_fitness_list returned an error!\n");
			printf("Pool %d\n", tmp_pool->position);
			return NULL;
		}
		if(display_level == MED || display_level == HIGH)
		{
			printf("Sorted List for pool %d:  ", tmp_pool->position);
			print_fitness_list(fitness_list);
		}

		//creates a totally new genereation of sessions, under the same pool
		if( sessions_crossover(tmp_pool, fitness_list, display_level, max_legs_per_session, max_toks_per_leg, proxy) == ERROR)
		{
			printf("In process_generation, sessions_crossover returned an error!\n");
			printf("Pool %d\n", tmp_pool->position);
			return NULL;
		}
		if(  destroy_fitness_list(fitness_list) == ERROR)
		{
			printf("In process_generation, destroy_fitness_list returned an error!\n");
			return NULL;
		}

                tmp_pool = tmp_pool->next;
        }

	/*
	   Mutate two tokens/session in each pool. One type mutation and one data mutation.
	   Leave the first session (will already be sorted) in each pool alone.
	*/
	if( (generations+1) % session_mutation_rate == 0 )
	{
	        tmp_pool = pools->first_pool;
		while ( tmp_pool != NULL)
	        {
	                tmp_session = tmp_pool->first_session;
	                while ( tmp_session != NULL)
	                {
				//Protect the first session
				if( tmp_session->position != 0)
				{
					//pick a random leg
					data_num = _rand(tmp_session->total_legs);
			  	   if( proxy == YES)
				   {
					if( display_level==MED || display_level==HIGH )
						printf("\tMutating: Pool %d, session %d, mutate data in leg %d\n",
							tmp_pool->position, tmp_session->position, data_num);
				   }
				   else
				   {
					type_num = _rand(tmp_session->total_legs);
					if( display_level==MED || display_level==HIGH )
						printf("\tMutating: Pool %d, session %d, mutate data in leg %d, and type in leg %d\n",
							tmp_pool->position, tmp_session->position, data_num, type_num);
				   }
		       	         	tmp_leg = tmp_session->first_leg;
					while ( tmp_leg != NULL)
		                        {
					   if( proxy == NO )
					   {
						if( tmp_leg->position == type_num )
						{
							//The read||write type could change... :)
							tmp_leg->type = _rand(2);
							tok_type_mutation(tmp_leg);
						}
					   }
	
						if( tmp_leg->position == data_num )
							tok_data_mutation(tmp_leg, seeds);
	
		                                tmp_leg = tmp_leg->next;
		                        }
				}
			
	                        tmp_session = tmp_session->next;
	                }

	                tmp_pool = tmp_pool->next;
	        }
	}


	/*
	   Evolve from a higher level.  Work with pools, instead of sessions within a pool.
		-Leave the first pool OF THE SORTED LIST alone.
		-Also Note: If the session number is fixed, and a crossover operation results
		in a pool with less sessions that the fixed number, newly created sessions
		will be used to fill the pool up.  In this case, the pool mutation operation
		isn't adding much to the game.
	*/
	if( (generations+1) % pool_crossover_rate == 0 )
	{
		/*
		   To crossover pools is to bascially reorder the sessions with other highly fit pools.
			-don't change best fit pool
			-from top x% tipcally recreate other pools with ordering from the highest
			-all but the top 1 die
		*/
		if( display_level==MED || display_level==HIGH )
			printf("\n\tCrossing Pools:\n");
		fitness_list = create_fitness_list( (struct session *) pools->first_pool, stalk_type);
		if ( sort_fitness_list(fitness_list, pools->total_pools) == ERROR)
			return NULL;
		if (display_level == HIGH)
		{
			printf("Sorted List of POOLS\n");
			print_fitness_list(fitness_list);
		}

		if( (pools = pool_crossover(pools, fitness_list, display_level, max_sessions_per_pool,
				sessions_fixed_or_max, max_legs_per_session, legs_fixed_or_max,
				max_toks_per_leg, toks_fixed_or_max, seeds, types, proxy)) == NULL)
		{
			printf("In process_generation, pool_crossover returned an error!\n");
			return NULL;
		}

		if(  destroy_fitness_list(fitness_list) == ERROR)
		{
			printf("In process_generation, destroy_fitness_list returned an error!\n");
			return NULL;
		}
	}

	/*
	   Mutate each pool (except 1st) by either adding or deleting a random session.
	   Leave the first session (should already be sorted) in each pool alone.
	*/
	if( (generations+1) % pool_mutation_rate == 0 )
	{
        	tmp_pool = pools->first_pool;
        	while ( tmp_pool != NULL)
        	{
		   if( proxy == YES )
		   {
			//leave 1st pool alone
			if( tmp_pool->position !=0 )
			{
				if( display_level==MED || display_level==HIGH )
					printf("\tMutating Pool %d, by deleting than readding a session.\n", tmp_pool->position);
				remove_random_session(tmp_pool);
				add_type_session(tmp_pool, max_sessions_per_pool, sessions_fixed_or_max, types, seeds);
			}
		   }
		   else
		   {
			//leave 1st pool alone
			if( tmp_pool->position !=0 )
			{
			   if(sessions_fixed_or_max == FIXED)
			   {
					if( display_level==MED || display_level==HIGH )
						printf("\tMutating Pool %d, by deleting than readding a session.\n", tmp_pool->position);
					remove_random_session(tmp_pool);
					add_random_session(tmp_pool, max_sessions_per_pool, sessions_fixed_or_max,
						max_legs_per_session, legs_fixed_or_max, max_toks_per_leg,
						legs_fixed_or_max, seeds);
			   }
			   else
			   {
				if (_rand(100) < 50)
				{
					if( display_level==MED || display_level==HIGH )
						printf("\tMutating Pool %d, by adding a session.\n", tmp_pool->position);
					add_random_session(tmp_pool, max_sessions_per_pool, sessions_fixed_or_max,
						max_legs_per_session, legs_fixed_or_max, max_toks_per_leg,
						toks_fixed_or_max, seeds);
				}
				else
				{
					if( display_level==MED || display_level==HIGH )
						printf("\tMutating Pool %d, by deleting a session.\n", tmp_pool->position);
					remove_random_session(tmp_pool);
				}
			   }
			}
		   }
                	tmp_pool = tmp_pool->next;
		}
	}

	return pools;
}
