/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

int _flag(int fd, int * flags)
{
        if (( (*flags) = fcntl(fd, F_GETFL)) == -1)
	{
                return(ERROR);
	}
}

int _non_blocking(int fd, int * flags)
{
        if (fcntl(fd, F_SETFL, (*flags)|O_NONBLOCK ) == -1)
	{
        	return(ERROR);
	}
}

int evolve_write(int fd, u_char * data, u_int size, int * flags, int * UDP_SERV, void * udpaddr, int display_level, u_int wait)
{
	int i=0; 
	int x=0; 
	int print_wait = 1;

	if ( _non_blocking(fd, flags) == ERROR)
		return ERROR;
	for(i=0; i<wait; i++)
	{
		if( *(UDP_SERV) == 1)
			x=sendto(fd, data, size, 0, (struct sockaddr *) udpaddr, sizeof(*udpaddr) );
		else
			x=write(fd, data, size);
		if (x == -1 || x == 0 )
			return x;
		else if (x == size)
			return x;
		if (display_level==HIGH && print_wait)
		{
			printf("\t~~waiting to write data~~\n");
			print_wait = 0;
		}
		usleep(10000);
	}
	if (display_level==HIGH)
		printf("\t~~~~Failed to write, moving on. %d sent.~~~~\n", x);
	return x;
}

int evolve_read(int fd, int * flags, int * UDP_SERV, int * UDP_SERV_READ, int * TFTP, struct leg * leg,
	u_int wait, void * udpaddr, u_int print_choice, int display_level, int position)
{
	u_int equal=0; //0 means the received and stored were the same
	int x=0;
	char tmp_buf[1500];
	int y=0;
	int retval=0;
	socklen_t structLen = sizeof(*udpaddr);
	int print_wait = 1;

	if( (*UDP_SERV == 1) && (*UDP_SERV_READ == 0) )
	{
		if (*(TFTP) > 0)
		{
			if ( _non_blocking(fd, flags) == ERROR)
				return ERROR;
			while( (x = recvfrom(fd, tmp_buf, 1500, 0, (struct sockaddr *) udpaddr, &structLen)) !=0)
			{
				y++;
				if(x > 0)
					break;
				usleep(1);
				if ( y > 13+wait )
				{
					if (display_level==HIGH)
					{
						printf("\t~~~~moving on without that data~~~~\n");
						fflush(stdout);
					}
					break;
				}
				if (y>13)
				{
					if (display_level==HIGH && print_wait)
					{
						printf("\t~~waiting to read data~~\n");
						print_wait = 0;
					}
					usleep(10000);
				}
			}
		}
		else
		{
			//we want blocking here, should we se it to that? should be that.
			x = recvfrom(fd, tmp_buf, 1500, 0, (struct sockaddr *) udpaddr, &structLen);
		}

		*(UDP_SERV_READ)=1;
	}
	else
	{
		if( _non_blocking(fd, flags) == ERROR)
			return ERROR;
		while( (x = read(fd, tmp_buf, 1500)) != 0)
		{
			y++;
			if(x > 0)
				break;
			usleep(1);
			if ( y > 13+wait )
			{
				if (display_level==HIGH)
				{
					printf("\t~~~~moving on without that data~~~~\n");
					fflush(stdout);
				}
				break;
			}
			if (y>13)
			{
				if (display_level==HIGH && print_wait)
				{
					printf("\t~~waiting to read data~~\n");
					print_wait = 0;
				}
				usleep(10000);
			}
		}
	}
	if( (x < 1) || (x > 1500) )
	{
		if(errno == EAGAIN)
			retval=-2;  ///Server didn't send data, but socket is still ok
		else
			retval=-1;
	}
	else
	{
		/*Right now this compare is just form printing, but could it be used for some type of AI?*/
		/*furthermore, should this whole block of code be up a level?*/
		if( display_level==HIGH || (display_level==MED && position==0) )
		{
			if(leg->data != NULL)
			{
				equal=memcmp(tmp_buf, leg->data, leg->data_len);
				if( equal==0 )
					printf("\t  [<-SameData-%d/", leg->data_len);
				else
					printf("\t  [<-DiffData-%d/", leg->data_len);
			}
			else
				printf("\t  [<-NoPrevData-0/", leg->data_len, leg->position);
		}
		
		/*free the old, and move in the new!*/
		if(leg->data != NULL)
			free(leg->data);
		leg->data_len = x;
		if( (leg->data = _malloc(leg->data_len)) == NULL)
		{
			printf("In evolve_read, _malloc returned and error!\n");
			return ERROR;
		}
		memcpy(leg->data, tmp_buf, leg->data_len);

		if( display_level==HIGH || (display_level==MED && position==0) )
		{
			printf("%d][%d] ", leg->data_len, leg->position);
			if( leg->data_len < 100)
			{
				_print(leg->data_len, print_choice, leg->data);
				printf("\n");
			}
			else
			{
				_print(100, print_choice, leg->data);
				printf("...\n");
			}
		}
	}

	return retval;
}
