/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com
    Copyright (c) 2007, VDA Labs, LLC
    Jared DeMott, jdemott@vdalabs.com, www.vdalabs.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

//THIS IS A VERY IMPORTANT FUNC TO HAVE GOOD HEURSITICS/WEIGHTS IN
//Is there a way I can weight state/path commands (i.e. USER) less to fuzz than the data after it?  Would we want to?
//-1=error. 0=ok.
int Fuzz(struct Tok * tok, int total_toks, int mutation_rate)
{
        int ret=0;
	int fuzz=0;
        u_int x=0;

        if( (tok->data == NULL) || (tok->dataLen==0) ) {
		TIME(); fprintf(stderr, "{{Null data or 0 dataLen in Fuzz!}}\n"); return -1;} //should this be an error??

	x=_rand(100);
	if(mutation_rate == HIGH)
	{
		//30% chance to fuzz general strings or cmdvars
		if( ((tok->type == ASCII) || (tok->type == STRING) || (tok->type==ASCII_CMDVAR)) && (x < 30) )
			fuzz=1;
		//10% to fuzz cmds
		else if( (tok->type == ASCII_CMD) && (x < 10) )
			fuzz=1;
		//20% to fuzz mixed 
		else if( (tok->type == MIXED) && (x < 20) )
			fuzz=1;
	}
	else if( mutation_rate == MED)
	{
		//15% chance to fuzz general strings or cmdvars
		if( ((tok->type == ASCII) || (tok->type == STRING) || (tok->type==ASCII_CMDVAR)) && (x < 15) )
			fuzz=1;
		//5% to fuzz cmds
		else if( (tok->type == ASCII_CMD) && (x < 5) )
			fuzz=1;
		//10% to fuzz mixed 
		else if( (tok->type == MIXED) && (x < 10) )
			fuzz=1;
	}
	else
	{
		//7% chance to fuzz general strings or cmdvars
		if( ((tok->type == ASCII) || (tok->type == STRING) || (tok->type==ASCII_CMDVAR)) && (x < 7) )
			fuzz=1;
		//2% to fuzz cmds
		else if( (tok->type == ASCII_CMD) && (x < 2) )
			fuzz=1;
		//5% to fuzz mixed 
		else if( (tok->type == MIXED) && (x < 5) )
			fuzz=1;
	}

	//the below code belongs to the above three
	if (fuzz)
	{
		x=_rand(100);
                if( tok->numeric==1 && (x < 25) ) //25% to fuzz if num
		{
                        if( (ret+=Fuzz_Numeric(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Numeric returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
                else if (tok->numeric==0 && (x < 3) ) //3% to fuzz as num even if not
		{
                        if( (ret+=Fuzz_Numeric(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Numeric2 returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
                x=_rand(100);
                if( (tok->bracket==1 || tok->quote==1 || tok->special==1 || tok->other==1) && (x < 25) ) //25% if
		{
                        if( (ret+=Fuzz_Parsers(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Parsers returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
                else if( (x < 3) )
		{
                        if( (ret+=Fuzz_Parsers(tok)) < 0){ //3% if not
				TIME(); fprintf(stderr, "{{Fuzz_Parsers2 returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
                if( _rand(100) < 25 )
		{
                        if( (ret+=Fuzz_BitFlip(tok)) < 0){ //25% to flip a bit
				TIME(); fprintf(stderr, "{{Fuzz_BitFlip returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
                if( _rand(100) < 25 )
		{
                        if( (ret+=Fuzz_Buffer(tok)) < 0){ //25% to add some junk
				TIME(); fprintf(stderr, "{{Fuzz_Buffer returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
                if( _rand(100) < 12 )
		{
                        if( (ret+=Fuzz_Format(tok)) < 0){ //12% to add a format char
				TIME(); fprintf(stderr, "{{Fuzz_Format returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
	}

	if( mutation_rate == HIGH)
	{
		//4% to fuzz spaces
		if( (tok->type == ASCII_SPACE) && (_rand(100) < 4) )
		{
			if( (ret+=Fuzz_ASCII_SPACE(tok)) < 0){
					TIME(); fprintf(stderr, "{{Fuzz_ASCII_SPACE returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
		//7% to fuzz string endings 
		if( ((tok->type == ASCII_END) || (tok->type == ASCII)) && (_rand(100) < 7) )  //should ascii be on here?
		{
			if( (ret+=Fuzz_END(tok)) < 0){
					TIME(); fprintf(stderr, "{{Fuzz_ASCII_END returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
		//4% to fuzz endings 
		if( ((tok->type == BINARY_END) || (tok->type == STRING)) && (_rand(100) < 4)  )
		{
			if( (ret+=Fuzz_END(tok)) < 0){
					TIME(); fprintf(stderr, "{{Fuzz_END returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
	}
	if( mutation_rate == MED)
	{
		//2% to fuzz spaces
		if( (tok->type == ASCII_SPACE) && (_rand(100) < 2) )
		{
			if( (ret+=Fuzz_ASCII_SPACE(tok)) < 0){
					TIME(); fprintf(stderr, "{{Fuzz_ASCII_SPACE returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
		//4% to fuzz string endings 
		if( ((tok->type == ASCII_END) || (tok->type == ASCII)) && (_rand(100) < 4) )  //should ascii be on here?
		{
			if( (ret+=Fuzz_END(tok)) < 0){
					TIME(); fprintf(stderr, "{{Fuzz_ASCII_END returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
		//2% to fuzz endings 
		if( ((tok->type == BINARY_END) || (tok->type == STRING)) && (_rand(100) < 2)  )
		{
			if( (ret+=Fuzz_END(tok)) < 0){
					TIME(); fprintf(stderr, "{{Fuzz_END returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
	}
	if( mutation_rate == LOW)
	{
		//1% to fuzz spaces
		if( (tok->type == ASCII_SPACE) && (_rand(100) < 1) )
		{
			if( (ret+=Fuzz_ASCII_SPACE(tok)) < 0){
					TIME(); fprintf(stderr, "{{Fuzz_ASCII_SPACE returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
		//1% to fuzz string endings 
		if( ((tok->type == ASCII_END) || (tok->type == ASCII)) && (_rand(100) < 1) )  //should ascii be on here?
		{
			if( (ret+=Fuzz_END(tok)) < 0){
					TIME(); fprintf(stderr, "{{Fuzz_ASCII_END returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
		//1% to fuzz endings 
		if( ((tok->type == BINARY_END) || (tok->type == STRING)) && (_rand(100) < 1)  )
		{
			if( (ret+=Fuzz_END(tok)) < 0){
					TIME(); fprintf(stderr, "{{Fuzz_END returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
	}

	fuzz=0;
	x=_rand(100);
	if( mutation_rate == HIGH)
	{
		if( (tok->type == SESSION_ID) && (x<50) )
			fuzz=1;
		else if( (tok->type == FLAGS) && (x<8) )
			fuzz=1;
		else if( (tok->type == PROTO_ID) && (x<8) )
			fuzz=1;
		else if( (tok->type == LEN) && (x<50) )
			fuzz=1;
		else if( (tok->type == HASH) && (x<50) )
			fuzz=1;
		else if( (tok->type == BINARY) && (x<15) )
			fuzz=1;
	}
	else if( mutation_rate == MED)
	{
		if( (tok->type == SESSION_ID) && (x<25) )
			fuzz=1;
		else if( (tok->type == FLAGS) && (x<5) )
			fuzz=1;
		else if( (tok->type == PROTO_ID) && (x<5) )
			fuzz=1;
		else if( (tok->type == LEN) && (x<25) )
			fuzz=1;
		else if( (tok->type == HASH) && (x<25) )
			fuzz=1;
		else if( (tok->type == BINARY) && (x<8) )
			fuzz=1;
	}
	else
	{
		if( (tok->type == SESSION_ID) && (x<10) )
			fuzz=1;
		else if( (tok->type == FLAGS) && (x<2) )
			fuzz=1;
		else if( (tok->type == PROTO_ID) && (x<2) )
			fuzz=1;
		else if( (tok->type == LEN) && (x<10) )
			fuzz=1;
		else if( (tok->type == HASH) && (x<10) )
			fuzz=1;
		else if( (tok->type == BINARY) && (x<4) )
			fuzz=1;
	}
	
	//the below code belongs to the above 6
	if (fuzz)
	{
		x=_rand(100);
                if( tok->numeric==1 && (x < 25) ) //25% to fuzz if num
		{
                        if( (ret+=Fuzz_Numeric(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Numeric returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
                else if (tok->numeric==0 && (x < 3) ) //3% to fuzz as num even if not
		{
                        if( (ret+=Fuzz_Numeric(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Numeric2 returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
                x=_rand(100);
                if( (tok->bracket==1 || tok->quote==1 || tok->special==1 || tok->other==1) && (x < 25) ) //25% if
		{
                        if( (ret+=Fuzz_Parsers(tok)) < 0){
				TIME(); fprintf(stderr, "{{Fuzz_Parsers returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
                else if( (x < 3) )
		{
                        if( (ret+=Fuzz_Parsers(tok)) < 0){ //3% if not
				TIME(); fprintf(stderr, "{{Fuzz_Parsers2 returned -1}}\n"); Print_Toks(tok, "1", "long");}
		}
                if( _rand(100) < 20 )
		{
                        if( (ret+=Fuzz_BitFlip(tok)) < 0){ //25% to flip a bit
				TIME(); fprintf(stderr, "{{Fuzz_BitFlip returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
                if( _rand(100) < 8 )
		{
                        if( (ret+=Fuzz_Buffer(tok)) < 0){ //25% to add some junk
				TIME(); fprintf(stderr, "{{Fuzz_Buffer returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
                if( _rand(100) < 4 )
		{
                        if( (ret+=Fuzz_Format(tok)) < 0){ //12% to add a format char
				TIME(); fprintf(stderr, "{{Fuzz_Format returned -1}}\n"); Print_Toks(tok, "1", "long");} 
		}
		//special case for bin fuzzing.  pass directly to Fuzz_Integer
                if( (_rand(100) < 20) && (tok->data != NULL) && (tok->data > 0) )
		{
			u_int16_t size8=0;
			u_int16_t size16=0;
			u_int16_t size32=0;
			if( tok->dataLen == 1)
			{
				size8=Fuzz_Integer( _ndata_to_size8(tok->data) );
				_size8_to_ndata(tok->data, size8);
			}
			else if( tok->dataLen == 2)
			{
				size16=Fuzz_Integer( _ndata_to_size16(tok->data) );
				_size16_to_ndata(tok->data, size16);
			}
			else if( tok->dataLen == 4)
			{
				size32=Fuzz_Integer( _ndata_to_size32(tok->data) );
				_size32_to_ndata(tok->data, size32);
			}
			else
				*((int*)(tok->data))=Fuzz_Integer( *((int*)(tok->data)) );
		}

	}

	if( tok->type == UNDEFINED ){
		TIME(); fprintf(stderr, "{{Undefined type in Fuzz}}\n"); Print_Toks(tok, "1", "long"); return -1;}

	//each function will need to report on it's own error
	if(ret < 0 ) {
		TIME(); fprintf(stderr, "{{ret=%d}}\n",ret); return -1;}
	else	
		return 0;
}
