/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

/*Evolve GPF*/
int evolve_mode(int argc, char ** argv)
{

	/*Variable Definitions*/
        int total_evaluations, err, x;
	time_t start, now, diff;
	double min, hr;
	int result;
	int unresponsive=0;

	u_int play_mode;
        u_char *host, *port, *sport;
	u_int protocol;
        u_int wait;
        u_int delay;
	u_int print_choice; //0=ascii; 1=hex; 2=auto
	u_int display_level; //0=low; 1=med; 2=high
	u_int total_pools;
	u_int sessions_per_pool;
	int sessions_fixed_or_max;
	u_int legs_per_session;
	int legs_fixed_or_max;
	u_int toks_per_leg;
	int toks_fixed_or_max;
	u_int current_generation;
	u_int generations;
	u_int session_mutation_rate;
	u_int pool_crossover_rate;
	u_int pool_mutation_rate;
	u_int diversity;
        u_int user_func=1; //0=no user func; 1=yes -- to increasing GPF's speed
	char * user_func_name=NULL;  //holds the name of the user_func passed from cmdline
	int started_from_zero;
	int already_deleted =0;
	int recieved_types = NO;
	
	char *mysql_host, *mysql_user, *mysql_passwd;
	int proxy = NO;
	u_int gpf_id;
	MYSQL * mysql = NULL;
	int first_time = 1;

	int paimei_fd;
	struct addrinfo paimei_hints, *paimei_res;
	char *paimei_host, *paimei_port;
        int stalk_type;
	struct seed * seeds = NULL;

	char * types = NULL;
	struct pools * pools = NULL;
	struct pool * tmp_pool = NULL;
	struct session * tmp_session = NULL;

	/*seed randomness*/
	struct timeval t1;
	gettimeofday(&t1,NULL);
	srand(t1.tv_usec);

	/*check number of command line arguments and handle them*/
	if (argc < 33){
		TIME(); fprintf(stderr, "{{{Not enough args}}}\n"); Evolving_GPF_usage();
	}

	/*handle args*/
	mysql_host = argv[2];
	mysql_user = argv[3];
	mysql_passwd = argv[4];
	gpf_id = (u_int)(atoi(argv[5]));
	current_generation = (u_int)(atoi(argv[6]));

	paimei_host = argv[7];
	paimei_port = argv[8];
	if ( strcasecmp(argv[9], "funcs") == 0)
		stalk_type = FUNCS;
	else if ( strcasecmp(argv[9], "basic_blocks") == 0 )
		stalk_type = BASIC_BLOCKS;
	else{ printf("{{{Bad Stalk Type}}}\n\n"); Evolving_GPF_usage(); }

	if( strcasecmp(argv[10], "client") == 0 )
		play_mode = CLIENT;
	else if ( strcasecmp(argv[10], "server") == 0)
		play_mode = SERVER;
	else{printf("{{{Bad play_mode.}}}\n\n"); Evolving_GPF_usage(); }
	host = argv[11];
	port = argv[12];
	sport = argv[13];
	if( strcasecmp(sport, "?") == 0 )
		sport = NULL;
	if ( strcasecmp(argv[14], "UDP") == 0)
		protocol = SOCK_DGRAM;
	else if ( strcasecmp(argv[14], "TCP") == 0 )
		protocol = SOCK_STREAM;
	else{printf("{{{Bad protocol}}}\n\n"); Evolving_GPF_usage(); }
	delay = (u_int)(atoi(argv[15]));
	wait = (u_int)(atoi(argv[16]));

	if( strcasecmp(argv[17], "low") == 0 )
		display_level = LOW;
	else if ( strcasecmp(argv[17], "med") == 0)
		display_level = MED;
	else if ( strcasecmp(argv[17], "high") == 0)
		display_level = HIGH;
	else{printf("{{{Bad display_level.}}}\n\n"); Evolving_GPF_usage(); }
	if( strcasecmp(argv[18], "ascii") == 0 )
		print_choice = TEXT;
	else if ( strcasecmp(argv[18], "hex") == 0)
		print_choice = HEX;
	else if ( strcasecmp(argv[18], "auto") == 0)
		print_choice = AUTO;
	else{printf("{{{Bad print_choice.}}}\n\n"); Evolving_GPF_usage(); }

	total_pools = (u_int)(atoi(argv[19]));

	sessions_per_pool = (u_int)(atoi(argv[20]));
	if( strcasecmp(argv[21], "Fixed") == 0 )
		sessions_fixed_or_max = FIXED;
	else if ( strcasecmp(argv[21], "Max") == 0)
		sessions_fixed_or_max = MAXI;
	else{printf("{{{Bad print_choice.}}}\n\n"); Evolving_GPF_usage(); }

	legs_per_session = (u_int)(atoi(argv[22]));
	if( strcasecmp(argv[23], "Fixed") == 0 )
		legs_fixed_or_max = FIXED;
	else if ( strcasecmp(argv[23], "Max") == 0)
		legs_fixed_or_max = MAXI;
	else{printf("{{{Bad print_choice.}}}\n\n"); Evolving_GPF_usage(); }

	toks_per_leg = (u_int)(atoi(argv[24]));
	if( strcasecmp(argv[25], "Fixed") == 0 )
		toks_fixed_or_max = FIXED;
	else if ( strcasecmp(argv[25], "Max") == 0)
		toks_fixed_or_max = MAXI;
	else{printf("{{{Bad print_choice.}}}\n\n"); Evolving_GPF_usage(); }

	generations = (u_int)(atoi(argv[26]));

	session_mutation_rate = (u_int)(atoi(argv[27]));
	pool_crossover_rate = (u_int)(atoi(argv[28]));
	pool_mutation_rate = (u_int)(atoi(argv[29]));

	if( (strcasecmp("none", argv[30]))==0)
		user_func=0;
	else
	{
		user_func=1;
		user_func_name = _malloc(strlen(argv[30]) );
		strncpy(user_func_name, argv[30], strlen(argv[30]) );
	}

	if( (strcasecmp("none", argv[31]))==0)
		seeds=NULL;
	else
	{
		if( (seeds=read_in_seeds(seeds, argv[31])) == NULL) //passes back the head of the chain
		{
			printf("In evole_mode, read_in_seeds returned an error!\n");
			return ERROR;
		}
	}

	if( (strcasecmp("no", argv[32]))==0)
		proxy=NO;
	else if( (strcasecmp("yes", argv[32]))==0)
	{
		proxy=YES;
		legs_per_session=1;
		legs_fixed_or_max = FIXED;
		toks_fixed_or_max = FIXED;
	}
	else{printf("{{{Bad proxy choice.}}}\n\n"); Evolving_GPF_usage(); }
	diversity = (u_int)(atoi(argv[33]));


	/*create/initialize pools-sessions based on proxy delievered type info*/
	if(proxy == YES && current_generation == 0)
	{
		if ( (types = recieve_type_information(host, port, display_level, print_choice)) == NULL)
		{
			printf("Failed to recieve type information from proxy Exiting...\n");
			return ERROR;
		}
		//get number of toks
		toks_per_leg=0;
		x=0;
		while(1)
		{
		        if( *((unsigned char *)(types+x)) == 0xff )
		                break;
		        toks_per_leg++;
		        x = x + 5;
		}
printf("toks_per_leg = %d\n", toks_per_leg);

		/*number of (Geno-)pools, sessions/pool (genotypes or candidate solutions), legs/session, and toks/leg*/
		if( (pools = types_create_pools(total_pools, sessions_per_pool, legs_per_session, toks_per_leg, seeds,
				sessions_fixed_or_max, legs_fixed_or_max, toks_fixed_or_max, types)) == NULL)
		{
			printf("Failed to properly create a group of pools (proxy type).  Exiting...\n");
			return ERROR;
		}

		if (display_level==MED)
		{
			printf("\n%d pools have been randomly filled with no more than %d sessions.\n",
				pools->total_pools, sessions_per_pool);	
			printf("Each session has not more than %d legs (pkts).\n", legs_per_session);	
			printf("Each leg has not more than %d tokens (data fields).\n", toks_per_leg);
			printf("The type/data of each token has been randomly determined.\n");
			if (seeds != NULL)
				printf("Seeds from %s were also used.\n", argv[31]); //don't forget to edit this argv if charge cmdline args!
		}
		started_from_zero = 1;
		recieved_types = YES;
	}

	/*start the connection to the mysql server*/
	if( (mysql = mysql_init(mysql)) == NULL)
	{
		printf("In evolve_mode, mysql_init returned an error!\n");
		return ERROR;
	}
	if( (mysql = mysql_real_connect(mysql, mysql_host, mysql_user, mysql_passwd, "gpf", 0, NULL, 0)) == NULL)
	{
		printf("In evolve_mode, mysql_real_connect returned an error!\n");
		return ERROR;
	}
	if (display_level == MED || display_level == HIGH) {printf("\nMysql connection successful.\n");}

	/* sem-stocastically create/initialize pools-sessions based on seed-file*/
	/*If new run, we create the pools*/
	if( current_generation == 0 && proxy == NO)
	{
		/*number of (Geno-)pools, sessions/pool (genotypes or candidate solutions), legs/session, and toks/leg*/
		if( (pools = create_pools(total_pools, sessions_per_pool, legs_per_session, toks_per_leg, seeds,
				sessions_fixed_or_max, legs_fixed_or_max, toks_fixed_or_max )) == NULL)
		{
			printf("Failed to properly create a group of pools.  Exiting...\n");
			return ERROR;
		}

		if (display_level==MED)
		{
			printf("\n%d pools have been randomly filled with no more than %d sessions.\n",
				pools->total_pools, sessions_per_pool);	
			printf("Each session has not more than %d legs (pkts).\n", legs_per_session);	
			printf("Each leg has not more than %d tokens (data fields).\n", toks_per_leg);
			printf("The type/data of each token has been randomly determined.\n");
			if (seeds != NULL)
				printf("Seeds from %s were also used.\n", argv[31]); //don't forget to edit this argv if charge cmdline args!
		}
		started_from_zero = 1;
	}
	else if( current_generation == 0 && proxy == YES)
	{
		started_from_zero = 1;
	}
	/*We continue from a previous run using this id*/
	else if (current_generation > 0 && recieved_types == NO && (proxy == YES || proxy == NO) )
	{
		printf("Peforming a restore with ID=%d, generation=%d\n", gpf_id, current_generation);
		if( (pools = restore(mysql, gpf_id, current_generation, proxy)) == NULL)
		{
			printf("Failure to restore, perhaps the id or generation was bad?\n");
			return ERROR;
		}
		printf("Done restoring genereation %d\n",current_generation);

		if( proxy == YES)
		{
			//recieve type info from proxy.
			if ( (types = recieve_type_information(host, port, display_level, print_choice)) == NULL)
			{
				printf("Failed to recieve type information from proxy Exiting...\n");
				return ERROR;
			}
			//get number of toks
			toks_per_leg=0;
			x=0;
			while(1)
			{
			        if( *(types+x)==0x00 )
			                break;
			        toks_per_leg++;
			        x = x + 5;
			}
		}

		started_from_zero = 0;
	}
	else{printf("{{{Bad generation or proxy parameters}}}\n\n"); Evolving_GPF_usage(); }
	

	if (display_level==HIGH)
		print_pools(pools);

	/*start connection to PaiMei (paimei) server*/
	bzero(&paimei_hints, sizeof(paimei_hints));
	paimei_hints.ai_socktype = SOCK_STREAM;
	if( (err = getaddrinfo(paimei_host, paimei_port, &paimei_hints, &paimei_res)) !=0) {
		TIME(); fprintf(stderr, "{{{Error in evolve_mode, getaddrinfo: %s}}}\n", gai_strerror(err)); exit(-1); }
	if( (paimei_fd = setup_client (paimei_res)) == ERROR) {
		TIME(); fprintf (stderr, "{{{PaiMei (paimei) connection failed!}}}\n\n"); return ERROR;}
	if (display_level == MED || display_level == HIGH) {printf("\nPaiMei connection successful.\n");}

	printf("\nDid people come from toads?  I don't think so, the world drips of intelligent design.\n");
	printf("However, I do believe a strong dog can become stronger due to micro-evolution or adaptation.\n");
	printf("So why shouldn't our fuzzers be able to evolve as well? :)\n\n");

	/*Do all of the below for each session in each pool for all generations*/
	total_evaluations=0;
	start = time(NULL);
	first_time = 1;
	for(; current_generation < generations; current_generation++)
	{
	   //This code allows us to "process the gen we re-start at before replay".
	   //I'm not sure we want to do that, so I'm taking it out for now.
	   //if(first_time==1 && current_generation > 0)
	  //	goto SKIP;

   	   /*If hits or crashes exist (for this id/gen or all) from previous runs delete them*/
	   if( started_from_zero == 1 && already_deleted==0)
	   {
	  	 if( delete_all_hits_and_crashes(mysql, mysql_host, mysql_user, mysql_passwd, "gpf", gpf_id,
				current_generation, pools, display_level, stalk_type) == ERROR)
		   {
	   		   printf("Failed to delete hits.\n");
		  	 if (paimei_send_done(paimei_fd, display_level, ERROR) == ERROR)
				printf("Failed to properly send Done.\n");
			   return ERROR;
		   }
		already_deleted = 1;
	   }
	   else if (started_from_zero ==0 && already_deleted == 0)
	   {
	  	 if( delete_hits_and_crashes(mysql, mysql_host, mysql_user, mysql_passwd, "gpf", gpf_id,
				current_generation, pools, display_level, stalk_type) == ERROR)
		   {
	   		   printf("Failed to delete hits.\n");
		  	 if (paimei_send_done(paimei_fd, display_level, ERROR) == ERROR)
				printf("Failed to properly send Done.\n");
			   return ERROR;
		   }
	   }
 
		tmp_pool = pools->first_pool;
		while ( tmp_pool != NULL )
		{
			if(display_level==MED || display_level==HIGH) {printf("Playing POOL %d\n", tmp_pool->position);}
			tmp_session = tmp_pool->first_session;
			unresponsive=0;
			while ( tmp_session != NULL )
			{
				//(1)// --> //*Send tag/target info to paimei*/
				if( paimei_send_target_tag(current_generation, tmp_pool->position, tmp_session->position,
					paimei_fd, display_level) == ERROR)
				{
					printf("Failed to send target_tag options!\n");
					return ERROR;
				}
	
				//(2)// <-- //*read in "ready to roll" from paimei*/
				if (is_paimei_ready(paimei_fd, display_level) == ERROR)
				{
					printf("Failed to recieve 'paimei ready'!\n");
					if( unresponsive < 6 )
					{
						unresponsive++;
						continue; //try again
					}
					else
					{
						printf("Too many failed attempts.  Quitting.\n");
						if (paimei_send_done(paimei_fd, display_level, ERROR) == ERROR)
							printf("Failed to properly send Done.\n");
						return ERROR;
					}
				}
				
				//(3)// <--> //*Send session to actual target under test*/  //on average how long does this take?
				result = play_session(tmp_session, play_mode, host, port, sport, protocol, delay,
                				wait, display_level, print_choice, user_func, user_func_name, seeds, types, proxy);
				if( result == ERROR)
				{
					printf("Session play error.\n");
					tmp_session->error = YES;
				}
				else if( result == APP_ERR )
				{
					if (paimei_send_app_err(paimei_fd, display_level) == ERROR)
						printf("Failed to properly send app err!\n");
					if( tmp_session->prev != NULL)
						tmp_session = tmp_session->prev;
					if( unresponsive < 6 )
					{
						unresponsive++;
						continue; // replay this one and the one before (sometimes)
					}
					else
					{
						printf("Too many failed attempts.  Quitting.\n");
						if (paimei_send_done(paimei_fd, display_level, ERROR) == ERROR)
							printf("Failed to properly send Done.\n");
						return ERROR;
					}
				}
				else if( result == FATAL )
				{
					if (paimei_send_done(paimei_fd, display_level, ERROR) == ERROR)
						printf("Failed to properly send Done.\n");
					return ERROR;
				}
	
				//(4)// --> //*Send aOK or err*/
				if (paimei_send_aOK(paimei_fd, display_level) == ERROR)
				{
					printf("Failed to properly send aOK\n");
					return ERROR;
				}
	
				tmp_session = tmp_session->next;
				total_evaluations++;
				unresponsive=0;
			}
		
			tmp_pool = tmp_pool->next;
		}

		/*calculate (hits) fitness for each session and pool*/
		if( calculate_fitness(mysql, mysql_host, mysql_user, mysql_passwd, "gpf", gpf_id,
					current_generation, pools, display_level, stalk_type) == ERROR)
		{
			printf("Failed to properly calculate fitness\n");
			if (paimei_send_done(paimei_fd, display_level, ERROR) == ERROR)
				printf("Failed to properly send Done.\n");
			return ERROR;
		}
		if( diversity == 1)
		{
			/*Calculate the diversity fitness for each session and pool? and add to total fitness*/
			if( calculate_diversity(mysql, mysql_host, mysql_user, mysql_passwd, "gpf", gpf_id,
						current_generation, pools, display_level, stalk_type) == ERROR)
			{
				printf("Failed to properly calculate diversity fitness\n");
				if (paimei_send_done(paimei_fd, display_level, ERROR) == ERROR)
					printf("Failed to properly send Done.\n");
				return ERROR;
			}
		}

		/*save*/
		printf("Successfully played generation %d. Saving to mysqldb.\n", current_generation);
		if( (err=save(mysql, gpf_id, current_generation, pools, display_level, started_from_zero)) != OK)
		{
			printf("In evolve_mode, save returned an error: %d\n", err);
			if ( err==2006 || err==2013 || err==1053)
			{
				printf("Try reconnecting, to mysql and retry!\n");
				sleep(4);
				if( (mysql=mysql_real_connect(mysql,mysql_host,mysql_user,mysql_passwd,"gpf",0,NULL,0)) == NULL)
				{
					printf("In evolve_mode, mysql_real_connect returned an error!\n");
					if (paimei_send_done(paimei_fd, display_level, ERROR) == ERROR)
						printf("Failed to properly send Done.\n");
					return ERROR;
				}
				if( save(mysql, gpf_id, current_generation, pools, display_level, started_from_zero) != OK)
				{	
					printf("Failed to fully save generation %d!\n", current_generation);
					if (paimei_send_done(paimei_fd, display_level, ERROR) == ERROR)
						printf("Failed to properly send Done.\n");
					return ERROR;
				}
			}
			else
			{
					printf("Failed to fully save generation %d!\n", current_generation);
					if (paimei_send_done(paimei_fd, display_level, ERROR) == ERROR)
						printf("Failed to properly send Done.\n");
					return ERROR;
			}
		}

		/*Now process this generation*/
		SKIP:
	   	first_time = 0;
		printf("Processing Generation %d ...\n", current_generation);
		if ( (pools = process_generation(pools, current_generation, stalk_type, session_mutation_rate,
			pool_crossover_rate, pool_mutation_rate, display_level, seeds,
			sessions_per_pool, legs_per_session, toks_per_leg, sessions_fixed_or_max,
			legs_fixed_or_max, toks_fixed_or_max, types, proxy)) == NULL)
		{
			printf("Failed to properly process generation!\n");
			if (paimei_send_done(paimei_fd, display_level, ERROR) == ERROR)
				printf("Failed to properly send Done.\n");
			return ERROR;
		}
		if( display_level==HIGH )
			print_pools(pools);
		now = time(NULL);
		diff = now - start;
		printf("Done processing. Time to play and process: %d total evaluations in %d seconds.\n", total_evaluations, diff);
		if( total_evaluations!=0 && diff!=0 )
			printf("%.2f sec/eval\n", (double)(now-start)/(double)total_evaluations );
		if( diff != 0 )
		{
			min = (double)diff/60;
			hr = min/60;
			printf("That's %.3f mins or %.3f hrs.\n", min, hr);
		}
		printf("\n");
	}

	if (paimei_send_done(paimei_fd, display_level, OK) == ERROR)
		printf("Failed to properly send Done.\n");

	printf("\nCompleted %d generations successfully.\n", current_generation);
	mysql_close(mysql);

	/*After we've processed all of the required generations, switch to a more aggressive fuzzing mode.*/
		//I don't think so, this should already have been done as a by-product
		//should this be handled one layer up?

	return OK;	
}
