/*  Copyright (c) 2007, VDA Labs, LLC
    Jared DeMott, jdemott@vdalabs.com, www.vdalabs.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

int filefuzz_mode(int argc, char **argv)
{
	/*Variable Definitions*/
        u_int in, none;
        u_short len;
        int err;
        u_int i, j, k, l, x, y;
        u_int loops=0;
	char filename[65];
	char dirname[45];

        u_int userFunc=1; //0=no user func; 1=yes -- to increasing GPF's speed
        char * example_filename=NULL;
        char * file_extension=NULL;
        char * dump_dir=NULL;
	int num_files = 0;
	int mutation_rate=MED;

	/*Global Variables*/
	u_int displayLevel=0; //0=low; 1=med; 2=high
	u_char userFuncName[64];  //holds the name of the userFunc passed from cmdline

	//It's a bit ugly have these globals static.
	//Up these if you have a maga huge session to replay
	u_int seq[3000];  //order in which legs took place  0=client speaking; 1=server speaking
	void * Data[1500]; //master list of each of the legs in order
	u_int DataSize[1500];  //used to hold the size of each pkt data
	u_int Count=0;  //total number of client and server legs
	void * readData[1500];  //used to hold a pointer to the data actually read from target
	u_int readDataSize[1500];  //used to hold the size of each pkt read from target
        u_int firstTime = 1;

	/*prepare the above args into a strange format for the pcap_loop func within, call_readin*/
	void * args[7];
        args[0] = seq;
        args[1] = Data;
        args[2] = DataSize;
        args[3] = &Count;
        args[4] = &firstTime;

	/*check number of command line arguments and handle them*/
	if (argc < 10){
		TIME(); fprintf(stderr, "{{{Not enough args}}}\n"); FileFuzz_usage(); }
	example_filename = argv[2];
	if( (file_extension = strstr(example_filename, ".") ) == NULL) {
		fprintf(stderr, "Bad file ... no extension on it!\n\n"); exit(-1); }
	file_extension++;
	/*Call the filefuzz_readin function*/
	if( filefuzz_readin(example_filename, seq, Data, DataSize, &Count) == ERROR){
		fprintf(stderr, "{{{Problem reading in example file!}}}\n\n"); exit(-1);}
	/*handle the rest of the args*/
	dump_dir = argv[3];
	umask(0);
	memset(dirname, 0x00, 40);
	if( (strcasecmp("same", argv[3]))==0)
	{
		snprintf(dirname, 40, "./_%s", example_filename);
		if( (mkdir(dirname, 0777) == ERROR) ) {
			fprintf(stderr, "Problem creating directory!\n"); exit(-1);}
	}
	else
	{
		snprintf(dirname, 40, "./%s", dump_dir);
		if( (mkdir(dirname, 0777) == ERROR) ) {
			fprintf(stderr, "Problem creating directory!\n"); exit(-1);}
	}
	srand( (u_int)(atoi(argv[4])) );
	num_files = atoi(argv[5]);
	if( (strcasecmp("none", argv[6]))==0)
		userFunc=0;
	else
		strncpy(userFuncName, argv[6], 60);
	u_char * tokPrint=argv[7];
	u_char * tokAid=argv[8];
	if( (strcasecmp("low",argv[9])) == 0 )
		mutation_rate=LOW;
	else if( (strcasecmp("med", argv[9])) == 0 )
		mutation_rate=MED;
	else if( (strcasecmp("high", argv[9])) == 0 )
		mutation_rate=HIGH;
	else{
		fprintf(stderr, "Bad mutation rate!\r"); exit(-1);}

	/*Tokenize*/
	struct Tok * Toks=NULL;
	struct Tok * tok=NULL;
	u_int size=0;
	int total_toks=0;
	u_char * data=NULL;
	data=Data[0];
	size=DataSize[0];

	printf("\nTokenizing File:\n");
	//change this and userFuncs code to dynamic code???
	//tokAids are listed here :: MAKE CHANGES HERE IF YOU ADDED A TOKAID ::
	if( (strcasecmp(tokAid, "GENERIC")) == 0)
		tok=GENERIC(tok, data, size, 0);
	else if( (strcasecmp(tokAid, "NORMAL_ASCII")) == 0)
		tok=NORMAL_ASCII(tok, data, size, 0);
	else if( (strcasecmp(tokAid, "AUTO_BINARY")) == 0)
		tok=AUTO_BINARY(tok, data, size, 0);
	else { TIME(); fprintf(stderr, "{{{No such tokAid!}}}\n\n"); exit(-1); }

	Toks=Find_First_Tok(tok);
	Print_Toks(Toks, "all", tokPrint);

	if( (total_toks = Count_Toks(Toks)) == ERROR){
		fprintf(stderr, "{{{No TOKENS!!}}}\n\n"); exit(-1);}
	/*FileFuzz fuzz*/
	printf("Beginning Fuzzy File Dump:\n");
	for(i=0; i<num_files; i++)
	{
		//Main fuzzy loop
		struct Tok * PktTok=NULL;

		//Build Fuzzy Leg
		if( (PktTok=Build_Fuzzy_Leg(Toks, RESPECT, 0, total_toks, mutation_rate)) == NULL )
		{
			TIME(); fprintf(stderr, "{{Build_Fuzzy_Leg failed!}}\n"); exit(-1);
		}
		//process attack??
		if(userFunc)
			USERFUNC(userFuncName, 0, 0, 0, 0, PktTok->data, PktTok->dataLen, 0, 0);
		//print attack
		printf("[%d][-p:%d->] ", loops, PktTok->dataLen);
		if(PktTok->dataLen<150)
		{
			_print(PktTok->dataLen, AUTO, PktTok->data);
				printf("\n");
		}
		else
		{
			_print(150, AUTO, PktTok->data);
				printf("...TRUNCATED...\n");
		}
		//write fuzzy files out to disk
		memset(filename, 0x00, 63);
		snprintf(filename, 63, "%s/%d.%s", dirname, loops, file_extension);
		if(write_file(filename, PktTok->data, PktTok->dataLen) == ERROR)
		{
			TIME(); fprintf(stderr, "{{{FileFuzz file write error}}}");
		}
		if ( Destroy_Toks(PktTok, "current") == NULL){
			TIME(); perror("{{PktTok Destroy error!}}");}

		if( (loops % 20) == 0 )
			printf("=================================Total Files: %d====================================\n\n", loops);
		loops++;
	}

	return(0);
}
