/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

/*Evolve GPF*/
int replay_evolve_mode(int argc, char ** argv)
{

	/*Variable Definitions*/
        u_int err, gen_count, session_count;
	time_t start, now;

	u_int play_mode;
        u_char *host, *port, *sport;
	u_int protocol;
        u_int wait;
        u_int delay;
	u_int print_choice; //0=ascii; 1=hex; 2=auto
	u_int display_level; //0=low; 1=med; 2=high
	u_int current_generation;
	u_int generations;
        u_int user_func=1; //0=no user func; 1=yes -- to increasing GPF's speed
	char * user_func_name=NULL;  //holds the name of the user_func passed from cmdline
	
	char *mysql_host, *mysql_user, *mysql_passwd;
	u_int mysql_id;
	MYSQL * mysql = NULL;

	struct pools * pools = NULL;
	struct pool * tmp_pool = NULL;
	struct session * tmp_session = NULL;

	int proxy = NO;

	/*seed randomness*/
	struct timeval t1;
	gettimeofday(&t1,NULL);
	srand(t1.tv_usec);

	/*check number of command line arguments and handle them*/
	if (argc < 18){
		TIME(); fprintf(stderr, "{{{Not enough args}}}\n"); Replay_evolving_GPF_usage();
	}

	/*handle args*/
	mysql_host = argv[2];
	mysql_user = argv[3];
	mysql_passwd = argv[4];
	mysql_id = (u_int)(atoi(argv[5]));
	current_generation = (u_int)(atoi(argv[6]));
	generations = (u_int)(atoi(argv[7]));

	if( strcasecmp(argv[8], "client") == 0 )
		play_mode = CLIENT;
	else if ( strcasecmp(argv[8], "server") == 0)
		play_mode = SERVER;
	else{printf("{{{Bad play_mode.}}}\n\n"); Replay_evolving_GPF_usage(); }
	host = argv[9];
	port = argv[10];
	sport = argv[11];
	if( strcasecmp(sport, "?") == 0 )
		sport = NULL;
	if ( strcasecmp(argv[12], "UDP") == 0)
		protocol = SOCK_DGRAM;
	else if ( strcasecmp(argv[12], "TCP") == 0 )
		protocol = SOCK_STREAM;
	else{printf("{{{Bad protocol}}}\n\n"); Replay_evolving_GPF_usage(); }
	delay = (u_int)(atoi(argv[13]));
	wait = (u_int)(atoi(argv[14]));

	if( strcasecmp(argv[15], "low") == 0 )
		display_level = LOW;
	else if ( strcasecmp(argv[15], "med") == 0)
		display_level = MED;
	else if ( strcasecmp(argv[15], "high") == 0)
		display_level = HIGH;
	else{printf("{{{Bad display_level.}}}\n\n"); Replay_evolving_GPF_usage(); }
	if( strcasecmp(argv[16], "ascii") == 0 )
		print_choice = TEXT;
	else if ( strcasecmp(argv[16], "hex") == 0)
		print_choice = HEX;
	else if ( strcasecmp(argv[16], "auto") == 0)
		print_choice = AUTO;
	else{printf("{{{Bad print_choice.}}}\n\n"); Replay_evolving_GPF_usage(); }


	if( (strcasecmp("none", argv[17]))==0)
		user_func=0;
	else
	{
		user_func=1;
		user_func_name = _malloc(strlen(argv[17]) );
		strncpy(user_func_name, argv[17], strlen(argv[17]) );
	}

	if( (strcasecmp("no", argv[18]))==0)
                proxy=NO;
        else if( (strcasecmp("yes", argv[18]))==0)
                proxy=YES;


	/*start the connection to the mysql server*/
	if( (mysql = mysql_init(mysql)) == NULL)
	{
		printf("In evolve_mode, mysql_init returned an error!\n");
		return ERROR;
	}
	if( (mysql = mysql_real_connect(mysql, mysql_host, mysql_user, mysql_passwd, "gpf", 0, NULL, 0)) == NULL)
	{
		printf("In evolve_mode, mysql_real_connect returned an error!\n");
		return ERROR;
	}
	if( (err = mysql_select_db(mysql, "gpf") != 0))
	{
		printf("In evolve_mode, mysql_select_db returned an error!\n");
		if (mysql != NULL) { printf("Error %u (%s)\n", mysql_errno(mysql), mysql_error(mysql)); }
		return ERROR;
	}
	if (display_level == MED || display_level == HIGH) {printf("\nMysql connection successful.\n");}

	start = time(NULL);
	gen_count = 0;
	session_count = 0;
	for(; current_generation < generations; current_generation++)
	{
		printf("Peforming a restore with ID=%d, generation=%d\n", mysql_id, current_generation);
		if( (pools = restore(mysql, mysql_id, current_generation, proxy)) == NULL)
		{
			printf("Failure to restore, perhaps the id or generation was bad?\n");
			return ERROR;
		}
		if (display_level==HIGH)
			print_pools(pools);

		tmp_pool = pools->first_pool;
		printf("\n\nPlaying Generation %d:\n\n", current_generation);
		while ( tmp_pool != NULL )
		{
			tmp_session = tmp_pool->first_session;
			printf("\n\nPlaying Pool %d:\n\n", tmp_pool->position);
			while ( tmp_session != NULL )  //on average, what's the total time for one session
			{
				if ( play_session(tmp_session, play_mode, host, port, sport, protocol, delay,
                				wait, HIGH, print_choice, user_func, user_func_name, NULL, NULL, proxy) == ERROR)
				{
					printf("Session play error.\n");
					tmp_session->error = YES;
				}
				tmp_session = tmp_session->next;
				session_count++;
			}
			tmp_pool = tmp_pool->next;
		}

		printf("Successfully played generation %d.\n", current_generation);
		now = time(NULL);
		printf("Done processing. Time to play: %d total sessions in %d seconds.\n\n",session_count, now - start);
		gen_count++;
	}

	printf("\nReplayed %d generations successfully.\n", gen_count);
	mysql_close(mysql);

	return OK;	
}
