/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

/*Function to store flags*/
int FLAG(int fd, int * flags)
{
	if (( (*flags) = fcntl(fd, F_GETFL)) == -1) {
		TIME(); fprintf(stderr, "{{{fcntl(F_GETFL)}}}"); return(-1); }
	
}

/*Function to change a file discriptor to non-blocking*/
int NONBLOCKING(int fd, int * flags)
{
	if (fcntl(fd, F_SETFL, (*flags)|O_NONBLOCK ) == -1){
		TIME(); fprintf(stderr, "{{{fcntl(F_SETL, 0_NONBLOCK)}}}\n"); return(-1); }
}

int _write(int fd, u_char * data, u_int size, int * flags, int * UDP_SERV, void * udpaddr)
{
	int i, x; 

	NONBLOCKING(fd, flags);
	for(i=0; i<2; i++)
	{
		if( *(UDP_SERV) == 1)
			x=sendto(fd, data, size, 0, (struct sockaddr *) udpaddr, sizeof(*udpaddr) );
		else
			x=write(fd, data, size);
		if (x == -1 || x == 0 )
			return x;
		else if (x == size)
			return x;
		printf("~~waiting to write data~~\n");
		sleep(1);
	}

	printf("~~~~Failed to write, moving on. %d sent.~~~~\n", x);
	return x;
}

int _read(int fd, int * flags, int * UDP_SERV, int * UDP_SERV_READ, int * TFTP, void ** Data, void ** readData,
	u_int * readDataSize, u_int index, u_int wait, void * udpaddr, u_int printChoice)
{
	u_int equal=0; //0 means the received and stored were the same
	int x=0;
	int y=0;
	int retval=0;
	socklen_t structLen = sizeof(*udpaddr);

	if( (*UDP_SERV == 1) && (*UDP_SERV_READ == 0) )
	{
		if (*(TFTP) > 0)
		{
			NONBLOCKING(fd, flags);
			while( (x=recvfrom(fd, readData[index], 1500, 0, (struct sockaddr *) udpaddr, &structLen)) !=0)
			{
				y++;
				if(x > 0)
					break;
				usleep(1);
				if ( y > (13+(wait)) )
				{
					printf("~~~~moving on without that data~~~~\n");
					fflush(stdout);
					break;
				}
				if (y>13)
				{
					printf("~~waiting to read data~~\n");
					sleep(1);
				}
			}
		}
		else
		{
			//should we explicitly set this to blocking, since that's the action we want?
			x=recvfrom(fd, readData[index], 1500, 0, (struct sockaddr *) udpaddr, &structLen);
		}

		*(UDP_SERV_READ)=1;
	}
	else
	{
		NONBLOCKING(fd, flags);
		while( (x=read(fd, readData[index], 1500)) != 0)
		{
			y++;
			if(x > 0)
				break;
			usleep(1);
			if ( y > (13+(wait)) )
			{
				printf("~~~~moving on without that data~~~~\n");
				fflush(stdout);
				break;
			}
			if (y>13)
			{
				printf("~~waiting to read data~~\n");
				sleep(1);
			}
		}
	}
	if( (x < 1) || (x > 1500) )
	{
		if(errno == EAGAIN)
			retval=-2;  ///Server didn't send data, but socket is still ok
		else
			retval=-1;
	}
	else
	{	
		readDataSize[index]=x;
		if(Data[index] !=0)
		{
			equal=memcmp(Data[index], readData[index], readDataSize[index] );
			if(equal==0)
				printf("  [<--][%d] ", index);
			else
				printf("  [<-DIFF-][%d] ", index);
		}
		else
			printf("  [<-NoDataCMP-]\n");
		_print(readDataSize[index], printChoice, readData[index]);
			printf("\n");
	}
	return retval;
}
