/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110ERROR301  USA
*/

#include "headers.h"

int setup_client_srcport(struct addrinfo *server_info, int srcport)
{
	int sockfd,i,b;

	/*Using the (possibly) multiple addrinfos, I now attempt to create and connect to a the
	server socket.  Each socket is given 7 trys before moving on to the next socket.  The first
	succesful connection is returned.
	*/	
	b=0;	
	do {
		sockfd = socket(server_info->ai_family, server_info->ai_socktype, server_info->ai_protocol);
		int on = 1;
		int len = sizeof(on);
		if( (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &on, len)) == ERROR)
		{
			TIME(); perror("{{{setsockopt}}}\n"); return(ERROR);
		}

		//This bit of code binds the socket to a particular source port. This was need for tftp.
		//But adversly affects -R option. I'll add an option to use srcport or not from cmdline.
		struct sockaddr_in tmp;
		int structLen=sizeof(tmp);
		bzero(&tmp, structLen);
		tmp.sin_family=AF_INET;
		tmp.sin_port=htons(srcport);
		if( (bind(sockfd, (struct sockaddr *) &tmp, structLen)) == ERROR)
		{
			TIME(); perror("{{{bind failed}}}\n"); return(ERROR);
		}

		if (sockfd<0)
			continue;

		for(i=0;i<7;i++)
		{
			if(connect(sockfd, server_info->ai_addr, server_info->ai_addrlen) == 0)
			{
				//printf ("Client successfully contacted Server");
				b=1;
				break;
			}
			else
			{	
				perror("connect");
				printf ("Failed to setup client\n");
				printf ("Let's sleep for 2 second and try again....\n");
				sleep(2);
			}
			if(i==6)
			{
				printf ("Giving up!!\n");
				return ERROR;
			}
		}
		if(b=1)
			break;
	
		if( (close(sockfd) < 0))
		{
			perror("close");
		}

	} while ( (server_info = server_info->ai_next) != NULL);
	
	if (server_info == NULL) {
		printf("CLIENT SETUP ERROR!!!"); return (ERROR);
	}

	return (sockfd);
}

//use this one for the -R option becuase it's faster, unless you really need to choose a srcport
int setup_client(struct addrinfo *server_info)
{
	int sockfd,i,b;

	/*Using the (possibly) multiple addrinfos, I now attempt to create and connect to a the
	server socket.  Each socket is given 7 trys before moving on to the next socket.  The first
	succesful connection is returned.
	*/	
	b=0;	
	do {
		sockfd = socket(server_info->ai_family, server_info->ai_socktype, server_info->ai_protocol);
		int on = 1;
		int len = sizeof(on);

		if (sockfd<0)
			continue;

		for(i=0;i<7;i++)
		{
			if(connect(sockfd, server_info->ai_addr, server_info->ai_addrlen) == 0)
			{
				//printf ("Client successfully contacted Server");
				b=1;
				break;
			}
			else
			{	
				perror("connect");
				printf ("Failed to setup client\n");
				printf ("Let's sleep for 2 second and try again....\n");
				sleep(2);
			}
			if(i==6)
			{
				printf ("Giving up!!\n");
				return ERROR;
			}
		}
		if(b=1)
			break;
	
		if( (close(sockfd) < 0))
		{
			perror("close");
		}

	} while ( (server_info = server_info->ai_next) != NULL);
	
	if (server_info == NULL) {
		printf("CLIENT SETUP ERROR!!!"); return (ERROR);
	}

	return (sockfd);
}

int setup_server(struct addrinfo *server_address)
{

	int listenfd;

	do {
	/*Using the (possibly) multiple addrinfos, I now attempt to create, listen, and bind to 
	server socket.  The first succesful connection is returned.
	*/	
		listenfd = socket(server_address->ai_family, server_address->ai_socktype, server_address->ai_protocol);
		int on = 1;
		int len = sizeof(on);
		if( (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &on, len)) == ERROR)
		{
			perror("setsockopt");
			return(ERROR);
		}
		if (listenfd<0)
			continue;
		if (bind(listenfd, server_address->ai_addr, server_address->ai_addrlen) == 0)
			break;

		if( (close(listenfd) < 0))
		{
			perror("close listenfd");
		}
	} while ( (server_address = server_address->ai_next) != NULL);

	if(server_address == NULL) {	
		printf("SERVER SETUP ERROR!!!"); return (ERROR);
	}
	
	if( server_address->ai_socktype==SOCK_STREAM )
	{
		if( (listen(listenfd, LISTENQ)) < 0) {
			perror("listen"); return (ERROR);
		}
		return (listenfd);
	}
	else
	{
		return (listenfd);
	}
}
