/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

//1=Yes. 0=No.
int Is_Ascii(struct Tok * tok)
{
	if (tok == NULL)
                return ERROR;

        int i;
        u_char c;

        u_int ret=1;

        for(i=0; i<tok->dataLen; i++)
        {
                c=(u_char)*(tok->data+i);
		if( c=='\t' || c=='\r' || c=='\n' )
			continue;
		else
		{
			if ( (c > (u_char)0x7e) )
			{
				ret=0;
				break;
			}
			if ( (c < (u_char)0x20) )
			{
				ret=0;
				break;
			}
		}
        }

        return ret;
}

//1=Yes. 0=No.
int char_is_ascii(char c)
{
        int i;

        u_int ret=1;

	if( c=='\t' || c=='\r' || c=='\n' )
		return ret;
	else
	{
		if ( (c > (u_char)0x7e) )
		{
			ret=0;
			return ret;
		}
		if ( (c < (u_char)0x20) )
		{
			ret=0;
			return ret;
		}
	}

        return ret;
}

//1=Yes. 0=No.
//will always return No if either is zero.
int same_aprox_size(int a, int b)
{
	int x;

	//for debugging
	//printf("a=%d, b=%d\n", a, b);
	//x = a - b;
	//printf("x=%d, abs(x)=%d\n", x, abs(x));
	//
	
	if( a==0 || b==0)
		return 0;

	//allow small strings, but the must be exact
	if( a==1 && b==1)
		return 1;
	if( a==2 && b==2)
		return 1;
	if( a==3 && b==3)
		return 1;
	if( a==4 && b==4)
		return 1;
	if( a<5 || b<5)
		return 0;

	//anything 5-8 must be only one off
	x = a - b;
	if( a < 9 || b < 9)
	{
		if( abs(x) < 2)
			return 1;
		else
			return 0;
	}

	//anything else can be off by two
	if( abs(x) < 3)
		return 1;
	else
		return 0;
}

/*Should be called multiple times, as it only reports the first found ascii*/
/*If a valid ascii struct is retuned it needs to be free'ed by the caller*/
//ret YES, NO, ERROR
int ascii_locator(char * data, int len, struct ascii * ascii)
{
        int i=0;
        int j=0;
        int _len=0;
        int _LEN=0;
        int head=0;
	int _len_size=0;
	int first_time = 1;
	int found = 0;
        char c=0;
	char * start=NULL;

	if (data==NULL || len<0)
                return ERROR;

        for(i=0; i<len; i++)
        {
                c=*(data+i);
		if( char_is_ascii(c) )
		{
			if(first_time == 1)
			{
				start=data+i;
				first_time = 0;
			}
			if( i == len-1)
			{
				i++;
				goto END;
			}
		}
		else
		{
			END:
			if( start != NULL)
			{
				//suspected _len of ascii
				_len = (data+i) - start;
				//bytes from head of data
				head = start - data;
				if( head == 0)
				{
				}
				else if ( head == 1)
				{
					_len_size = 1;
					_LEN = _ndata_to_size8(start - _len_size);
					if( same_aprox_size(_len, _LEN) )
					{
						found = 1;
						break;
					}
				}
				else if (head==2 || head==3)
				{
					_len_size = 1;
					_LEN = _ndata_to_size8(start - _len_size);
					if( same_aprox_size(_len, _LEN) )
					{
						found = 1;
						break;
					}
					_len_size = 2;
					_LEN = _ndata_to_size16(start - _len_size);
					if( same_aprox_size(_len, _LEN) )
					{
						found = 1;
						break;
					}
				}
				else if (head > 3)
				{
					_len_size = 1;
					_LEN = _ndata_to_size8(start - _len_size);
					if( same_aprox_size(_len, _LEN) )
					{
						found = 1;
						break;
					}
					_len_size = 2;
					_LEN = _ndata_to_size16(start - _len_size);
					if( same_aprox_size(_len, _LEN) )
					{
						found = 1;
						break;
					}
					_len_size = 4;
					_LEN = _ndata_to_size32(start - _len_size);
					if( same_aprox_size(_len, _LEN) )
					{
						found = 1;
						break;
					}
				}
				else
				{
					printf("In ascii_locator, bad head count!\n");
					return ERROR;
				}

				first_time = 1;
				start = NULL;
			} //if
		} //else
        } //for

	if( found )
	{
		ascii->size_of_len = _len_size;
		ascii->start_of_len = start - _len_size;
		ascii->len = _LEN;
		ascii->data = start;
			
		return YES;
	}
	else
		return NO;
}

