/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "headers.h"

/*to aid in the tokenization of a a string based protocol capture for the -P option*/
struct Tok * NORMAL_ASCII(struct Tok * tok, u_char * data, u_int size, u_int leg)
{

	//error checking
        if( data == NULL ) {
                TIME(); fprintf(stderr, "{{{NULL data passed to normal_ascii tokAid}}}\n"); exit(-1);
        }
        if( (size < 0) || (size > 999999999) ) {
                TIME(); fprintf(stderr, "{{{Bad size passed to normal_ascii tokAid: %d}}}\n", size); exit(-1);
        }
        if( (leg < 0) || (leg > 3000) ) {
                TIME(); fprintf(stderr, "{{{Bad leg passed to normal_ascii tokAid}}}\n"); exit(-1);
        }
	//local vars
	u_char * ptr=NULL;
	u_char * ptr2=NULL;
	
	int first=1;
	int i=1;

	do
	{
		//if first that treat as cmd
		tok=Create_Next_Tok(tok, leg);
		if(first) {
			tok->type=ASCII_CMD;
			first=0;
		}
		else
			tok->type=ASCII_CMDVAR;
		int space=0;
		int crlf=0;
		if ( (ptr=memchr(data+(tok->currentTotal), ' ', size - tok->currentTotal)) != NULL)
		{
			space=1;
			for(i=1; i<(size -1 -tok->currentTotal); i++)
			{
				if( *(ptr+i) == ' ')
					space++;
				else
					break;
			}
		}
		if ( (ptr2=memchr(data+(tok->currentTotal), '\r', size - tok->currentTotal)) != NULL)
		{
			if ( *(ptr2+1) == '\n')
				crlf=1;
		}
		//found both
		if( (space > 0) && (crlf == 1) )
		{
			//since we found both with ptr is lesser in value
			if( ptr < ptr2 ) //the space came first
			{
				//grab cmd
				tok->dataLen=ptr - (data+(tok->currentTotal));
				if(tok->dataLen == 0)
				{
					tok->type=ASCII_SPACE;
					tok->dataLen=space;
					Slurp_Into_Tok(tok, data);
				}
				else
				{
					Slurp_Into_Tok(tok, data);
					//grab space
					tok=Create_Next_Tok(tok, leg);
					tok->type=ASCII_SPACE;
					tok->dataLen=space;
					Slurp_Into_Tok(tok, data);
				}
			}
			else //the crlf came first.  an odd case, but can happen validly.
			{
				//grab cmd
				tok->dataLen=ptr2 - (data+(tok->currentTotal));
				if(tok->dataLen == 0)
				{
					tok->type=ASCII_END;
					tok->dataLen=2;
					Slurp_Into_Tok(tok, data);
				}
				else
				{
					Slurp_Into_Tok(tok, data);
					//grab \r\n
					tok=Create_Next_Tok(tok, leg);
					tok->type=ASCII_END;
					tok->dataLen=2;
					Slurp_Into_Tok(tok, data);
					first=1;
				}
			}
		}	
		//found just clfr
		else if( (space == 0) && (crlf == 1) )
		{
			//grab cmd
			tok->dataLen=ptr2 - (data+(tok->currentTotal));
			if(tok->dataLen == 0)
			{
				tok->type=ASCII_END;
				tok->dataLen=2;
				Slurp_Into_Tok(tok, data);
			}
			else
			{
				Slurp_Into_Tok(tok, data);
				//grab \r\n
				tok=Create_Next_Tok(tok, leg);
				tok->type=ASCII_END;
				tok->dataLen=2;
				Slurp_Into_Tok(tok, data);
				first=1;
			}
		}	
		//found just a space
		else if( (space > 0) && (crlf == 0) )
		{
			//grab cmd
			tok->dataLen=ptr - (data+(tok->currentTotal));
			if(tok->dataLen == 0)
			{
				tok->type=ASCII_SPACE;
				tok->dataLen=space;
				Slurp_Into_Tok(tok, data);
			}
			else
			{
				Slurp_Into_Tok(tok, data);
				//grab space
				tok=Create_Next_Tok(tok, leg);
				tok->type=ASCII_SPACE;
				tok->dataLen=space;
				Slurp_Into_Tok(tok, data);
			}
       		}
		//found no space or \r\n so just grab cmd
		else if( (space == 0) && (crlf == 0) )
		{
			//grab cmd
			tok->dataLen=size - tok->currentTotal;
			if(tok->dataLen == 0) {
				if( (tok=Destroy_Toks(tok, "current")) == NULL){
                			TIME(); fprintf(stderr, "{{{problem with destroy in normal ascii}}}\n"); exit(-1);
				}
				break;
			}
			else		
				Slurp_Into_Tok(tok, data);
		}
	} while (tok->currentTotal < size);

	//roll back to first tok and add extra attributes if there are any
	if ( (tok=Find_First_Tok(tok)) == NULL) {
	                TIME(); fprintf(stderr, "{{{Find_First_Tok error in normal_ascii}}}\n"); exit(-1);
       	}
	while(tok->next !=NULL)
	{
                //If so, does the string contain ascii numbers? (0x30-0x39)
                if(Is_Numeric(tok))
                        tok->numeric=1;
                //does it contain bracket?
                if(Is_Bracket(tok))
                        tok->bracket=1;
                //does it contain quote?
                if(Is_Quote(tok))
                        tok->quote=1;
                //does it contain special?
                if(Is_Special(tok))
                        tok->special=1;
                //does it others
                if(Is_Other(tok))
                        tok->other=1;
		tok=tok->next;
	}
	return tok;
}
