/*  Createright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include "headers.h"

/*to aid in the tokenization of a dns capture for the -P option*/
struct Tok * ZONE_TRANSFER(struct Tok * tok, u_char * data, u_int size, u_int leg)
{
	//error checking
        if( data == NULL ) {
                TIME(); fprintf(stderr, "{{{NULL data passed to DNS tokAid}}}\n"); exit(-1);
        }
        if( (size < 0) || (size > 1500) ) {
                TIME(); fprintf(stderr, "{{{Bad size passed to DNS tokAid}}}\n"); exit(-1);
        }
        if( (leg < 0) || (leg > 3000) ) {
                TIME(); fprintf(stderr, "{{{Bad leg passed to DNS tokAid}}}\n"); exit(-1);
        }
	//local vars
	u_int16_t qd=0;
	u_int16_t an=0;
	u_int16_t ns=0;
	u_int16_t ar=0;
	//master len
	tok=Create_Next_Tok(tok, leg);
	tok->type=LEN;
	tok->covered=0;
	tok->dataLen=2;
	Slurp_Into_Tok(tok, data);
	//id
	tok=Create_Next_Tok(tok, leg);
	tok->type=SESSION_ID;
	tok->dataLen=2;
	Slurp_Into_Tok(tok, data);
	//flags
	tok=Create_Next_Tok(tok, leg);
	tok->type=FLAGS;
	tok->dataLen=2;
	Slurp_Into_Tok(tok, data);
	//qd count
	tok=Create_Next_Tok(tok, leg);
	tok->type=PROTO_ID;
	tok->dataLen=2;
	Slurp_Into_Tok(tok, data);
	qd=_ndata_to_size16(tok->data);
	//an count
	tok=Create_Next_Tok(tok, leg);
	tok->type=PROTO_ID;
	tok->dataLen=2;
	Slurp_Into_Tok(tok, data);
	an=_ndata_to_size16(tok->data);
	//ns count
	tok=Create_Next_Tok(tok, leg);
	tok->type=PROTO_ID;
	tok->dataLen=2;
	Slurp_Into_Tok(tok, data);
	ns=_ndata_to_size16(tok->data);
	//ar count
	tok=Create_Next_Tok(tok, leg);
	tok->type=PROTO_ID;
	tok->dataLen=2;
	Slurp_Into_Tok(tok, data);
	ar=_ndata_to_size16(tok->data);
	//parse queries
	while(qd > 0)
	{
		//Figure out len
		tok=Create_Next_Tok(tok, leg);
		tok->type=LEN;
		tok->covered=0;
		tok->dataLen=1;
		Slurp_Into_Tok(tok, data);
		//Then the next token will be the string that len is associated with
		tok=Create_Next_Tok(tok, leg);
		tok->type=ASCII;
		tok->dataLen=_ndata_to_size8(tok->prev->data);
		Slurp_Into_Tok(tok, data);
		//check to see if we're at the end of the dns name
		if ( *(data+(tok->currentTotal)) == 0x00)
		{
			//the null is it's own token
			tok=Create_Next_Tok(tok, leg);
			tok->type=BINARY_END;
			tok->dataLen=1;
			Slurp_Into_Tok(tok, data);
			//Figure out len
			tok=Create_Next_Tok(tok, leg);
			tok->type=BINARY;
			tok->dataLen=2;
			Slurp_Into_Tok(tok, data);
			//Figure out len
			tok=Create_Next_Tok(tok, leg);
			tok->type=BINARY;
			tok->dataLen=2;
			Slurp_Into_Tok(tok, data);
			qd--;
		}
	}
	//parse answers, authoritative nameservers, and additional records
	while ( (an > 0) || (ns > 0) || (ar > 0) )
	{
		//name
		tok=Create_Next_Tok(tok, leg);
		tok->type=BINARY;
		tok->dataLen=2;
		Slurp_Into_Tok(tok, data);
		//type
		tok=Create_Next_Tok(tok, leg);
		tok->type=BINARY;
		tok->dataLen=2;
		Slurp_Into_Tok(tok, data);
		//class
		tok=Create_Next_Tok(tok, leg);
		tok->type=BINARY;
		tok->dataLen=2;
		Slurp_Into_Tok(tok, data);
		//time to live
		tok=Create_Next_Tok(tok, leg);
		tok->type=BINARY;
		tok->dataLen=4;
		Slurp_Into_Tok(tok, data);
		//len
		tok=Create_Next_Tok(tok, leg);
		tok->type=LEN;
		tok->covered=1;
		tok->dataLen=2;
		Slurp_Into_Tok(tok, data);
		//Then the next token will be the string that len is associated with
		tok=Create_Next_Tok(tok, leg);
		tok->type=MIXED;
		tok->dataLen=_ndata_to_size16(tok->prev->data);
		Slurp_Into_Tok(tok, data);
		if (an > 0)
			an--;
		else if (ns > 0)
			ns--;
		else if (ar > 0)
			ar--;
	}

	u_int covered = tok->tokNum;
	struct Tok * tmp_tok = Find_First_Tok(tok);
	tmp_tok->covered = covered - 1;
	
	return tok;
}
