/*  Copyright (c) 2005, Applied Security, Inc.
    Jared DeMott, jdemott@appliedsec.com, www.appliedsec.com

    This file is part of GPF.

    GPF is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GPF is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GPF; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "headers.h"

/*If you pass in a NULL ptr than this is the first tok of the double linked list*/
//NULL=error
//good ptr = ok
struct Tok * Create_Next_Tok(struct Tok * tok, u_int leg)
{
	//error check
	if( (leg < 0) || (leg > 500000) ) {
        	TIME(); fprintf(stderr, "{{{Invalid leg passed to Create_Next_Tok}}}\n"); return(NULL);
        }
	
        struct Tok * new_tok;
        new_tok=_malloc(sizeof(struct Tok));
	if(new_tok == NULL) {
        	TIME(); fprintf(stderr, "{{{Bad new new_tok in Create_Next_Tok}}}\n"); return(NULL);
        }
        
	new_tok->dataLen=-1;
        new_tok->data=NULL;
        new_tok->type=UNDEFINED;
        new_tok->numeric=-1;
        new_tok->bracket=-1;
        new_tok->quote=-1;
        new_tok->special=-1;
        new_tok->other=-1;
        new_tok->covered=-1;
        new_tok->leg=leg;
        new_tok->next=NULL;

	if(tok == NULL)
	{
        	new_tok->prev=NULL;
		new_tok->tokNum=0;
        	new_tok->currentTotal=0;
	}
	else
	{
		//set linked list ptrs
        	new_tok->prev=tok;
		tok->next=new_tok;
		//updated tokNum if we're on the same leg, 0 if new leg.
		if(tok->leg == new_tok->leg)
			new_tok->tokNum=(tok->tokNum)+1;
		else
        		new_tok->tokNum=0;
		//current total should be the one from before, if we're in the same leg.  If new leg, start back at 0.
		if(tok->leg == new_tok->leg)
        		new_tok->currentTotal=tok->currentTotal;
		else
        		new_tok->currentTotal=0;
	}

	if(new_tok == NULL) {
        	TIME(); fprintf(stderr, "{{{Bad new_tok in Create_Next_Tok}}}\n"); return(NULL);
        }
	return new_tok;
}

/*was added for evolve code*/
int type_initialize_tok(struct Tok * tok, int remaining_toks, struct seed * seeds, char * type)
{
	/*error check*/
	if(tok == NULL) {
        	TIME(); fprintf(stderr, "{{{Null tok passed to type_initialize_tok}}}\n"); return(ERROR);
	}

	int x, i;

	if( type == NULL)
	{
		printf("Null type, returning.\r");
		return 0;
	}
	// not sure why this is here?
	//else if( (int)type < 17 )
	//{
	//	tok->type = (int)type;
       	// 	tok->dataLen = 0x00000000;
	//}
	else
	{
		tok->type = *(type);	
        	tok->dataLen = *((u_int32_t*)(type+1));
	}

	//additionaly processing on the len
	if( tok->dataLen == 0x00000000 )
	{
		x = _rand(8);
		switch (x)
		{
			case 0:
				tok->dataLen = _rand(16) + 1;
				break;
			case 1:
				tok->dataLen = _rand(32) + 1;
				break;
			case 2:
				tok->dataLen = _rand(64) + 1;
				break;
			case 3:
				tok->dataLen = _rand(128) + 1;
				break;
			case 4:
				tok->dataLen = _rand(256) + 1;
				break;
			case 5:
				tok->dataLen = _rand(512) + 1;
				break;
			case 6:
				tok->dataLen = _rand(2048) + 1;
				break;
			case 7:
				tok->dataLen = _rand(36356) + 1;
				break;
		}
	}
	
	//get data from seeds sometimes if they exits
	//a little cluggy ...
	if( seeds!=NULL && (_rand(100) > 65) ) //use a seed
	{
		x = _rand(seeds->total_seeds);

		while(seeds != NULL)
		{
			if( seeds->position == x )
			{
				tok->data = _malloc(tok->dataLen);
				if(seeds->data_len < tok->dataLen)
					memcpy(tok->data, seeds->data, seeds->data_len);
				else
					memcpy(tok->data, seeds->data, tok->dataLen);
				break;
			}

			seeds = seeds->next;
		}
	}
	//randomly create data

	else
	{
		/*fill the data with totally random bytes*/
		if( tok->type==LEN || tok->type==HASH || tok->type==MIXED || tok->type==BINARY || tok->type==SESSION_ID || tok->type==FLAGS || tok->type==PROTO_ID)
		{
			tok->data = _malloc(tok->dataLen);
			for(i=0; i < tok->dataLen; i++)
			{
				tok->data[i] = (u_char)_rand(256);
			}
		}
		if (tok->type==STRING || tok->type==ASCII)
		{
			tok->data = _malloc(tok->dataLen);
			for(i=0; i<(tok->dataLen); i++)
			{
				tok->data[i] = (u_char)_rand(119) + 9;
			}
		}
		//relook at sometime
		if (tok->type == WIDESTRING_2)
		{
			tok->data = _malloc(tok->dataLen);
			for(i=0; i<(tok->dataLen); i=i+2)
			{
				tok->data[i] = 0x00;
				tok->data[i+1] = (u_char)_rand(119) + 9;
			}
		}
		//relook at sometime
		if (tok->type == WIDESTRING_4)
		{
			tok->data = _malloc(tok->dataLen);
			for(i=0; i<(tok->dataLen); i=i+4)
			{
				tok->data[i] = 0x00;
				tok->data[i+1] = 0x00;
				tok->data[i+2] = 0x00;
				tok->data[i+3] = (u_char)_rand(119) + 9;
			}
		}
		//right now these are almost the same, but I could make this one more intelligent, like picking from a dic?
		if(tok->type==ASCII_CMD || tok->type==ASCII_CMDVAR)
		{
			tok->data = _malloc(tok->dataLen);
			for(i=0; i < tok->dataLen; i++)
			{
				tok->data[i] = (u_char)_rand(105) + 21;
			}
		}
	
		/*set a data size and fill these guys with data*/
		if (tok->type == ASCII_SPACE)
		{
			tok->dataLen = _rand(5)+1;
			tok->data = _malloc(tok->dataLen);
			for(i=0; i<(tok->dataLen); i++)
       	 	{
				x = _rand(4);
				switch (x)
				{
					case 0:
						tok->data[i]=' ';
						break;
					case 1:
						tok->data[i]='\t';
						break;
					case 2:
						tok->data[i]='\r';
						break;
					case 3:
						tok->data[i]=' ';
						break;
				}
			}       
		}
	
		if (tok->type == ASCII_END)
		{
			x = _rand(3);
			switch (x)
			{
				case 0:
					tok->dataLen = 1;
					tok->data = _malloc(tok->dataLen);
					memcpy(tok->data, "\n", 1);
					break;
				case 1:
					tok->dataLen = 2;
					tok->data = _malloc(tok->dataLen);
					memcpy(tok->data, "\r\n", 2);
					break;
				case 2:
					tok->dataLen = 4;
					tok->data = _malloc(tok->dataLen);
					memcpy(tok->data, "\r\n\r\n", 4);
					break;
			}
		}
	
		if (tok->type == BINARY_END)
		{
			x = _rand(3);
			switch (x)
			{
				case 0:
					tok->dataLen = 1;
					tok->data = _malloc(tok->dataLen);
					memcpy(tok->data, "\x00", 1);
					break;
				case 1:
					tok->dataLen = 2;
					tok->data = _malloc(tok->dataLen);
					memcpy(tok->data, "\x00\x00", 2);
					break;
				case 2:
					tok->dataLen = 4;
					tok->data = _malloc(tok->dataLen);
					for(i=0; i<(tok->dataLen); i++)
       	 			{
						tok->data[i]= (u_char)_rand(256);
					}       
					break;
			}
		}
	
	  } //we went random

	
	/*if it's a LEN or HASH, how many of the following toks will it cover*/
	if( tok->type == LEN || tok->type == HASH)
	{
		//tok->covered = _rand(remaining_toks) + 1;
		tok->covered = 1; //hard coded for proxy
	}

	/*now that we've built data, find out what it is*/
	//If so, does the string contain ascii numbers? (0x30-0x39)
	if(Is_Numeric(tok))
	        tok->numeric=1;
	//does it contain bracket?
	if(Is_Bracket(tok))
	        tok->bracket=1;
	//does it contain quote?
	if(Is_Quote(tok))
	        tok->quote=1;
	//does it contain special?
	if(Is_Special(tok))
	        tok->special=1;
	//does it contain others
	if(Is_Other(tok))
	        tok->other=1;

	if( tok->dataLen < 1)
	{
		printf("In randomly_init_toks, bad dataLen!\n"); 
		Print_Toks(tok, "all", "long");
	}
	
	return OK;
}

/*was added for evolve code*/
int randomly_initialize_tok(struct Tok * tok, int remaining_toks, struct seed * seeds)
{
	/*error check*/
	if(tok == NULL) {
        	TIME(); fprintf(stderr, "{{{Null tok passed to randomly_initialize_tok}}}\n"); return(ERROR);
        }

	/*vars*/
	char x;
	int i;

  if( seeds!=NULL && (_rand(100) > 65) ) //use a seed
  {
	x = _rand(seeds->total_seeds);

	while(seeds != NULL)
	{
		if( seeds->position == x )
		{
			tok->type = seeds->type;
			tok->dataLen = seeds->data_len;
			tok->data = _malloc(seeds->data_len);
			memcpy(tok->data, seeds->data, seeds->data_len);
			break;
		}

		seeds = seeds->next;
	}
  }
  else
  {
	/*first pick type*/
	tok->type = _rand(14);  ////we're only using the first 13 types.  Add code for wide strings?

	/*now, query that type and create a dataLen accordingly*/
	if (tok->type == LEN)
	{
		x = _rand(5);
		switch (x)
		{
			case 0:
				tok->dataLen = 1;
				break;
			case 1:
				tok->dataLen = 2;
				break;
			case 2:
				tok->dataLen = 4;
				break;
			case 3:
				tok->dataLen = 8;
				break;
			case 4:
				tok->dataLen = 16;
				break;
		}
	}

	if (tok->type == HASH)
	{
		x = _rand(8);
		switch (x)
		{
			case 0:
				tok->dataLen = 16;
				break;
			case 1:
				tok->dataLen = 20;
				break;
			case 2:
				tok->dataLen = 24;
				break;
			case 3:
				tok->dataLen = 28;
				break;
			case 4:
				tok->dataLen = 32;
				break;
			case 5:
				tok->dataLen = 40;
				break;
			case 6:
				tok->dataLen = 48;
				break;
			case 7:
				tok->dataLen = 64;
				break;
		}
		/*add code here to pick a (sub-type) hash type*/
	}

	if (tok->type==MIXED || tok->type==BINARY || tok->type==STRING || tok->type==ASCII)
	{
		x = _rand(8);
		switch (x)
		{
			case 0:
				tok->dataLen = _rand(16) + 1;
				break;
			case 1:
				tok->dataLen = _rand(32) + 1;
				break;
			case 2:
				tok->dataLen = _rand(64) + 1;
				break;
			case 3:
				tok->dataLen = _rand(128) + 1;
				break;
			case 4:
				tok->dataLen = _rand(256) + 1;
				break;
			case 5:
				tok->dataLen = _rand(512) + 1;
				break;
			case 6:
				tok->dataLen = _rand(2048) + 1;
				break;
			case 7:
				tok->dataLen = _rand(36356) + 1;
				break;
		}
	}

	if (tok->type==SESSION_ID || tok->type==FLAGS || tok->type==PROTO_ID)
	{
		x = _rand(8);
		switch (x)
		{
			case 0:
				tok->dataLen = 1;
				break;
			case 1:
				tok->dataLen = 2;
				break;
			case 2:
				tok->dataLen = 4;
				break;
			case 3:
				tok->dataLen = 8;
				break;
			case 4:
				tok->dataLen = 10;
				break;
			case 5:
				tok->dataLen = 12;
				break;
			case 6:
				tok->dataLen = 16;
				break;
			case 7:
				tok->dataLen = 20;
				break;
		}
	}

	if(tok->type==ASCII_CMD || tok->type==ASCII_CMDVAR)
	{
		x = _rand(9);
		switch (x)
		{
			case 0:
				tok->dataLen = 1;
				break;
			case 1:
				tok->dataLen = 2;
				break;
			case 2:
				tok->dataLen = 4;
				break;
			case 3:
				tok->dataLen = 8;
				break;
			case 4:
				tok->dataLen = 16;
				break;
			case 5:
				tok->dataLen = _rand(16) + 1;
				break;
			case 6:
				tok->dataLen = _rand(32) + 2;
				break;
			case 7:
				tok->dataLen = _rand(64) + 4;
				break;
			case 8:
				tok->dataLen = _rand(128) + 8;
				break;
		}
	}

	/*fill the data with totally random bytes*/
	if( tok->type==LEN || tok->type==HASH || tok->type==MIXED || tok->type==BINARY || tok->type==SESSION_ID || tok->type==FLAGS || tok->type==PROTO_ID)
	{
		tok->data = _malloc(tok->dataLen);
		for(i=0; i < tok->dataLen; i++)
		{
			tok->data[i] = (u_char)_rand(256);
		}
	}

	if (tok->type==STRING || tok->type==ASCII)
	{
		tok->data = _malloc(tok->dataLen);
		for(i=0; i<(tok->dataLen); i++)
		{
			tok->data[i] = (u_char)_rand(119) + 9;
		}
	}

	//right now these are almost the same, but I could make this one more intelligent, like picking from a dic?
	if(tok->type==ASCII_CMD || tok->type==ASCII_CMDVAR)
	{
		tok->data = _malloc(tok->dataLen);
		for(i=0; i < tok->dataLen; i++)
		{
			tok->data[i] = (u_char)_rand(105) + 21;
		}
	}

	/*set a data size and fill these guys with data*/
	if (tok->type == ASCII_SPACE)
	{
		tok->dataLen = _rand(5)+1;
		tok->data = _malloc(tok->dataLen);
		for(i=0; i<(tok->dataLen); i++)
        	{
			x = _rand(4);
			switch (x)
			{
				case 0:
					tok->data[i]=' ';
					break;
				case 1:
					tok->data[i]='\t';
					break;
				case 2:
					tok->data[i]='\r';
					break;
				case 3:
					tok->data[i]=' ';
					break;
			}
		}       
	}

	if (tok->type == ASCII_END)
	{
		x = _rand(3);
		switch (x)
		{
			case 0:
				tok->dataLen = 1;
				tok->data = _malloc(tok->dataLen);
				memcpy(tok->data, "\n", 1);
				break;
			case 1:
				tok->dataLen = 2;
				tok->data = _malloc(tok->dataLen);
				memcpy(tok->data, "\r\n", 2);
				break;
			case 2:
				tok->dataLen = 4;
				tok->data = _malloc(tok->dataLen);
				memcpy(tok->data, "\r\n\r\n", 4);
				break;
		}
	}

	if (tok->type == BINARY_END)
	{
		x = _rand(3);
		switch (x)
		{
			case 0:
				tok->dataLen = 1;
				tok->data = _malloc(tok->dataLen);
				memcpy(tok->data, "\x00", 1);
				break;
			case 1:
				tok->dataLen = 2;
				tok->data = _malloc(tok->dataLen);
				memcpy(tok->data, "\x00\x00", 2);
				break;
			case 2:
				tok->dataLen = 4;
				tok->data = _malloc(tok->dataLen);
				for(i=0; i<(tok->dataLen); i++)
        			{
					tok->data[i]= (u_char)_rand(256);
				}       
				break;
		}
	}

  } //we went random

	/*if it's a LEN or HASH, how many of the following toks will it cover*/
	if( tok->type == LEN || tok->type == HASH)
	{
		tok->covered = _rand(remaining_toks) + 1;
	}

	/*now that we've built data, find out what it is*/
	//If so, does the string contain ascii numbers? (0x30-0x39)
        if(Is_Numeric(tok))
	        tok->numeric=1;
        //does it contain bracket?
        if(Is_Bracket(tok))
                tok->bracket=1;
        //does it contain quote?
        if(Is_Quote(tok))
                tok->quote=1;
        //does it contain special?
        if(Is_Special(tok))
                tok->special=1;
        //does it contain others
        if(Is_Other(tok))
                tok->other=1;

	if( tok->dataLen < 1)
	{
		printf("In randomly_init_toks, bad dataLen!\n"); 
		Print_Toks(tok, "all", "long");
	}

	return OK;
}

//num= "current" or "all"
//NULL=error
//1=ok
struct Tok * Destroy_Toks(struct Tok * tok, char * which)
{
	//error check
	if(tok == NULL) {
        	TIME(); fprintf(stderr, "{{{Null tok passed to Destroy_Tok}}}\n"); return(NULL);
        }
	struct Tok * temp_tok=NULL;

	if( strcasecmp(which, "current") == 0) 
	{
		//need to fix up (known as unlink) list if this is in the middle or on an end
		if( (tok->next == NULL) && (tok->prev == NULL) ) //simple case, only tok
		{
			temp_tok=(struct Tok *)(1);
			//just drop to bottom
		}
		else if( (tok->next == NULL) && (tok->prev != NULL) ) //last tok
		{
			temp_tok=tok->prev; //returns "2nd to last" tok
			temp_tok->next=NULL;
		}
		else if( (tok->next != NULL) && (tok->prev == NULL) ) //first tok
		{
			temp_tok=tok->next; //returns "2nd" tok
			temp_tok->prev=NULL; 
		}
		else if( (tok->next != NULL) && (tok->prev != NULL) ) //middle tok
		{
			temp_tok=tok->next;
			temp_tok->prev=tok->prev;
			temp_tok=tok->prev;  //returns the tok before the unlinked tok
			temp_tok->next=tok->next;
		}
		if(tok->data != NULL)
	       		free(tok->data);
		tok->data=NULL;
		free(tok);
		tok=NULL;
		return temp_tok;
	}
	else if( strcasecmp(which, "all") == 0) 
	{
		tok=Find_First_Tok(tok);
		while(tok != NULL)
		{
			temp_tok = tok->next;
			if(tok->data != NULL)
				free(tok->data);
			else
				printf("In Destroy_Toks, found NULL data\n");
			free(tok);
			tok = temp_tok;
		}
	}
	else{
        	TIME(); fprintf(stderr, "{{{Bad 'which' in Destroy_Tok}}}\n"); return(NULL);
        }
	return (struct Tok *)(1);	
}

//exact copy of one tok to another
//NULL=error
//good ptr = ok
struct Tok * Dup_Tok(struct Tok * tok)
{
	//error check
	if(tok == NULL) {
        	TIME(); fprintf(stderr, "{{{Null tok passed to Dup_Tok}}}\n"); return(NULL);
        }

        struct Tok * new_tok;
        new_tok=_malloc(sizeof(struct Tok));

        new_tok->dataLen=tok->dataLen;
        new_tok->data=_malloc(new_tok->dataLen);
        memcpy(new_tok->data, tok->data, new_tok->dataLen);
        new_tok->type=tok->type;
        new_tok->numeric=tok->numeric;
        new_tok->bracket=tok->bracket;
        new_tok->quote=tok->quote;
        new_tok->special=tok->special;
        new_tok->other=tok->other;
        new_tok->covered=tok->covered;
        new_tok->leg=tok->leg;
        new_tok->tokNum=tok->tokNum;
        new_tok->currentTotal=tok->currentTotal;
        new_tok->prev=tok->prev;
        new_tok->next=tok->next;

        return new_tok;
}

//transfer data from one tok to a tok in another list of toks
//NULL=error
//good ptr = ok
struct Tok * Transfer_Tok(struct Tok * tok, struct Tok *tok2)
{
	//error check
	if(tok == NULL) {
        	TIME(); fprintf(stderr, "{{{Null tok passed to Transfer_Tok}}}\n"); return(NULL);
        }

	if( tok2 == NULL )
	{
		tok2=Create_Next_Tok(tok2, tok->leg);
	        
		tok2->dataLen=tok->dataLen;
	        tok2->data=_malloc(tok2->dataLen);
	        memcpy(tok2->data, tok->data, tok2->dataLen);
	        tok2->type=tok->type;
		tok2->numeric=tok->numeric;
		tok2->bracket=tok->bracket;
	        tok2->quote=tok->quote;
	        tok2->special=tok->special;
	        tok2->other=tok->other;
	        tok2->covered=tok->covered;
		tok2->tokNum=tok->tokNum;
		//set to zero or keep wrong tally, becuase of fuzz funcs, or fix up total in fuzz funcs?
		//tok2->currentTotal=tok2->dataLen;
		tok2->currentTotal=0;
		
		return tok2;
	}
	else
	{
		//create a new tok and copy
		struct Tok * new_tok;
		new_tok=_malloc(sizeof(struct Tok));

	        new_tok->dataLen=tok->dataLen;
	        new_tok->data=_malloc(new_tok->dataLen);
	        memcpy(new_tok->data, tok->data, new_tok->dataLen);
	        new_tok->type=tok->type;
		new_tok->numeric=tok->numeric;
		new_tok->bracket=tok->bracket;
	        new_tok->quote=tok->quote;
	        new_tok->special=tok->special;
	        new_tok->other=tok->other;
	        new_tok->covered=tok->covered;
	        new_tok->leg=tok->leg;
		
		//set linked list ptrs
	       	new_tok->prev=tok2;
		new_tok->next=NULL;
		tok2->next=new_tok;
		//updated tokNum to whatever was??
		new_tok->tokNum=tok->tokNum;
		//set to zero or keep wrong tally, becuase of fuzz funcs, or fix up total in fuzz funcs?
       		//new_tok->currentTotal=(new_tok->prev->currentTotal) + (new_tok->dataLen);
       		new_tok->currentTotal=0;
        	
		return new_tok;
	}

}
//specifiy current tok, and the number of the tok (in this leg) that you wish to find)
//NULL=error
//good ptr = ok
struct Tok * Find_Tok(struct Tok * tok, u_int tokNum)
{
        struct Tok * new_tok;
	
	//error check
	if(tok == NULL) {
        	TIME(); fprintf(stderr, "{{{Null tok passed to Find_Tok}}}\n"); return(NULL);
        }
	if( (tokNum < 0) || (tokNum > 10000) ) {
        	TIME(); fprintf(stderr, "{{{Invalid tokNum passed to Find_Tok}}}\n"); return(NULL);
        }

        new_tok=_malloc(sizeof(struct Tok));
	new_tok=tok;	

	while (1)
	{
		if(tokNum > tok->tokNum)
		{
			new_tok=new_tok->next;
		}
		else if (tokNum < tok->tokNum)
		{
			new_tok=new_tok->prev;
		}
		else //found
			break;
	}
	if(new_tok == NULL) {
        	TIME(); fprintf(stderr, "{{{Bad new_tok in Find_Tok}}}\n"); return(NULL);
        }
        return new_tok;
}

//give the list of toks starting at the FIRST tok
//returns the number of toks
int Count_Toks(struct Tok * tok)
{
	int count=0;

	//error check
	if(tok == NULL) {
        	TIME(); fprintf(stderr, "{{{Null tok passed to Count_Toks}}}\n"); return(ERROR);
        }

	while(tok->next != NULL)
	{
		count++;
		tok=tok->next;
	}
		count++;

        return count;
}

//slurp from *data to *tok, and create a new token
// -1 = error
// 0 = ok
int Slurp_Into_Tok(struct Tok * tok, u_char * data)
{
	//error check
	if(tok == NULL) {
		TIME(); fprintf(stderr, "{{{Null tok passed to Slurp_Into_Tok}}}\n"); return(-1);
        }
	if(data == NULL) {
		TIME(); fprintf(stderr, "{{{Null data passed to Slurp_Into_Tok}}}\n"); return(-1);
	}
	//malloc and copy data
	tok->data=_malloc(tok->dataLen);
	memcpy(tok->data,data+(tok->currentTotal),tok->dataLen);
  	tok->currentTotal+=(tok->dataLen);
}

//find first tok
//NULL=error
//good ptr = ok
struct Tok * Find_First_Tok(struct Tok * tok)
{
	//error check
	if(tok == NULL) {
		TIME(); fprintf(stderr, "{{{Null tok passed to Find_First_Tok}}}\n"); return(NULL);
        }
	//find it
	while(tok->prev != NULL)
		tok=tok->prev;

	return tok;
}

//find last tok
//NULL=error
//good ptr = ok
struct Tok * Find_Last_Tok(struct Tok * tok)
{
	//error check
	if(tok == NULL) {
		TIME(); fprintf(stderr, "{{{Null tok passed to Find_First_Tok}}}\n"); return(NULL);
        }
	//find it
	while(tok->next != NULL)
		tok=tok->next;

	return tok;
}

//convert a list of toks into a single buffer and totalLen which are passed in as ptrs
//NULL=error
//good ptr = ok
struct Tok * Toks_To_Buf(struct Tok * tok)
{
	//error check
	if(tok == NULL) {
		TIME(); fprintf(stderr, "{{{Null tok passed to Toks_To_Buf}}}\n"); return NULL;
        }
	
	//create a struct tok to hold pkt, len, and currentTotal (i) info
	struct Tok * new_tok=NULL;
	new_tok=Create_Next_Tok(new_tok, 0);
	new_tok->dataLen = 0;

	//calc total len by adding up each tok->dataLen and storing in new_tok->dataLen
	tok=Find_First_Tok(tok);
	while(1)
	{
		new_tok->dataLen+=tok->dataLen;
		if(tok->next == NULL)
			break;
		tok=tok->next;
	}

	//malloc memory
	new_tok->data=_malloc(new_tok->dataLen);

	//cp data from each tok->data to new_tok->data
	tok=Find_First_Tok(tok);
	while(1)
	{
		memcpy((new_tok->data)+new_tok->currentTotal, tok->data, tok->dataLen);
		new_tok->currentTotal+=tok->dataLen;
		if(tok->next == NULL)
			break;
		tok=tok->next;
	}

	return new_tok;
}

//slurp from *data to *tok, and create a new token
// -1 = error
// 0 = ok
// >= 1 = unable to fix len a x lens
int Fix_Lens(struct Tok * tok)
{
	struct Tok * temp_tok=NULL;
	int ret=0, i=0, x=0;
	u_int16_t size8=0;
	u_int16_t size16=0;
	u_int16_t size32=0;
	//error check
	if(tok == NULL) {
		TIME(); fprintf(stderr, "{{{Null tok passed to Fix_Lens}}}\n"); return(-1);
        }
	
	//start at first tok
	tok=Find_First_Tok(tok);

	//scan through the list and fixup lens
	while(1)
	{
		if( tok->type == LEN)
		{
			if( (temp_tok=tok->next) != NULL)
			{
				x=0;
				for(i=0; i<tok->covered; i++)
				{
					if( temp_tok == NULL) {
						ret+=1;
						break;
					}
					x+=temp_tok->dataLen;
					temp_tok=temp_tok->next;
				}
				if( tok->dataLen == 1)
				{
					size8=x;
					_size8_to_ndata(tok->data, size8); 
				}
				else if( tok->dataLen == 2)
				{
					size16=x;
					_size16_to_ndata(tok->data, size16); 
				}
				else if( tok->dataLen == 4)
				{
					size32=x;
					_size32_to_ndata(tok->data, size32); 
				}
				else{
					ret+=1;
				}
			}
			else{
				ret+=1;
			}
		}
		if(tok->next == NULL)
			break;
		tok=tok->next;
	}
	return ret;
}

//num="1" to print just the tok you pass
//num="2" for that one and the next, ...
//num="all" to print all toks
//len: "short" = short print; "long" = long print
//-1 = error
// 0 = ok
int Print_Toks(struct Tok * tok, char * num, char *len)
{
        u_int i, _num;
	void *NumToName[]={"STRING","MIXED","ASCII","ASCII_CMD","ASCII_CMDVAR","ASCII_SPACE","ASCII_END","SESSION_ID","FLAGS","PROTO_ID","LEN","HASH","BINARY","BINARY_END","WIDESTRING_2","WIDESTRING_4","UNDEFINED"};

	//error check
	if(tok == NULL) {
        	TIME(); fprintf(stderr, "{{{Null tok passed to Print_Toks}}}\n"); return(-1);
        }

        if( (strcasecmp("all", num)) == 0)
	{
		tok=Find_First_Tok(tok);
                _num=5000;
	}
        else
                _num=atoi(num);
	if( (_num < 0) || (_num > 80000) ) {
        	TIME(); fprintf(stderr, "{{{Invalid num passed to Print_Toks}}}\n"); return(-1);
        }

        for(i=0; i<_num; i++)
        {
		if( (strcasecmp("long", len)) == 0)
		{
	                printf("tok:%x\n", tok);
	                printf("\ttok->dataLen=%d\n", tok->dataLen);
	                printf("\ttok->data=");
			if(tok->data == NULL)
				printf("NULL\n");
			else { 
				_print(tok->dataLen, AUTO, tok->data); printf("\n"); }
	                printf("\ttok->type=%s\n", NumToName[tok->type]);
	                printf("\ttok->numeric=%d\n", tok->numeric);
	                printf("\ttok->bracket=%d\n", tok->bracket);
	                printf("\ttok->quote=%d\n", tok->quote);
	                printf("\ttok->special=%d\n", tok->special);
	                printf("\ttok->other=%d\n", tok->other);
	                printf("\ttok->covered=%d\n", tok->covered);
	                printf("\ttok->leg=%d\n", tok->leg);
			printf("\ttok->tokNum=%d\n", tok->tokNum);
			printf("\ttok->currentTotal=%d\n", tok->currentTotal);
	                printf("\ttok->prev=%x\n", tok->prev);
			printf("\ttok->next=%x\n", tok->next);
		}
		else if( (strcasecmp("short", len)) == 0)
		{
                	printf("Tok[%d][%d]: type=%13.13s, dataLen=%d, currentTotal=%d, data=", tok->leg, tok->tokNum, NumToName[tok->type], tok->dataLen, tok->currentTotal);
			if(tok->data == NULL)
				printf("NULL\n");
			else
			{
				int pc=AUTO;
				if( tok->type > ASCII_END)
					pc=HEX; 
				if( tok->dataLen < 100) {
					_print(tok->dataLen, pc, tok->data); printf("\n");
				}
				else {
					_print(100, pc, tok->data); printf("...\n");
				}
			}
		}
		else {
        		TIME(); fprintf(stderr, "{{{Bad len option to Print_Toks}}}\n"); return(-1);
        	}
                if( tok->next ==  NULL)
                        break;
                else
                        tok=tok->next;
        }
}

